/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.auditlog;

import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.audit.AuditBuilder.TypeOfInformation;
import dk.stamkort.util.RequestContextVO;

import java.util.Set;
import java.util.function.Supplier;

public class AuditLogger {
    private static final String COMPONENT = "FSK";
    private static final String RETRIEVE_DOCUMENT_SET_CONTEXT = "retrieveDocumentSet";

    private final Supplier<AuditBuilder> auditBuilderSupplier;

    public AuditLogger() {
        this(Audit::createAuditBuilder);
    }

    public AuditLogger(Supplier<AuditBuilder> auditBuilderSupplier) {
        this.auditBuilderSupplier = auditBuilderSupplier;
    }

    public void logRetrieveDocumentSet(Set<String> documentIds, Set<String> patientIds, RequestContextVO requestContext) {
        if (documentIds.isEmpty()) {
            return;
        }
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = RETRIEVE_DOCUMENT_SET_CONTEXT;
        documentIds.forEach(documentId -> log(auditBuilder, context, Type.DOCUMENT_ID, documentId));
        patientIds.forEach(patientId -> log(auditBuilder, context, Type.CPR, patientId));
        logServiceActor(auditBuilder, context, requestContext);
    }

    private void logServiceActor(AuditBuilder auditBuilder, String context, RequestContextVO requestContext) {
        log(auditBuilder, context, Type.ACTOR_CPR, requestContext.getUserCprIdentifier());
        log(auditBuilder, context, Type.ACTOR_USERTYPE, requestContext.getUserType());
    }

    private void log(AuditBuilder auditBuilder, String context, Type type, Object value) {
        if (value != null) {
            auditBuilder.addAuditInformation(COMPONENT, context, type.typeOfInformation, type.value, value);
        }
    }

    private enum Type {
        DOCUMENT_ID("document-id", TypeOfInformation.NonPersonalInformation),
        CPR("cpr", TypeOfInformation.SensitivePersonalInformation),
        ACTOR_CPR("actor-cpr", TypeOfInformation.SensitivePersonalInformation),
        ACTOR_USERTYPE("actor-usertype", TypeOfInformation.RegularPersonalInformation);

        private final String value;
        private final TypeOfInformation typeOfInformation;

        Type(String value, TypeOfInformation typeOfInformation) {
            this.value = value;
            this.typeOfInformation = typeOfInformation;
        }
    }
}
