/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.conf;

import dk.nsp.minlog.api.RegisterService;
import dk.nsp.minlog.api.RegisterServiceProvider;
import dk.nsp.minlog.helper.PropertyFilter;
import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.api.SLALogger;
import dk.stamkort.auditlog.AuditLogger;
import dk.stamkort.health.LastStateHealthIndicator;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.integrations.*;
import dk.stamkort.integrations.fgvhr.FgvhrClient;
import dk.stamkort.integrations.fgvhr.FgvhrRequestBuilder;
import dk.stamkort.integrations.minlog.MinLogClient;
import dk.stamkort.integrations.minlog.MinLogException;
import dk.stamkort.util.DataSourceHealthIndicator;
import dk.stamkort.util.HealthIndicator;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextProvider;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;

import javax.sql.DataSource;
import javax.xml.bind.JAXBException;
import java.io.IOException;

@Configuration
@ComponentScan("dk.stamkort")
public class FskConfig {

    private static final Logger log = LogManager.getLogger(FskConfig.class);

    private static final String IT_SYSTEM_NAME = "Fælles Stamkort";

    public static final String MINLOG_CONFIG_FILE_NAME = "/minlogclient.properties";


    @Bean
    public MinLogClient fskMinLogClient(RegisterService registerService, Properties properties) throws MinLogException {
        return new MinLogClient(registerService, Boolean.parseBoolean(properties.getMinLogOnIdcardLevel3Enabled()));
    }

    @Bean
    public RegisterService registerService() throws MinLogException {
        java.util.Properties properties = new java.util.Properties();
        try {
            log.debug("Loading Minlog configuration from " + MINLOG_CONFIG_FILE_NAME);
            properties.load(MinLogClient.class.getResourceAsStream(MINLOG_CONFIG_FILE_NAME));
        } catch (IOException e) {
            throw new MinLogException("Error reading MinLog configuration from: " + MINLOG_CONFIG_FILE_NAME, e);
        }

        try {
            java.util.Properties filterProperties = new PropertyFilter().filterProperties("kafka.producer.", properties);

            RegisterService registerService = RegisterServiceProvider.getRegisterService();
            registerService.init(properties.get("kafka.topic").toString(), filterProperties);
            return registerService;
        } catch (Exception e) {
            throw new MinLogException("Could not initialize MinLog client", e);
        }
    }

    @Bean
    public OrganDonorClient organDonorClient(Properties properties) throws JAXBException {
        return new OrganDonorClient(
                properties.getOdrEndpoint(),
                properties.getOdrConnectTimeoutMillis(),
                properties.getOdrReadTimeoutMillis());
    }

    @Bean
    public LivingWillClient livingWillClient(Properties properties) throws JAXBException {
        return new LivingWillClient(
                properties.getLtrEndpoint(),
                properties.getLtrConnectTimeoutMillis(),
                properties.getLtrReadTimeoutMillis());
    }

    @Bean
    public TreatmentWillClient treatmentWillClient(Properties properties) throws JAXBException {
        return new TreatmentWillClient(
                properties.getBtrEndpoint(),
                properties.getBtrConnectTimeoutMillis(),
                properties.getBtrReadTimeoutMillis());
    }

    @Bean
    public PersonalDataCardRegisterClient personalDataCardRegisterClient(Properties properties) throws JAXBException {
        return new PersonalDataCardRegisterClient(
                properties.getSkrEndpoint(),
                properties.getSkrConnectTimeoutMillis(),
                properties.getSkrReadTimeoutMillis());
    }

    @Bean
    public YderClient yderClient(Properties properties) throws JAXBException {
        return new YderClient(
                properties.getSyesEndpoint(),
                properties.getSyesConnectTimeoutMillis(),
                properties.getSyesReadTimeoutMillis());
    }


    @Bean
    public SCESClient scesClient(Properties properties) throws JAXBException {
        return new SCESClient(
                properties.getScesEndpoint(),
                properties.getScesConnectTimeoutMillis(),
                properties.getScesReadTimeoutMillis());
    }

    @Bean
    public FgvhrClient fgvhrClient(Properties properties)  {
        TransportUtil transport = new TransportUtil(
                properties.getFgvhrConnectTimeoutMillis(),
                properties.getFgvhrReadTimeoutMillis());
        return new FgvhrClient(
                properties.getFgvhrEndpoint(),
                new DGWSService(
                        properties.getStsEndpoint(),
                        properties.getFgvhrIdcardSubjectName(),
                        properties.getFgvhrIdcardSubjectId(),
                        properties.getFgvhrStsKeystore(),
                        properties.getFgvhrStsKeystorePassword(),
                        properties.getFgvhrIdcardSystemName(),
                        transport),
                transport,
                new FgvhrRequestBuilder()
        );
    }

    @Bean
    public SLALogger slaLogger() {
        return new SLALogConfig("Fælles stamkort", "fsk").getSLALogger();
    }

    /*
     * Actuator HealthIndicator beans:
     * This HealthIndicator beans are automatically added to Actuator and the name of the bean (except for the HealthIndicator
     * suffix) is used as the entry name in the health information.
     */

    @Bean
    public HealthIndicator organDonorClientHealthIndicator(@Qualifier("organDonorClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("organDonorClient");
    }

    @Bean
    public HealthIndicator livingWillClientHealthIndicator(@Qualifier("livingWillClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("livingWillClient");
    }

    @Bean
    public HealthIndicator treatmentWillClientHealthIndicator(@Qualifier("treatmentWillClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("treatmentWillClient");
    }

    @Bean
    public HealthIndicator personalDataCardRegisterClientHealthIndicator(@Qualifier("personalDataCardRegisterClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("personalDataCardRegisterClient");
    }

    @Bean
    public HealthIndicator yderClientHealthIndicator(@Qualifier("yderClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("yderClient");
    }


    @Bean
    public HealthIndicator scesClientHealthIndicator(@Qualifier("scesClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("scesClient");
    }

    @Bean
    public HealthIndicator minLogClientHealthIndicator(@Qualifier("fskMinLogClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build("fskMinLogClient");
    }

    @Bean
    public HealthIndicator fgvhrClientHealthIndicator(@Autowired FgvhrClient client) {
        return LastStateHealthIndicator.reporter(client).build(FgvhrClient.class.getSimpleName());
    }

    @Bean
    public DataSourceHealthIndicator fskDataSourceHealthIndicator(@Qualifier("fsk")DataSource fskDatasource) {
        return new DataSourceHealthIndicator("fsk", fskDatasource);
    }

    @Bean
    public DataSourceHealthIndicator stamdataDataSourceHealthIndicator(@Qualifier("stm")DataSource stmDatasource) {
        return new DataSourceHealthIndicator("stm", stmDatasource);
    }

    @Bean
    RequestContextProvider requestContextProvider() {
        return new RequestContextProvider();
    }

    @Bean
    public AuditLogger auditLogger() {
        return new AuditLogger();
    }
}
