/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.dao.impl;

import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.dao.impl.abstr.AbstractDAOImpl;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.util.Properties;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.stereotype.Repository;

import java.sql.Types;
import java.time.Instant;
import java.util.List;

@Repository
public class RegistryIndexDAOImpl extends AbstractDAOImpl implements RegistryIndexDAO {

    @Autowired
    Properties properties;

    private final String INSERT_SQL = "INSERT INTO RegistryIndex (FSKDocumentId, TypeCode, FormatCode, PersonIdentifier, FirstName, LastName, HashValue, DeleteAfter) " +
            "VALUES (:FSKDocumentId, :TypeCode, :FormatCode, :PersonIdentifier, :FirstName, :LastName, :HashValue, :DeleteAfter)";

    @Override
    public RegistryIndexVO getRegistryIndexVO(String cpr) throws DAOException {
        try {
            return fetchSingleRowMatchingSingleColumn("PersonIdentifier", cpr);
        } catch (DAOException e) {
            throw new DAOException("An error occurred querying " + getTableName() + " with PersonIdentifier: " + cpr, e);
        }
    }

    @Override
    public Integer createRegistryIndexVO(RegistryIndexVO registryIndexVO) throws DAOException {
        final Integer pid = createSingleObject(INSERT_SQL, registryIndexVO);
        registryIndexVO.setRegistryIndexPid(pid);
        return pid;
    }

    @Override
    public void markForDeletion(String cpr, Instant deleteAfter) throws DAOException {
        try {
            String sql = "UPDATE " + getTableName() + " SET DeleteAfter = :deleteAfter WHERE PersonIdentifier = :cpr";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();

            parameters.addValue("deleteAfter", java.sql.Timestamp.from(deleteAfter), Types.TIMESTAMP);
            parameters.addValue("cpr", cpr);
            executeUpdate(sql, parameters);
        } catch (DAOException e) {
            throw new DAOException("An error occurred updating " + getTableName() + " with cpr: " + cpr + " and deleteAfter: " + deleteAfter, e);
        }
    }
    
    @Override
    public List<RegistryIndexVO> getFSKDocumentIdsToBeDeleted(Instant jobStartTime) throws DAOException {
        try {
            String sql = "SELECT * FROM " + getTableName() + " WHERE deleteAfter < :deleteDate";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            
            parameters.addValue("deleteDate", java.sql.Timestamp.from(jobStartTime), Types.TIMESTAMP);
            return queryForList(sql, parameters);
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred counting from " + getTableName(), e);
        }
    }

    @Override
    public void deleteByDocumentId(String fskDocumentId) throws DAOException {
        try {
            String sql = "DELETE FROM " + getTableName() + " WHERE FSKDocumentId = :FSKDocumentId";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("FSKDocumentId", fskDocumentId);
            executeUpdate(sql, parameters);
        } catch (DAOException e) {
            throw new DAOException("An error occurred deleting from " + getTableName(), e);
        }
    }

    @Override
    public RegistryIndexVO getRegistryIndexByDocumentId(String fskDocumentId) throws DAOException {
        try {
            return fetchSingleRowMatchingSingleColumn("FSKDocumentId", fskDocumentId);
        } catch (DAOException e) {
            throw new DAOException("An error occurred querying " + getTableName() + " with FSKDocumentId: " + fskDocumentId, e);
        }
    }

    @Override
    public String getTableName() {
        return "RegistryIndex";
    }

    @Override
    protected String getIdColumnName() {
        return "RegistryIndexPid";
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new RegistryIndexRowMapper();
    }
}
