/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.dao.impl.abstr;

import dk.stamkort.dao.exceptions.DAOException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSourceUtils;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.util.ArrayList;
import java.util.List;

/**
 * Base class for DAO implementations
 */
public abstract class AbstractDAOImpl {

    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;

    @Autowired
    private DataSource dataSource;

    @PostConstruct
    private void init() {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }

    public abstract String getTableName();
    protected abstract String getIdColumnName();
    protected abstract RowMapper getDefaultRowMapper();

    protected <T> Integer createSingleObject(String insertSql, T object) throws DAOException {
        if (object == null) {
            return null;
        }
        try {
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            ArrayList<T> a = new ArrayList<>();
            a.add(object);
            SqlParameterSource[] params = SqlParameterSourceUtils.createBatch(a.toArray());
            namedParamJdbcTemplate.update(insertSql, params[0], generatedKeyHolder);
            return generatedKeyHolder.getKey().intValue();
        } catch (Exception e) {
            throw new DAOException("An error occurred creating object: " + insertSql, e);
        }
    }

    protected void executeUpdate(String updateSql, SqlParameterSource parameters) throws DAOException {
        try {
            final long startTime = System.nanoTime();
            namedParamJdbcTemplate.update(updateSql, parameters);
            recordMetrics(updateSql, startTime);
        } catch (Exception e) {
            throw new DAOException("An error occurred executing update: " + updateSql, e);
        }
    }

    protected <T> T fetchSingleRowMatchingSingleColumn(String columnName, String filterValue) throws DAOException {
        List<T> result = fetchRowsMatchingSingleColumn(columnName, filterValue);
        if (result.size() == 0) {
            return null;
        } else if (result.size() > 1) {
            throw new DAOException("Multiple rows were found in table " + getTableName()
                    + " looking in column " + columnName + " for the value " + filterValue);
        } else {
            return result.get(0);
        }
    }

    protected <T> List<T> fetchRowsMatchingSingleColumn(String columnName, String filterValue) throws DAOException {
        final MapSqlParameterSource parms = new MapSqlParameterSource().addValue("val", filterValue);
        return queryForList("SELECT * FROM " + getTableName() + " WHERE " + columnName + "=:val", parms);
    }

    protected <T> List<T> queryForList(String sql, SqlParameterSource parms) throws DAOException {
        return queryForList(sql, parms, getDefaultRowMapper());
    }

    protected <T> List<T> queryForList(String sql, SqlParameterSource parms, RowMapper rowMapper) throws DAOException {
        try {
            final long startTime = System.nanoTime();
            final List<T> list = namedParamJdbcTemplate.query(sql, parms, rowMapper);
            recordMetrics(sql, startTime);
            return list;
        } catch (Exception e) {
            throw new DAOException("An error occurred executing query: " + sql, e);
        }
    }

    private void recordMetrics(String sql, long startTime) {
        final String metricName = sql.substring(0, Math.min(sql.length(), 50));
        final double durationMillis = (System.nanoTime() - startTime)/1000000;
    }

}
