/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Properties;
import java.util.UUID;

public class DGWSService {
	private static final Logger logger = Logger.getLogger(DGWSService.class);

	private final String endpoint;
	private final String organizationName;
	private final String organizationCode;
	private final String systemName;
	private final SOSIFactory sosiFactory;
	private final TransportUtil httpClientService;
	private SystemIDCard stsSignedIdCard;

	public DGWSService(String endpoint, String organizationName, String organizationCode, String keystorePath,
					   String keystorePassword, String systemName, TransportUtil httpClientService) {
		this.endpoint = endpoint;
		this.organizationName = organizationName;
		this.organizationCode = organizationCode;
		this.systemName = systemName;
		this.httpClientService = httpClientService;

		Properties properties = SignatureUtil.setupCryptoProviderForJVM();
		properties.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));
		CredentialVault vault = getVault(properties, keystorePath, keystorePassword);

		if (Boolean.parseBoolean(System.getenv("NSP_TEST_FEDERATION")) || Boolean.getBoolean(DGWSService.class.getName() + ".test")) {
			logger.warn("Using NSP Test Federation");
			sosiFactory = getSosiFactoryTestFederation(vault, properties);
		} else {
			logger.info("Using NSP Production Federation");
			sosiFactory = getSosiFactoryFederation(vault, properties);
		}
	}

	public void ensureValidIdCard() {
		if (stsSignedIdCard != null) {
			Instant expiryDate = stsSignedIdCard.getExpiryDate().toInstant();
			if (expiryDate.minus(5, ChronoUnit.HOURS).isAfter(Instant.now())) {
				// We have a valid IDCard with at least 5 hours remaining.
				return;
			} else {
				logger.info("Renewing IDCard from " + endpoint);
			}
		} else {
			logger.info("Getting first IDCard from " + endpoint);
		}

		try {
			Document securityTokenRequest = buildSecurityTokenRequest();
			String requestXml = XmlUtil.node2String(securityTokenRequest, false, true);
			String responseXml = httpClientService.sendRequest(endpoint, "\"Issue\"", requestXml);
			stsSignedIdCard = buildSystemIDCard(responseXml);

			logger.info("Got new IDCard based on '" + stsSignedIdCard.getAlternativeIdentifier() + "' issued by " + stsSignedIdCard.getIssuer());
		} catch (Exception e) {
			throw new IllegalStateException(e);
		}
	}

	public Request createRequest() {
		Request sosiRequest = sosiFactory.createNewRequest(false, UUID.randomUUID().toString());
		sosiRequest.setIDCard(stsSignedIdCard);
		sosiRequest.setDGWSVersion("1.0.1");
		Document sosiRequestDocument = sosiRequest.serialize2DOMDocument();
		sosiRequest.setBody(sosiRequestDocument.getDocumentElement());
		return sosiRequest;
	}

	protected CredentialVault getVault(Properties properties, String keystorePath, String keystorePassword) {
		return new ClasspathCredentialVault(properties, keystorePath, keystorePassword);
	}

	protected SOSIFactory getSosiFactoryTestFederation(CredentialVault vault, Properties properties) {
		return new SOSIFactory(new SOSITestFederation(new Properties()), vault, properties);
	}

	protected SOSIFactory getSosiFactoryFederation(CredentialVault vault, Properties properties) {
		return new SOSIFactory(new SOSIFederation(new Properties()), vault, properties);
	}

	protected SystemIDCard buildSystemIDCard(String responseXml) {
		SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);
		SystemIDCard idCard = (SystemIDCard) securityTokenResponse.getIDCard();

		boolean isFault = securityTokenResponse.isFault();
		if (isFault || idCard == null) {
			logger.error("We got a SOAP Fault from the STS. The detailed message can be seen in the xml response.");
			if (isFault) {
				logger.error("A SOAP Fault was received: " + securityTokenResponse.getFaultString());
			}
			throw new IllegalStateException("SOAP Fault from the STS");
		}

		return idCard;
	}

	protected Document buildSecurityTokenRequest() {
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, organizationCode, organizationName);
		X509Certificate certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();
		SystemIDCard systemIdCard = sosiFactory.createNewSystemIDCard(systemName, careProvider,
				AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, certificate, null);

		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(systemIdCard);
        return securityTokenRequest.serialize2DOMDocument();
	}
}
