/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.time.Instant;
import java.util.Optional;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.TransformerException;

import org.apache.commons.io.Charsets;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.util.RequestContextEnricher;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.SlaLogger;

public abstract class ReusingDGWSClient<REQUEST, RESPONSE> implements LastStateHealthReporter {
	private static final Logger log = LogManager.getLogger(ReusingDGWSClient.class);

	private String name;
	private Class<REQUEST> requestClass;
	private Class<RESPONSE> responseClass;
	private SlaLogger slaLogger;
	private String endpoint;
	private TransportUtil transport;

	private Boolean lastExecutionSucceeded;
	private Instant timeOfLastExecution;
	private Exception lastException;

	private JAXBContext jaxbContext;

	public ReusingDGWSClient(String name, Class<REQUEST> requestClass, Class<RESPONSE> responseClass,
							 String endpoint, int connectTimeoutMillis, int readTimeoutMillis) throws JAXBException {
		this.name = name;
		this.requestClass = requestClass;
		this.responseClass = responseClass;
		this.slaLogger = new SlaLogger(name);
		this.transport = new TransportUtil(connectTimeoutMillis, readTimeoutMillis);
		this.endpoint = endpoint;

		jaxbContext = JAXBContext.newInstance(Header.class, requestClass, responseClass);
	}

	public RESPONSE call(REQUEST request, RequestContextVO requestContext) throws DGWSClientException {

		JAXBElement<REQUEST> bodyJaxbElm = new JAXBElement<>(getRequestQName(), requestClass, request);

		return call(bodyJaxbElm, requestContext);
	}

	public RESPONSE call(JAXBElement<REQUEST> request, RequestContextVO requestContext) throws DGWSClientException {
		long startMillis = System.currentTimeMillis();
		long startNanos = System.nanoTime();
		try {
			String responseXml = callInternal(request, requestContext.getRequestContextEnricher());
			addSlaLog(requestContext, startNanos, startMillis, null);

			RESPONSE response = fromXml(responseXml);

			lastExecutionSucceeded = true;
			timeOfLastExecution = Instant.now();

			return response;
		} catch (Exception e) {
			String errorMessage = "Error calling " + name;
			String msg = e.getMessage();
			if (msg != null) {
				errorMessage += ":" + msg;
			}
			addSlaLog(requestContext, startNanos, startMillis, errorMessage);

			lastExecutionSucceeded = false;
			timeOfLastExecution = Instant.now();
			lastException = e;

			throw new DGWSClientException(errorMessage, e);
		}
	}

	public String callInternal(JAXBElement<REQUEST> request, RequestContextEnricher requestContextEnricher) throws IOException, JAXBException, SOAPException, TransformerException, ParserConfigurationException {

		DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
		Document document = documentBuilder.newDocument();

		jaxbContext.createMarshaller().marshal(request, document);

		Optional<String> requestAsString = requestContextEnricher.createRequestWithContext(document.getDocumentElement());

		log.debug(name + " request: " + requestAsString.orElse(null));
		String responseXml = transport.sendRequest(endpoint, getSOAPAction(), requestAsString.get());
		log.debug(name + " response: " + responseXml);

		return responseXml;
	}

	protected abstract QName getRequestQName();

	protected abstract String getSOAPAction();

	private RESPONSE fromXml(String xml) throws SOAPException, IOException, JAXBException, DGWSClientException {
		SOAPMessage soapMessage = MessageFactory.newInstance().createMessage(null, new ByteArrayInputStream(xml.getBytes(Charsets.UTF_8)));
		SOAPBody soapBody = soapMessage.getSOAPBody(); // throws an exception if body is malformed
		log.debug("SOAPMessage parsed successfully");

		if (soapBody.hasFault()) {
			SOAPFault fault = soapBody.getFault();
			throw new DGWSClientException("Service " + name + " returned fault: " + fault.getFaultString());
		}
		JAXBElement<RESPONSE> jaxbResponse = jaxbContext.createUnmarshaller().unmarshal(soapBody.getFirstChild(), responseClass);

		return jaxbResponse.getValue();
	}

	private void addSlaLog(RequestContextVO requestContext, long startNanos, long startMillis, String errorMessage) {
		if (slaLogger != null) {
			slaLogger.addSlaLog(requestContext, startNanos, startMillis, errorMessage);
		}
	}

	@Override
	public Boolean lastExecutionSucceeded() {
		return lastExecutionSucceeded;
	}

	@Override
	public Instant timeOfLastExecution() {
		return timeOfLastExecution;
	}

	@Override
	public Exception lastException() {
		return lastException;
	}
}
