/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.fgvhr;

import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.integrations.DGWSService;
import dk.stamkort.integrations.TransportUtil;
import dk.sundhedsdatastyrelsen.fravalg_af_genoplivning._2023._06._01.InputIdentifierType;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.ws.soap.SOAPFaultException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.time.Instant;

import static java.util.Optional.ofNullable;

public class FgvhrClient implements LastStateHealthReporter {
	private static final Logger logger = Logger.getLogger(FgvhrClient.class);

	private final DGWSService dgwsService;
	private final TransportUtil httpClientService;
	private final FgvhrRequestBuilder requestBuilder;
	private final JAXBContext jaxbContext;
	private final DocumentBuilder documentBuilder;

	private final String endpoint;
	private Instant timeOfLastExecution;
	private Boolean lastExecutionSucceeded;
	private Exception lastException;

	public FgvhrClient(String endpoint, DGWSService dgwsService, TransportUtil httpClientService, FgvhrRequestBuilder requestBuilder) {
		this.dgwsService = dgwsService;
		this.httpClientService = httpClientService;
		this.requestBuilder = requestBuilder;
		this.endpoint = endpoint;

		try {
			this.jaxbContext = JAXBContext.newInstance(InputIdentifierType.class);

		} catch (JAXBException e) {
			throw new IllegalStateException(e);
		}

		try {
			DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
			documentBuilderFactory.setNamespaceAware(true);
			this.documentBuilder = documentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	public boolean hasNoResuscitation(String patientId) throws FgvhrClientException {
		try {
			boolean response = callService(requestBuilder.createHasNoResuscitationRequest(patientId));
			lastExecutionSucceeded = true;
			return response;
		}
		catch (Exception e) {
			lastException = e;
			lastExecutionSucceeded = false;
			throw e;
		}
		finally { timeOfLastExecution = Instant.now(); }
	}

	private boolean callService(InputIdentifierType requestBody) throws FgvhrClientException {
		try {
			// Tilføj DGWS-headeren
			dgwsService.ensureValidIdCard();
			Request req = dgwsService.createRequest();

			// Lav rod elementet
			JAXBElement<InputIdentifierType> jaxbRootElement =
					new JAXBElement<>(
							new QName(SoapAction.HAS_NO_RESUSCITATION.getSoapNamespace(), SoapAction.HAS_NO_RESUSCITATION.getSoapRootElement()),
							InputIdentifierType.class,
							requestBody);

			// Marshal elementet til et DOM-dokument
			Document document = documentBuilder.newDocument();
			Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
			jaxbMarshaller.marshal(jaxbRootElement, document);

			// Tilføj elementet til requestet
			req.setBody(document.getDocumentElement());
			Document requestDocument = req.serialize2DOMDocument();

			// Send requestet
			String requestXml = XmlUtil.node2String(requestDocument, false, true);
			String responseXml = httpClientService.sendRequest(endpoint, SoapAction.HAS_NO_RESUSCITATION.getSoapAction(), requestXml);

			logger.debug(responseXml);

            return unmarshalResponse(responseXml);
		} catch (JAXBException | IOException | SOAPException e) {
            throw new FgvhrClientException(e.getMessage(), e);
        }
    }

	private boolean unmarshalResponse(String responseXml)
			throws SOAPException, IOException, JAXBException {
		SOAPBody soapBody = MessageFactory.newInstance()
				.createMessage(null, new ByteArrayInputStream(responseXml.getBytes()))
				.getSOAPBody();

		ofNullable(soapBody.getFault())
				.map(SOAPFaultException::new)
				.ifPresent(ex -> { throw ex; });

		Document message = soapBody
				.extractContentAsDocument();

        return jaxbContext.createBinder()
				.unmarshal(message, Boolean.class).getValue();
	}

	@Override
	public Boolean lastExecutionSucceeded() { return lastExecutionSucceeded; }

	@Override
	public Instant timeOfLastExecution() { return timeOfLastExecution; }

	@Override
	public Exception lastException() { return lastException; }
}