/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.minlog;

import java.time.Instant;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import dk.nsp.minlog.api.RegisterService;
import dk.nsp.minlog.domain.Destination;
import dk.nsp.minlog.domain.LogEntry;
import dk.nsp.minlog.domain.RegistrationResult;
import dk.nsp.minlog.domain.Source;
import dk.stamkort.domain.OrganizationTypeVO;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.RequestContextVO.UserType;
import dk.stamkort.util.SlaLogger;

public class MinLogClient implements LastStateHealthReporter {

    private static final Logger log = LogManager.getLogger(MinLogClient.class);

    private final RegisterService registerService;
    private SlaLogger slaLogger = new SlaLogger("MinLog");

    private boolean minLogOnIdcardLevel3Enabled;
    private Boolean lastExecutionSucceeded;
    private Instant timeOfLastExecution;
    private Exception lastException;


    public MinLogClient(RegisterService registerService, boolean minLogOnIdcardLevel3Enabled) {
        this.registerService = registerService;
        this.minLogOnIdcardLevel3Enabled = minLogOnIdcardLevel3Enabled;
    }

    void logDataAdd(RequestContextVO context, MinLogData minLogData) throws MinLogException {
        LogEntry minlogEntry;

        if (UserType.SystemUser.equals(context.getUserType()) && !minLogOnIdcardLevel3Enabled) {
            log.info("Skipper kald af minlog for systembruger");
            return;
        }
        try {
            throwIfNull(minLogData.getCitizenCpr(), "CPR number must be specified for the person whose data is being viewed/changed.");
            throwIfNull(minLogData.getActivity(), "The activity being performed must be specified");
            throwIfNull(minLogData.getEventDateTime(), "Event time must be specified");
            throwIfNull(minLogData.getSourceSystemIdentifier(), "The source system must be specified");
            throwIfNull(minLogData.getUserCPR(), "The user CPR must be specified");
        } catch (IllegalArgumentException e) {
            throw new MinLogException("Invalid argument when calling the MinLog service", e);
        }

        try {
            minlogEntry = createLogDataEntry(minLogData);
        } catch (Exception e) {
            throw new MinLogException("Error creating MinLog data.", e);
        }

        long startMillis = System.currentTimeMillis();
        long startNanos = System.nanoTime();
        List<LogEntry> logData = new ArrayList<>();
        logData.add(minlogEntry);
        RegistrationResult result = registerService.addRegistrations(logData);
        registerService.flush();

        if(result.getNumberFailed() > 0) {
            String errorMessage = "Der opstod en fejl under invokeringen af MinLog registreringsservice. Failed log entry: ";
            log.warn(errorMessage);
            String msg = result.getFailedLogDataEntries().toString();
            if (msg != null) {
                errorMessage += ": " + msg;
            }
            addSlaLog(context, startNanos, startMillis, errorMessage);
            lastExecutionSucceeded = false;
            timeOfLastExecution = Instant.now();
            lastException = new MinLogException(errorMessage);
            throw new MinLogException(errorMessage);
        }
        lastExecutionSucceeded = true;
        timeOfLastExecution = Instant.now();
        addSlaLog(context, startNanos, startMillis, null);
    }

    private LogEntry createLogDataEntry(MinLogData minLogData) {
        final LogEntry logEntry;
        logEntry = new LogEntry();
        logEntry.setDestination(new Destination());
        logEntry.setSource(new Source());

        logEntry.getDestination().setCriticality("Ingen");
        logEntry.getDestination().setPersonIdentifier(minLogData.getCitizenCpr());
        logEntry.getDestination().setPersonIdentifierSrc("CPR");
        logEntry.getDestination().setActivity(minLogData.getActivity());
        logEntry.getDestination().setDateTime(minLogData.getEventDateTime());
        logEntry.getDestination().setSystemName(minLogData.getSourceSystemIdentifier());
        logEntry.getDestination().setUserPersonIdentifier(minLogData.getUserCPR());
        logEntry.getDestination().setUserPersonIdentifierSrc("CPR");

        if (minLogData.getOrganizationId() != null) {
            logEntry.getDestination().setOrganisationId(minLogData.getOrganizationId());
            logEntry.getDestination().setOrganisationIdSrc(getMinlogOrganizationType(minLogData.getOrganizationType()));
        }

        if (minLogData.getOrganizationName() != null) {
            logEntry.getDestination().setOrganisationName(minLogData.getOrganizationName());
        }

        if (minLogData.getOnBehalfOfCpr() != null) {
            logEntry.getDestination().setOnBehalfOfPersonIdentifier(minLogData.getOnBehalfOfCpr());
            logEntry.getDestination().setOnBehalfOfPersonIdentifierSrc("CPR");
        }
        return logEntry;
    }

    private static String getMinlogOrganizationType(OrganizationTypeVO orgType) {
        switch (orgType) {
            case COMMUNALNUMBER:
            case CVRNUMBER:
            case PNUMBER:
            case SKS:
            case SOR:
            case YDERNUMMER:
                return orgType.toString();
            default:
                throw new RuntimeException("Unexpected organization type " + orgType);
        }
    }

    private <T> void throwIfNull(T value, String errorMessage) throws IllegalArgumentException {
        if (value == null) {
            throw new IllegalArgumentException(errorMessage);
        }
    }

    private void addSlaLog(RequestContextVO context, long startNanos, long startMillis, String errorMessage) {
        if (slaLogger != null) {
            slaLogger.addSlaLog(context, startNanos, startMillis, errorMessage);
        }
    }

    @Override
    public Boolean lastExecutionSucceeded() {
        return lastExecutionSucceeded;
    }

    @Override
    public Instant timeOfLastExecution() {
        return timeOfLastExecution;
    }

    @Override
    public Exception lastException() {
        return lastException;
    }
}
