/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.service;

import java.time.ZonedDateTime;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;

import dk.stamkort.domain.Dentist;
import dk.stamkort.domain.GeneralPractitioner;
import dk.stamkort.integrations.fgvhr.FgvhrClient;
import dk.stamkort.integrations.fgvhr.FgvhrClientException;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.HasTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.v3.II;
import org.hl7.v3.POCDMT000040Component3;
import org.hl7.v3.POCDMT000040Entry;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

// Note: No references to HL7 namespaces in this file.

import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.ChildCustodyCustodyData;
import dk.stamkort.integrations.CustodyByCustodyData;
import dk.stamkort.integrations.CustodyData;
import dk.stamkort.integrations.DGWSClientException;
import dk.stamkort.integrations.LivingWillClient;
import dk.stamkort.integrations.OrganDonorClient;
import dk.stamkort.integrations.PersonalDataCardRegisterClient;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.integrations.SCESClient;
import dk.stamkort.integrations.TreatmentWillClient;
import dk.stamkort.integrations.YderClient;
import dk.stamkort.util.FSKException;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.YderNrUtil;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.HasLivingWillResponse;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.HasOrganDonorRegistrationResponse;
import oio.medcom.cprservice._1_0.CurrentGeneralPractitionerStructureType;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import oio.medcom.cprservice._1_0.PersonCustodyRelationType;
import oio.medcom.cprservice._1_0.PersonCustodyStructureType;
import oio.medcom.cprservice._1_0.PersonHealthCareInformationStructureType;
import oio.medcom.cprservice._1_0.PersonInformationStructureType;
import oio.medcom.cprservice._1_0.PersonWardRelationType;
import oio.medcom.cprservice._1_0.PersonWardStructureType;
import oio.medcom.cprservice._1_0.PersonWithHealthCareInformationStructureType;

@Service
public class PersonalDataCardServiceImpl implements PersonalDataCardService {
	private static final Logger log = LogManager.getLogger(PersonalDataCardServiceImpl.class);
	private static ExecutorService executorService;

	@Autowired
	private Properties properties;

	@Autowired
	public OrganDonorClient organDonorClient;

	@Autowired
	public LivingWillClient livingWillClient;

	@Autowired
	public TreatmentWillClient treatmentWillClient;

	@Autowired
	public PersonalDataCardRegisterClient personalDataCardRegisterClient;

	@Autowired
	public SCESClient cpre;

	@Autowired
	public YderClient yderClient;

	@Autowired
	public FgvhrClient fgvhrClient;

	@Value("${testing.disableAccessCheck:false}")
	private boolean disableAccessCheck;

	@Value("${testing.disableMinLogging:false}")
	private boolean disableMinLogging;



	@PostConstruct
	public void init() {
		executorService = Executors.newFixedThreadPool(properties.getDgwsClientPoolSize());
	}

	@Override
	public RegisterData getRegisterData(String cprnr, RequestContextVO requestContext) {
		RegisterData data = new RegisterData(cprnr);

		// call services in parallel
		CallWrapper scesCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isScesIntegrationEnabled()) {
					fetchSCESData(cprnr, data, requestContext);
				}
			}
		};
		scesCallable.submit();

		CallWrapper skrCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isSkrIntegrationEnabled()) {
					fetchSKRData(cprnr, data, requestContext);
				}
			}
		};
		skrCallable.submit();

		CallWrapper ltrCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isLtrIntegrationEnabled()) {
					fetchLTRData(cprnr, data, requestContext);
				}
			}
		};
		ltrCallable.submit();

		CallWrapper btrCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isBtrIntegrationEnabled()) {
					if (ZonedDateTime.now().isAfter(properties.getBtrIntegrationStartDateTime())) {
						fetchBTRData(cprnr, data, requestContext);
					} else {
						log.debug("Skipping BTR call (enabled " + properties.getBtrIntegrationStartDateTime() + ")");
					}
				}
			}
		};
		btrCallable.submit();

		CallWrapper odrCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isOdrIntegrationEnabled()) {
					fetchODRData(cprnr, data, requestContext);
				}
			}
		};
		odrCallable.submit();

		CallWrapper fgvhrCallable = new CallWrapper() {
			@Override
			public void callClient() throws DGWSClientException {
				if (properties.isFgvhrIntegrationEnabled()) {
					try {
						fetchFGVHRData(cprnr, data);
					} catch (FgvhrClientException e) {
						throw new DGWSClientException(e.getMessage(), e);
					}
				}
			}
		};
		fgvhrCallable.submit();

		// wait for termination of threads
		scesCallable.waitForTermination(FaultCode.CALL_SCES_ERROR);
		skrCallable.waitForTermination(FaultCode.CALL_PERSONAL_DATA_CARD_REGISTER_ERROR);
		ltrCallable.waitForTermination(FaultCode.CALL_LIVING_WILL_ERROR);
		btrCallable.waitForTermination(FaultCode.CALL_TREATMENT_WILL_ERROR);
		odrCallable.waitForTermination(FaultCode.CALL_ORGAN_DONOR_REGISTER_ERROR);
		fgvhrCallable.waitForTermination(FaultCode.CALL_FGVHR_ERROR);

		// check for exceptions
		scesCallable.checkForExceptions(FaultCode.CALL_SCES_ERROR);
		skrCallable.checkForExceptions(FaultCode.CALL_PERSONAL_DATA_CARD_REGISTER_ERROR);
		ltrCallable.checkForExceptions(FaultCode.CALL_LIVING_WILL_ERROR);
		btrCallable.checkForExceptions(FaultCode.CALL_TREATMENT_WILL_ERROR);
		odrCallable.checkForExceptions(FaultCode.CALL_ORGAN_DONOR_REGISTER_ERROR);

		return data;
	}

	private void fetchSCESData(String cprnr, RegisterData data, RequestContextVO requestContext) throws DGWSClientException {
		log.debug("Fetching SCES data for " + cprnr);
		GetPersonWithHealthCareInformationOut scesData = cpre.call(cprnr, requestContext);
		if (scesData != null) {
			data.setScesData(scesData);

			// get custodies
			PersonWithHealthCareInformationStructureType personWithHealthCareInformationStructure = scesData.getPersonWithHealthCareInformationStructure();
			if (personWithHealthCareInformationStructure != null) {
				PersonInformationStructureType personInformationStructure = personWithHealthCareInformationStructure.getPersonInformationStructure();
				if (personInformationStructure != null) {
					// Fetch ChildCustody custody data
					List<ChildCustodyCustodyData> childCustodyCustodyDataList = personInformationStructure.getPersonWardStructure()
							.stream()
							.filter(PersonalDataCardServiceImpl::filterChildren)
							.map(PersonalDataCardServiceImpl::mapChildCustodyCustodyData)
							.collect(Collectors.toList());
					fetchSCESCustodyData(requestContext, childCustodyCustodyDataList)
					.forEach(data::addChildCustodyCustodyData);

					// Fetch CustodyBy custody data
					List<CustodyByCustodyData> custodyByCustodyDataList = personInformationStructure.getPersonCustodyStructure()
							.stream()
							.filter(PersonalDataCardServiceImpl::filterParents)
							.map(PersonalDataCardServiceImpl::mapCustodyByCustodyData)
							.collect(Collectors.toList());
					fetchSCESCustodyData(requestContext, custodyByCustodyDataList)
					.forEach(data::addCustodyByCustodyData);
				}

				PersonHealthCareInformationStructureType personHealthCareInformationStructureType = personWithHealthCareInformationStructure.getPersonHealthCareInformationStructure();
				if (personHealthCareInformationStructureType != null) {

					CurrentGeneralPractitionerStructureType currentGeneralPractitionerStructureType = personHealthCareInformationStructureType.getCurrentGeneralPractitionerStructure();
					if (currentGeneralPractitionerStructureType != null) {
						// Lookup GP by ydernr
						String yderNr = currentGeneralPractitionerStructureType.getAssociatedGeneralPractitionerStructure().getAssociatedGeneralPractitionerIdentifier().toString();
						String paddedYderNr = YderNrUtil.padZero(yderNr);
						YderVO responseYder = yderClient.call(paddedYderNr, requestContext);
						if (responseYder != null) {
							data.addYder(paddedYderNr, responseYder);
							data.setGeneralPractitioner(new GeneralPractitioner(responseYder));
						}
					}
				}
			}
		}
	}

	private static ChildCustodyCustodyData mapChildCustodyCustodyData(PersonWardStructureType personWardStructureType) {
		ChildCustodyCustodyData custodyData = new ChildCustodyCustodyData();
		custodyData.setCpr(personWardStructureType.getPersonWardIdentifier());
		custodyData.setPersonWardRelationType(personWardStructureType.getPersonWardRelation());
		return custodyData;
	}

	private static CustodyByCustodyData mapCustodyByCustodyData(PersonCustodyStructureType personCustodyStructureType) {
		CustodyByCustodyData custodyData = new CustodyByCustodyData();
		custodyData.setCpr(personCustodyStructureType.getPersonCustodyIdentifier());
		custodyData.setPersonCustodyRelationType(personCustodyStructureType.getPersonCustodyRelation());
		return custodyData;
	}

	private static boolean filterChildren(PersonWardStructureType personWardStructureType) {
		return personWardStructureType.getPersonWardRelation() == PersonWardRelationType.BARN;
	}

	private static boolean filterParents(PersonCustodyStructureType personCustodyStructureType) {
		return personCustodyStructureType.getPersonCustodyRelation() == PersonCustodyRelationType.FAR
				|| personCustodyStructureType.getPersonCustodyRelation() == PersonCustodyRelationType.MOR
				|| personCustodyStructureType.getPersonCustodyRelation() == PersonCustodyRelationType.ANDEN;
	}

	private <T extends CustodyData> List<T> fetchSCESCustodyData(RequestContextVO requestContext, List<T> custodyDataList) {
		// load custody data in parallel
		LinkedList<CustodyCallWrapper<T>> callableList = new LinkedList<>();
		for (T custodyData : custodyDataList) {
			CustodyCallWrapper<T> callable = new CustodyCallWrapper<>(requestContext, custodyData);
			callableList.add(callable);
			callable.submit();
		}

		// wait for termination of threads
		for (CustodyCallWrapper<T> callable : callableList) {
			callable.waitForTermination(FaultCode.CALL_SCES_CUSTODY_ERROR);
		}

		// check for exceptions
		for (CustodyCallWrapper<T> callable : callableList) {
			callable.checkForExceptions(FaultCode.CALL_SCES_CUSTODY_ERROR);
		}

		// fetch custody data
		return callableList.stream()
				.filter(Objects::nonNull)
				.map(custodyCallWrapper -> custodyCallWrapper.custodyData)
				.collect(Collectors.toList());
	}

	private void fetchSKRData(String cprnr, RegisterData data, RequestContextVO requestContext) throws DGWSClientException {
		log.debug("Feching PersonalDataCard register data for " + cprnr);
		GetPersonalDataCardResponse response = personalDataCardRegisterClient.call(cprnr, requestContext);
		if (response != null) {
			data.setClinicalDocument(response.getClinicalDocument());

			// lookup dentists by ydernr
			if (response.getClinicalDocument() != null && response.getClinicalDocument().getComponent() != null && response.getClinicalDocument().getComponent().getStructuredBody() != null) {
				Set<String> yderNrSet = getAndPadYderNr(response);
				for (String yderNr : yderNrSet) {
					YderVO responseYder = yderClient.call(yderNr, requestContext);
					if (responseYder != null) {
						data.addYder(yderNr, responseYder);
						data.setDentist(new Dentist(responseYder));
					}
				}
			}
		}
	}

	private Set<String> getAndPadYderNr(GetPersonalDataCardResponse response) throws RuntimeException {
		Set<String> yderNrSet = new HashSet<>();
		for (POCDMT000040Component3 component3 : response.getClinicalDocument().getComponent().getStructuredBody()
				.getComponents()) {
			if (component3 != null && component3.getSection() != null) {
				for (POCDMT000040Entry entry : component3.getSection().getEntries()) {
					HealthProvider healthProvider = entry.getHealthProvider();
					if (healthProvider != null && healthProvider.getOrganization() != null) {
						for (II id : healthProvider.getOrganization().getIds()) {
							String idType = id.getAssigningAuthorityName();
							if (!idType.equals("Yder")) {
								log.warn("Dentist information not fetched because SKR returned a health provider " +
												 "with an ID that is not of type 'Yder'. Type was '" + idType +
												 "' with ID '" + id.getExtension() + "'.");
							} else {
								String yderNr = id.getExtension();
								if (yderNr != null && !yderNr.isEmpty()) {
									String paddedYderNr = YderNrUtil.padZero(yderNr);
									yderNrSet.add(paddedYderNr);
									id.setExtension(paddedYderNr);
								}
							}
						}
					}
				}
			}
		}
		return yderNrSet;
	}

	private void fetchLTRData(String cprnr, RegisterData data, RequestContextVO requestContext) throws DGWSClientException {
		log.debug("Feching living will data for " + cprnr);
		HasLivingWillResponse response = livingWillClient.call(cprnr, requestContext);
		if (response != null) {
			data.setHasLivingWill(response.isWillExists());
		}
	}

	private void fetchBTRData(String cprnr, RegisterData data, RequestContextVO requestContext) throws DGWSClientException {
		log.debug("Feching treatment will data for " + cprnr);
		HasTreatmentWillResponse response = treatmentWillClient.call(cprnr, requestContext);
		if (response != null) {
			data.setHasTreatmentWill(response.isWillExists());
		}
	}

	private void fetchODRData(String cprnr, RegisterData data, RequestContextVO requestContext) throws DGWSClientException {
		log.debug("Feching organ donor data for " + cprnr);
		HasOrganDonorRegistrationResponse response = organDonorClient.call(cprnr, requestContext);
		if (response != null) {
			data.setHasOrganDonorRegistration(response.isRegistrationExists());
		}
	}

	private void fetchFGVHRData(String cprnr, RegisterData data) throws FgvhrClientException {
		log.debug("Feching FGVHR data for " + cprnr);
		boolean hasNoResuscitation = fgvhrClient.hasNoResuscitation(cprnr);
		data.setHasNoResuscitation(hasNoResuscitation);
	}

	private abstract class CallWrapper implements Callable<Void> {
		protected Exception ex;
		protected Future future;

		public abstract void callClient() throws DGWSClientException;

		@Override
		public Void call() throws Exception {
			try {
				callClient();
			} catch (Exception ex) {
				this.ex = ex;
			}
			return null;
		}

		public void checkForExceptions(FaultCode faultCode) {
			if (ex != null) {
				FaultUtil.throwFSKFault(faultCode, ex.getMessage(), ex);
			}
		}

		public void submit() {
			future = executorService.submit(this);
		}

		public void waitForTermination(FaultCode faultCode) throws FSKException {
			try {
				future.get(properties.getDgwsClientTimeoutMillis(), TimeUnit.MILLISECONDS);
			} catch (InterruptedException | ExecutionException | TimeoutException ex) {
				FaultUtil.throwFSKFault(faultCode, ex);
			}
		}
	}

	private class CustodyCallWrapper<T extends CustodyData> extends CallWrapper {

		private RequestContextVO requestContext;
		private T custodyData;

		public CustodyCallWrapper(RequestContextVO requestContext, T custodyData) {
			this.requestContext = requestContext;
			this.custodyData = custodyData;
		}

		@Override
		public void callClient() throws DGWSClientException {
			log.debug("Fetching custody SCES data for " + custodyData.getCpr());
			GetPersonWithHealthCareInformationOut result = cpre.call(custodyData.getCpr(), requestContext);

			custodyData.setGetPersonWithHealthCareInformationOut(result);
		}

		@Override
		public void checkForExceptions(FaultCode faultCode) {
			if (ex instanceof FSKException) {
				FSKException e = (FSKException) ex;
				if (e.getFaultCode() == FaultCode.NO_SCES_DATA_FOUND) {
					// When the exception means "no data", we just continue
					log.warn("No SCES data found for custody person: " + custodyData.getCpr(), ex);
					return;
				}
			}
			super.checkForExceptions(faultCode);
		}
	}
}