/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util;

import dk.nsi.hsuid.HealthcareServiceUserIdentificationHeaderUtil;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.hsuid.hsuidattributes.ValidatedHsuidAttributes;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.Organisation;
import dk.sds.nsp.security.SecurityContext.User;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.util.Optional;
import java.util.UUID;

@Component
public class RequestContextProvider {

    @Autowired
    private Properties properties;

    private static final Logger log = LogManager.getLogger(RequestContextProvider.class);

    private SecurityContext securityContext;

    public RequestContextVO getAndValidateSecuityData(HsuidHeader hsuidHeader) {

        this.securityContext = Security.getSecurityContext();

        RequestContextVO rc = new RequestContextVO();

        // Validerer gyldigheden af sikkerhedsbilletten
        if (securityContext == null || !securityContext.getTicket().isPresent() || !securityContext.getTicket().get().isValid()) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, "Sikkerhedsbilletten var ikke gyldig");
        }

        // Tjek forskellige brugertyper
        if (isSystemUser(hsuidHeader)) {
            rc.setUserType(RequestContextVO.UserType.SystemUser);
            mapAndValidateSystemUser(rc, securityContext, hsuidHeader);
        } else if (isHealthcareProfessional()) {
            rc.setUserType(RequestContextVO.UserType.HealthCareProfessional);
            mapAndValidateHealthCareProfessional(rc, securityContext, hsuidHeader);
        } else if (isCitizenUser(hsuidHeader)) {
            rc.setUserType(RequestContextVO.UserType.Citizen);
            mapAndValidateCitizenUser(rc, securityContext, hsuidHeader);
        } else {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_UNKNOWN_USER_TYPE);
        }

        enrichContext(rc);

        return rc;
    }

    private boolean isSystemUser(HsuidHeader hsuidHeader) {
        if (!securityContext.getActingUser().isPresent()) {
            if(hsuidHeader != null) {
                ValidatedHsuidAttributes validatedHsuidAttributes = HealthcareServiceUserIdentificationHeaderUtil.getValidatedHsuidAttributes(hsuidHeader);
                if (!validatedHsuidAttributes.isCitizen()) {
                    return true;
                }
            } else {
                return true;
            }
        }
        return false;
    }

    private boolean isHealthcareProfessional() {
        if (securityContext.getActingUser().isPresent() && securityContext.getActingUser().get().getUserType().equals(User.UserType.HealthcareProfessional)) {
            return true;
        }
        return false;
    }

    private boolean isCitizenUser(HsuidHeader hsuidHeader) {
        if (securityContext.getActingUser().isPresent() && securityContext.getActingUser().get().getUserType().equals(User.UserType.Citizen)) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_CITIZEN_NOT_ALLOWED);;
        }
        if(hsuidHeader != null) {
            ValidatedHsuidAttributes validatedHsuidAttributes = HealthcareServiceUserIdentificationHeaderUtil.getValidatedHsuidAttributes(hsuidHeader);
            if (validatedHsuidAttributes.isCitizen()) {
                return true;
            }
        }
        return false;
    }


    private RequestContextVO mapAndValidateSystemUser(RequestContextVO rc, SecurityContext securityContext, HsuidHeader hsuidHeader) {
        mapDataFromClient(rc, securityContext);
        mapDataFromMessage(rc, securityContext);
        mapAndValidateOrganisation(rc, securityContext, true);

        // Whitelistning check
        cvrIsWhiteListed(rc.getOrganisation().getIdentifier());

        return rc;
    }


    private RequestContextVO mapAndValidateHealthCareProfessional(RequestContextVO rc, SecurityContext securityContext, HsuidHeader hsuidHeader) {
        mapDataFromClient(rc, securityContext);
        mapDataFromMessage(rc, securityContext);
        mapAndValidateOrganisation(rc, securityContext, true);
        mapAndValidateUser(rc, securityContext);

        rc.setPatientInContext(hsuidHeader);
        rc.setUserCprIdentifier(hsuidHeader);

        // Whitelistning check not relevant for HealthcareProfessional

        return rc;
    }

    private RequestContextVO mapAndValidateCitizenUser(RequestContextVO rc, SecurityContext securityContext, HsuidHeader hsuidHeader) {
        mapDataFromClient(rc, securityContext);
        mapDataFromMessage(rc, securityContext);
        mapAndValidateOrganisation(rc, securityContext, true);

        rc.setPatientInContext(hsuidHeader);
        rc.setUserCprIdentifier(hsuidHeader);

        // Whitelistning check
        cvrIsWhiteListed(rc.getOrganisation().getIdentifier());

        return rc;
    }


    private void mapAndValidateUser(RequestContextVO rc, SecurityContext securityContext) {
        if (!securityContext.getActingUser().isPresent()) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, "Udførende bruger skal være med i sikkerhedsbilletten");
        }

        if (securityContext.getPrincipalUser().isPresent()) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_NO_DELEGATION);
        }

        User user = securityContext.getActingUser().get();
        if (user.getIdentifierFormat() == SecurityContext.User.PersonIdentifierFormat.CPR) {
            rc.setUserCprIdentifier(securityContext.getActingUser().get().getIdentifier());
        } else {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, "Bruger skal have angivet CPR nummer");
        }
        rc.setGivenName(user.getGivenName().orElse(null));
        rc.setSurName(user.getSurname().orElse(null));
        if (user.getCredentials().isPresent()) {
            SecurityContext.User.Credentials credentials = user.getCredentials().get();
            rc.setAuthorizationCode(credentials.getAuthorizationCode().orElse(null));
            rc.setEducationCode(credentials.getEducationCode().orElse(null));
        }
    }


    private void mapDataFromMessage(RequestContextVO rc, SecurityContext securityContext) {
        if (securityContext.getMessage().isPresent()) {
            rc.setMessageId(securityContext.getMessage().get().getIdentifier().orElse(UUID.randomUUID().toString()));
            rc.setConversationid(securityContext.getMessage().get().getConversationIdentifier().orElse(UUID.randomUUID().toString()));
        }
    }

    private void mapDataFromClient(RequestContextVO rc, SecurityContext securityContext) {
        if (securityContext.getClient().isPresent()) {
            rc.setItSystemName(securityContext.getClient().get().getName().orElse(null));
        }
    }

    private void mapAndValidateOrganisation(RequestContextVO rc, SecurityContext securityContext, boolean requireCvr) {
        if (requireCvr) {
            if (!securityContext.getOrganisation().isPresent() || securityContext.getOrganisation().get().getIdentifierFormat() == null) {
                FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, "Organisation og organisationsidentifikator skal være tilstede i sikkehedsbilletten");
            } else if (!Organisation.OrganisationIdentifierFormat.CVR.equals(securityContext.getOrganisation().get().getIdentifierFormat())) {
                FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, String.format("Ukendt organisationsidentifikator i sikkerhedsbilletten: %s", securityContext.getOrganisation().get().getIdentifierFormat().toString()));
            }
        }
        rc.setOrganisation(securityContext.getOrganisation().get());
    }

    private RequestContextVO.UserType getUserType(HsuidHeader hsuidHeader) {

        if(hsuidHeader != null) {
            ValidatedHsuidAttributes validatedHsuidAttributes = HealthcareServiceUserIdentificationHeaderUtil.getValidatedHsuidAttributes(hsuidHeader);
            if (validatedHsuidAttributes.isCitizen()) {
                return RequestContextVO.UserType.Citizen;
            }
        }
        return RequestContextVO.UserType.SystemUser;
    }

    public boolean cvrIsWhiteListed(String cvr) {
        if (StringUtils.isEmpty(cvr)) {
            log.warn("CVR number is not provided for system user: " + cvr);
            return false;
        }
        String whitelistedLevel3CVRs = properties.getWhitelistedLevel3CVRs();
        boolean contains = whitelistedLevel3CVRs.contains(cvr);
        if (!contains) {
            log.warn("CVR number is not allowed for system user: " + cvr);
        }
        return contains;
    }

    private void enrichContext(RequestContextVO rc) {
        RequestContextEnricher requestContextEnricher = new RequestContextEnricher() {


            @Override
            public Optional<String> createRequestWithContext(Element request) {

                HsuidHeader hsuid = rc.getHsuidHeader();

                Document hsuidDoc = null;
                if (hsuid != null) {
                    try {
                        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                        hsuidDoc = dbf.newDocumentBuilder().newDocument();
                        rc.getJaxbContext().createMarshaller().marshal(hsuid, hsuidDoc);
                    } catch (ParserConfigurationException | JAXBException e) {
                        throw new RuntimeException(e);
                    }
                }
                return hsuidDoc != null ? createRequestWithContextSynchronized(request, hsuidDoc.getDocumentElement()) : createRequestWithContextSynchronized(request);
            }

            private synchronized Optional<String> createRequestWithContextSynchronized(Element requestBody, Element ... headers) {
                return securityContext.getTicket().get().createRequestWithContext(requestBody, headers);
            }
        };

        rc.setRequestContextEnricher(requestContextEnricher);
    }
}
