/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda;

import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.service.PersonalDataCardService;
import dk.stamkort.util.*;
import dk.stamkort.webservice.hl7cda.assembler.PDC_2_0ClinicalDocumentAssembler;
import dk.stamkort.webservice.hl7cda.assembler.PDC_3_0ClinicalDocumentAssembler;
import oio.medcom.cprservice._1_0.PersonCivilRegistrationStatusStructureType;
import oio.medcom.cprservice._1_0.PersonHealthCareInformationStructureType;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.stereotype.Service;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.namespace.QName;
import java.io.StringWriter;
import java.math.BigInteger;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Optional;

import static java.util.Optional.ofNullable;

@Service
public class ClinicalDocumentProviderImpl implements ClinicalDocumentProvider {

    private final PersonalDataCardService personalDataCardService;
    private final Properties properties;

    private JAXBContext jaxbContext;

    public ClinicalDocumentProviderImpl(PersonalDataCardService personalDataCardService, Properties properties) throws JAXBException {
        this.personalDataCardService = personalDataCardService;
        this.properties = properties;

        jaxbContext = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
    }

    @Override
    public String getClinicalDocument(String cprnr, RequestContextVO requestContext, String formatCode) throws JAXBException {
        RegisterData registerData = personalDataCardService.getRegisterData(cprnr, requestContext);
        checkStatus(getStatus(registerData));

        PersonHealthCareInformationStructureType healthCareInformation = registerData.getScesData()
                .getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure();

        String yderNr = getYderNr(healthCareInformation);

        POCDMT000040ClinicalDocument doc =
                createPdcClinicalDocument(registerData, registerData.getYder(yderNr), ofNullable(formatCode));
        JAXBElement<POCDMT000040ClinicalDocument> documentJAXBElement =
                new JAXBElement<>(new QName("urn:hl7-org:v3", "ClinicalDocument"),
                                  POCDMT000040ClinicalDocument.class,
                                  doc);
        StringWriter sw = new StringWriter();
        jaxbContext.createMarshaller().marshal(documentJAXBElement, sw);
        return enhanceXml(sw.toString());
    }

    private POCDMT000040ClinicalDocument createPdcClinicalDocument(RegisterData registerData,
                                                                   YderVO yder,
                                                                   Optional<String> formatCode) {
        AuthorProperties authorProperties = properties.createAuthorProperties();
        return formatCode.filter(c -> c.equals("urn:ad:dk:medcom:pdc-v3.0:full")).isPresent()
                ? PDC_3_0ClinicalDocumentAssembler.assembleClinicalDocument(
                        PdcDataMapperV3.mapData(registerData), authorProperties)
                : PDC_2_0ClinicalDocumentAssembler.assembleClinicalDocument(
                        PdcDataMapperV2.mapDataHolder(registerData, yder), authorProperties);
    }

    private String enhanceXml(String toString) {
        return new XmlEnhancer().enhance(toString);
    }

    private void checkStatus(PersonCivilRegistrationStatusStructureType statusStructure) {
        if (statusStructure == null) {
            return;
        }
        String status = statusStructure.getPersonCivilRegistrationStatusCode();
        switch (status) {
            case "30":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_30, properties.getCallScesStatus30());
                break;
            case "50":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_50, properties.getCallScesStatus50());
                break;
            case "60":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_60, properties.getCallScesStatus60());
                break;
            case "90":
                XMLGregorianCalendar statusDate = statusStructure.getPersonCivilRegistrationStatusStartDate();
                LocalDateTime statusDateTime = statusDate.toGregorianCalendar().toZonedDateTime().toLocalDateTime();
                LocalDateTime oneYear = LocalDateTime.now().minus(1, ChronoUnit.YEARS);
                if (statusDateTime.isBefore(oneYear)) {
                    throwFaultStatus(FaultCode.CALL_SCES_STATUS_90, properties.getCallScesStatus90());
                }

                break;
        }

    }

    private PersonCivilRegistrationStatusStructureType getStatus(RegisterData registerData) {
        return registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().getPersonCivilRegistrationStatusStructure();
    }

    private void throwFaultStatus(FaultCode faultCode, String faultText) {
        if (org.apache.commons.lang.StringUtils.isEmpty(faultText)) {
            FaultUtil.throwFSKFault(faultCode);
        } else {
            FaultUtil.throwFSKFault(faultCode, faultText);
        }
    }

    private String getYderNr(PersonHealthCareInformationStructureType healthCareInformation) {
        String yderNr = "";
        if (healthCareInformation != null) {
            AssociatedGeneralPractitionerStructureType associatedGeneralPractitioner = healthCareInformation.getCurrentGeneralPractitionerStructure().getAssociatedGeneralPractitionerStructure();
            if (associatedGeneralPractitioner != null) {
                BigInteger ydernrBI = associatedGeneralPractitioner.getAssociatedGeneralPractitionerIdentifier();
                if (ydernrBI != null && !ydernrBI.equals(new BigInteger("0"))) {
                    yderNr = ydernrBI.toString();
                }
            }
        }

        return yderNr;
    }
}
