/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda;

import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.CountryIdentificationCodeType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.stamkort.OID;
import dk.stamkort.domain.Dentist;
import dk.stamkort.domain.GeneralPractitioner;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.webservice.hl7cda.model.*;
import dk.stamkort.webservice.hl7cda.model.v3.*;
import dk.sundhedsdatastyrelsen.fsk.extensions.*;
import oio.medcom.cprservice._1_0.*;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.hl7.v3.*;

import javax.xml.bind.JAXBElement;
import javax.xml.datatype.XMLGregorianCalendar;
import java.io.Serializable;
import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

import static dk.stamkort.webservice.hl7cda.LanguageExt.toDanishLanguageName;
import static dk.stamkort.webservice.hl7cda.model.DataEnterer.*;
import static java.util.Collections.emptyList;
import static java.util.Optional.*;

public class PdcDataMapperV3 {

    private PdcDataMapperV3() { }

    private static final DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ");
    private static final DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd");

    public static PdcV3Data mapData(RegisterData registerData) {
        PrimaryPersonInformationHolder personHolder = mapPersonHolder(registerData);
        Optional<ProviderOrganizationHolder> providerOrganizationHolder =
                registerData.getGeneralPractitioner().map(PdcDataMapperV3::mapProviderOrganizationHolder);
        List<CustodyHolder> custodyHolder = mapCustodyHolder(registerData);
        CoverageGroupHolder coverageGroupHolder = mapCoverageGroupHolder(registerData);
        OrganDonorRegistrationHolder organDonorRegistrationHolder = mapOrganDonorRegistrationHolder(registerData);
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = mapTreatmentWillRegistrationHolder(registerData);
        LivingWillRegistrationHolder livingWillRegistrationHolder = mapLivingWillRegistrationHolder(registerData);
        Optional<LanguageTypedInV3> languageTypedInV3 = mapLanguageTypedInV3(registerData);
        Optional<TempAddrTypedInV3> tempAddrTypedInV3 = mapTempAddrTypedInV3(registerData);
        Optional<DentistTypedInV3> dentistTypedInV3 = mapDentistTypedInV3(registerData);
        Optional<PatientContactTypedInV3> patientContactTypedInV3 = mapPatientContactTypedInV3(registerData);
        List<RelativeTypedInV3> relativeTypedInV3 = mapRelativeTypedInV3(registerData);

        return new PdcV3Data(registerData.getHasNoResuscitation())
                .setPrimaryPersonInformation(personHolder)
                .setProviderOrganizationHolder(providerOrganizationHolder)
                .setCustodyHolderList(custodyHolder)
                .setCoverageGroupHolder(coverageGroupHolder)
                .setOrganDonorRegistrationHolder(organDonorRegistrationHolder)
                .setTreatmentWillRegistrationHolder(treatmentWillRegistrationHolder)
                .setLivingWillRegistrationHolder(livingWillRegistrationHolder)
                .setLanguageTypedInV3(languageTypedInV3)
                .setTempAddrTypedInV3(tempAddrTypedInV3)
                .setDentistTypedInV3(dentistTypedInV3)
                .setPatientContactTypedInV3(patientContactTypedInV3)
                .setRelativeTypedInV3(relativeTypedInV3);
    }

    private static List<CustodyHolder> mapCustodyHolder(RegisterData registerData) {
        List<CustodyHolder> custodyHolderList = new ArrayList<>();

        // ChildCustody data
        // This custody code should maybe get mapped in the CDA assembler classes instead
        CodeHolder childCustodyCustodyCode = new CodeHolder();
        childCustodyCustodyCode.setCode("ChildCustody");
        childCustodyCustodyCode.setDisplayName("Forældremyndighed over");

        registerData.getChildCustodyCustodyData()
                .stream()
                .map(custodyData ->
                        mapCustodyHolderElement(
                                custodyData.getGetPersonWithHealthCareInformationOut(),
                                childCustodyCustodyCode,
                                mapChildCustodyCustodyRelationCode(
                                        registerData.getCprnr(),
                                        custodyData.getGetPersonWithHealthCareInformationOut())))
                .forEach(custodyHolderList::add);

        // CustodyBy data
        // This custody code should maybe get mapped in the CDA assembler classes instead
        CodeHolder custodyByCustodyCode = new CodeHolder();
        custodyByCustodyCode.setCode("CustodyBy");
        custodyByCustodyCode.setDisplayName("Forældremyndighedshaver");

        registerData.getCustodyByCustodyData()
                .stream()
                .map(custodyData ->
                        mapCustodyHolderElement(
                                custodyData.getGetPersonWithHealthCareInformationOut(),
                                custodyByCustodyCode,
                                mapCustodyByCustodyRelationCode(
                                        custodyData.getPersonCustodyRelationType())))
                .forEach(custodyHolderList::add);

        return custodyHolderList;
    }

    private static CustodyHolder mapCustodyHolderElement(GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformationOut, CodeHolder custodyCode, CodeHolder custodyRelationCode) {
        CustodyHolder custodyHolder = new CustodyHolder();

        // code
        custodyHolder.setCode(custodyCode);

        RegularCPRPersonType regularCPRPerson = getPersonWithHealthCareInformationOut
                .getPersonWithHealthCareInformationStructure() 
                .getPersonInformationStructure() 
                .getRegularCPRPerson(); 

        SimpleCPRPersonType simpleCPRPerson = regularCPRPerson
                .getSimpleCPRPerson(); 

        // cpr
        custodyHolder.setCpr(simpleCPRPerson.getPersonCivilRegistrationIdentifier());

        PersonNameStructureType personNameStructure = simpleCPRPerson
                .getPersonNameStructure(); 

        if (personNameStructure != null) {
            // givens
            // Split given names by whitespace
            String givenNames = Optional.ofNullable(personNameStructure.getPersonGivenName())
                    .filter(s -> !s.isEmpty())
                    .orElse(null);
            Arrays.stream(ArrayUtils.nullToEmpty(StringUtils.split(givenNames)))
                    .forEach(custodyHolder::addGiven);

            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonMiddleName()))) 
                    .forEach(custodyHolder::addGiven); // add middle name to given names

            // family
            String familyName = Optional.ofNullable(personNameStructure.getPersonSurnameName())
                    .filter(s -> !s.isEmpty())
                    .orElse(null);
            custodyHolder.setFamily(familyName);
        }

        // relation
        custodyHolder.setRelation(custodyRelationCode);

        return custodyHolder;
    }

    /**
     * @param cprOfParent                                  cpr of the parent
     * @param getPersonWithHealthCareInformationOutOfChild GetPersonWithHealthCareInformationOut object of the Child
     * @return code
     */
    private static CodeHolder mapChildCustodyCustodyRelationCode(String cprOfParent, GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformationOutOfChild) {

        List<PersonCustodyStructureType> personCustodyStructure = getPersonWithHealthCareInformationOutOfChild
                .getPersonWithHealthCareInformationStructure()
                .getPersonInformationStructure()
                .getPersonCustodyStructure();

        // Loop Person Custodies of the child to find the relation to the parent for the current parent
        PersonCustodyRelationType personCustodyRelationType = personCustodyStructure
                .stream()
                .filter(personCustodyStructureType ->
                        personCustodyStructureType.getPersonCustodyIdentifier().equals(cprOfParent))
                .reduce((personCustodyStructure1, personCustodyStructure2) -> {
                    throw new IllegalStateException("Multiple children with same CPR found");
                })
                .map(PersonCustodyStructureType::getPersonCustodyRelation)
                .orElseThrow(() -> new IllegalStateException("Expected Person Custody not found for child"));

        String code;
        String displayName;

        switch (personCustodyRelationType) {
            case MOR:
                code = "mor";
                displayName = "Mor";
                break;
            case FAR:
                code = "far";
                displayName = "Far";
                break;
            case ANDEN:
                code = "anden";
                displayName = "Anden relation";
                break;
            default:
                throw new IllegalStateException("Unexpected value: " + personCustodyRelationType);
        }

        CodeHolder codeHolder = new CodeHolder();

        // code
        codeHolder.setCode(code);

        // displayName
        codeHolder.setDisplayName(displayName);

        return codeHolder;
    }

    @SuppressWarnings("unused") // unused for now, but maybe relevant in future
    private static CodeHolder mapCustodyByCustodyRelationCode(PersonCustodyRelationType personCustodyRelationType) {
        CodeHolder codeHolder = new CodeHolder();

        // According to "Teknisk Implementeringsvejledning" (https://www.nspop.dk/x/YgsaBg)
        // Always use "barn" and "Barn" as values for CustodyBy

        // code
        codeHolder.setCode("barn");

        // displayName
        codeHolder.setDisplayName("Barn");

        return codeHolder;
    }

    private static CoverageGroupHolder mapCoverageGroupHolder(RegisterData registerData) {
        PersonPublicHealthInsuranceType personPublicHealthInsurance = registerData.getScesData()
                .getPersonWithHealthCareInformationStructure() 
                .getPersonHealthCareInformationStructure() 
                .getPersonPublicHealthInsurance(); 

        CoverageGroupHolder coverageGroupHolder = new CoverageGroupHolder();
        String group = ofNullable(personPublicHealthInsurance.getPublicHealthInsuranceGroupIdentifier())
                .map(PublicHealthInsuranceGroupIdentifierType::value)
                .map(str -> str.replace("sygesikringsgruppe_", "") )
                .orElse(null);

        // group
        coverageGroupHolder.setGroup(group);

        return coverageGroupHolder;
    }

    private static List<RelativeTypedInV3> mapRelativeTypedInV3(RegisterData registerData) {
        return getSKREntryElementsOfType(RelatedPerson.class, registerData.getClinicalDocument()).stream()
                .map(PdcDataMapperV3::mapRelativeTypedInV3Element)
                .collect(Collectors.toList());
    }

    private static RelativeTypedInV3 mapRelativeTypedInV3Element(RelatedPerson relatedPerson) {
        DataEnterer dataEnterer = toDataEnterer(relatedPerson.getDataEnterer());
        RelativeTypedInV3 relativeTypedInV3 =
                new RelativeTypedInV3(new IdHolder(relatedPerson.getId().getRoot(),
                                                       relatedPerson.getId().getExtension(),
                                                       relatedPerson.getId().getAssigningAuthorityName()),
                                      new AssignedAuthor(dataEnterer,
                                                         toRepresentedOrg(relatedPerson.getDataEnterer())));

        POCDMT000040AssociatedEntity associatedEntity = relatedPerson.getAssociatedEntity();

        POCDMT000040Person associatedPerson = associatedEntity.getAssociatedPerson();

        ofNullable(associatedPerson).ifPresent(value -> {
            relativeTypedInV3.setFamily(PdcDataMapperCommon.getPersonFamily(value));
            PdcDataMapperCommon.getPersonGivens(value).forEach(relativeTypedInV3::addGiven);
        });

        CD relationshipType = relatedPerson.getRelationshipType();
        CodeHolder relation = new CodeHolder();
        relation.setCode(relationshipType.getCode());
        relation.setDisplayName(relationshipType.getDisplayName());

        // relationCode
        relativeTypedInV3.setRelation(relation);

        String note = relatedPerson.getNote();

        associatedEntity.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("tel:"))
                .forEach(tel ->
                        relativeTypedInV3.getPhoneHolders().add((mapPhoneHolder(tel))));

        relativeTypedInV3.setNote(note);

        POCDMT000040Author skrDataEnterer = relatedPerson.getDataEnterer();
        POCDMT000040AssignedAuthor assignedAuthor = skrDataEnterer.getAssignedAuthor();
        relativeTypedInV3.setAuthorTime(getDataEntererTime(skrDataEnterer));
        relativeTypedInV3.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor.getAssignedPerson();
        PdcDataMapperCommon.getPersonGivens(assignedPerson).forEach(relativeTypedInV3::addAuthorGiven);
        relativeTypedInV3.setAuthorFamily(PdcDataMapperCommon.getPersonFamily(assignedPerson));
        return relativeTypedInV3;
    }

    private static Optional<PatientContactTypedInV3> mapPatientContactTypedInV3(RegisterData registerData) {
        List<PatientContact> contact =
                getSKREntryElementsOfType(PatientContact.class, registerData.getClinicalDocument());
        if(contact.size() > 1) { FaultUtil.throwFSKFault(FaultCode.MAPPING_ERROR_INVALID_DATA); }
        return contact.stream().findFirst().map(PdcDataMapperV3::mapPatientContactTypedInV3Element);
    }

    public static PatientContactTypedInV3 mapPatientContactTypedInV3Element(PatientContact patientContact) {
        DataEnterer dataEnterer = toDataEnterer(patientContact.getDataEnterer());
        PatientContactTypedInV3 patientContactTypedInV3 =
                new PatientContactTypedInV3(new AssignedAuthor(dataEnterer,
                                                               toRepresentedOrg(patientContact.getDataEnterer())));

        patientContact.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("tel:"))
                .forEach(tel ->
                        patientContactTypedInV3.addPhone(mapPhoneHolder(tel)));

        patientContact.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("mailto:"))
                .forEach(tel ->
                        patientContactTypedInV3.addEmailAddress(mapPhoneHolder(tel)));

        POCDMT000040Author skrDataEnterer = patientContact.getDataEnterer();
        patientContactTypedInV3.setAuthorTime(getDataEntererTime(skrDataEnterer));
        POCDMT000040AssignedAuthor assignedAuthor = skrDataEnterer.getAssignedAuthor();
        patientContactTypedInV3.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));
        POCDMT000040Person assignedPerson = assignedAuthor.getAssignedPerson();
        PdcDataMapperCommon.getPersonGivens(assignedPerson).forEach(patientContactTypedInV3::addAuthorGiven);
        patientContactTypedInV3.setAuthorFamily(PdcDataMapperCommon.getPersonFamily(assignedPerson));

        return patientContactTypedInV3;
    }

    private static Optional<DentistTypedInV3> mapDentistTypedInV3(RegisterData registerData) {
        List<HealthProvider> healthProviders =
                getSKREntryElementsOfType(HealthProvider.class, registerData.getClinicalDocument());
        if (healthProviders.size() > 1) {
            FaultUtil.throwFSKFault(FaultCode.MAPPING_ERROR_INVALID_DATA);
        }

        // Se mapningstabel i SDS-7203
        Optional<Dentist> dentist = registerData.getDentist();
        Optional<String> personName = dentist.map(Dentist::getPersonName);
        Optional<String> orgName = dentist.map(Dentist::getOrganizationName);
        Optional<String> dentistAddress = dentist.map(Dentist::getAddress);
        Optional<String> dentistPostcode = dentist.map(Dentist::getPostcode);
        Optional<String> dentistCity = dentist.map(Dentist::getCity);
        final String dentistCountry = "DK";
        Optional<String> phoneNumber = dentist.map(Dentist::getPhone);

        return healthProviders.stream().findFirst().map(
                healthProvider -> mapDentistTypedInV3Element(healthProvider,
                        personName,
                        orgName,
                        phoneNumber,
                        dentistAddress,
                        dentistPostcode,
                        dentistCity,
                        of(dentistCountry)
                )
        );
    }

    private static DentistTypedInV3 mapDentistTypedInV3Element(HealthProvider healthProvider,
                                                               Optional<String> givenName,
                                                               Optional<String> orgName,
                                                               Optional<String> phoneNumber,
                                                               Optional<String> dentistAddress,
                                                               Optional<String> dentistPostcode,
                                                               Optional<String> dentistCity,
                                                               Optional<String> dentistCountry

    ) {
        // Hvem der har indtastet data
        DentistTypedInV3 dentistTypedInV3 =
                new DentistTypedInV3(new AssignedAuthor(toDataEnterer(healthProvider.getDataEnterer()),
                                                        toRepresentedOrg(healthProvider.getDataEnterer())));

        // Tandlægens adresse
        dentistTypedInV3.setAddress(
                mapAddressHolderSimple(
                        "WP",
                        dentistAddress.map(Collections::singletonList).orElse(emptyList()),
                        dentistCity.orElse(null),
                        dentistPostcode.orElse(null),
                        dentistCountry.orElse(null)
                )
        );

        // Telefonnummer
        phoneNumber.ifPresent(tel -> dentistTypedInV3.setPhone(mapPhoneHolder(tel)));

        // Tidspunkt for indtastning
        POCDMT000040Author skrDataEnterer = healthProvider.getDataEnterer();
        POCDMT000040AssignedAuthor assignedAuthor = skrDataEnterer.getAssignedAuthor();
        dentistTypedInV3.setAuthorTime(getDataEntererTime(skrDataEnterer));

        // Navn på person der har indtastet data
        POCDMT000040Person assignedPerson = assignedAuthor.getAssignedPerson();
        PdcDataMapperCommon.getPersonGivens(assignedPerson).forEach(dentistTypedInV3::addAuthorGiven);

        healthProvider.getOrganization()
                .getIds()
                .stream()
                .filter(x -> OID.YDERNUMMER.equals(x.getRoot()))
                .findFirst()
                .ifPresent(x -> dentistTypedInV3.setYder(x.getExtension()));

        dentistTypedInV3.setAuthorFamily(PdcDataMapperCommon.getPersonFamily(assignedPerson));
        givenName.ifPresent(dentistTypedInV3::setName);
        orgName.ifPresent(dentistTypedInV3::setOrganizationName);
        dentistTypedInV3.setIdExtension(healthProvider.getId().getExtension());
        return dentistTypedInV3;
    }

    private static Optional<TempAddrTypedInV3> mapTempAddrTypedInV3(RegisterData registerData) {
        List<TemporaryAddress> tempAddr =
                getSKREntryElementsOfType(TemporaryAddress.class, registerData.getClinicalDocument());
        if(tempAddr.size() > 1) { FaultUtil.throwFSKFault(FaultCode.MAPPING_ERROR_INVALID_DATA); }
        return tempAddr.stream().findFirst().map(PdcDataMapperV3::mapTempAddrTypedInV3Element);
    }

    private static TempAddrTypedInV3 mapTempAddrTypedInV3Element(TemporaryAddress temporaryAddress) {
        DataEnterer dataEnterer = toDataEnterer(temporaryAddress.getDataEnterer());
        TempAddrTypedInV3 tempAddrTypedInV3 =
                new TempAddrTypedInV3(new IdHolder(temporaryAddress.getId().getRoot(),
                                                       temporaryAddress.getId().getExtension(),
                                                       temporaryAddress.getId().getAssigningAuthorityName()),
                                      new AssignedAuthor(dataEnterer,
                                                         toRepresentedOrg(temporaryAddress.getDataEnterer())));
        AD addr = temporaryAddress.getAddr();

        List<SXCMTS> usablePeriodElements = getJAXBElementsFromName(SXCMTS.class, "urn:hl7-org:v3", "useablePeriod", addr.getContent());

        if (usablePeriodElements.size() >= 1) {
            // timeFrom
            ofNullable(usablePeriodElements.get(0).getValue())
                    .ifPresent(value ->
                            tempAddrTypedInV3.setPeriodFrom(
                                    LocalDate.parse(value, dateFormatter))); // expects date format (not date time)
        }

        if (usablePeriodElements.size() == 2) {
            // timeFrom
            ofNullable(usablePeriodElements.get(1).getValue())
                    .ifPresent(value ->
                            tempAddrTypedInV3.setPeriodTo(
                                    LocalDate.parse(value, dateFormatter))); // expects date format (not date time)
        }

        List<String> uses = addr.getUses();
        String use = uses.isEmpty() ? null : uses.get(0);
        List<AdxpStreetAddressLine> streetAddressLines = getJAXBElementsFromName(AdxpStreetAddressLine.class, "urn:hl7-org:v3", "streetAddressLine", addr.getContent());
        AdxpPostalCode postalCode = getSingleJAXBElementFromName(AdxpPostalCode.class, "urn:hl7-org:v3", "postalCode", addr.getContent());
        AdxpCity city = getSingleJAXBElementFromName(AdxpCity.class, "urn:hl7-org:v3", "city", addr.getContent());
        AdxpCountry country = getSingleJAXBElementFromName(AdxpCountry.class, "urn:hl7-org:v3", "country", addr.getContent());

        // address
        tempAddrTypedInV3.setAddress(
                mapAddressHolderSimple(
                        use,
                        streetAddressLines.stream()
                                .map(PdcDataMapperV3::binToString)
                                .collect(Collectors.toList()),
                        binToString(city),
                        binToString(postalCode),
                        binToString(country)));

        POCDMT000040Author skrDataEnterer = temporaryAddress.getDataEnterer();

        POCDMT000040AssignedAuthor assignedAuthor = skrDataEnterer.getAssignedAuthor();

        tempAddrTypedInV3.setAuthorTime(getDataEntererTime(skrDataEnterer));

        tempAddrTypedInV3.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor.getAssignedPerson();

        PdcDataMapperCommon.getPersonGivens(assignedPerson).forEach(tempAddrTypedInV3::addAuthorGiven);

        tempAddrTypedInV3.setAuthorFamily(PdcDataMapperCommon.getPersonFamily(assignedPerson));

        return tempAddrTypedInV3;
    }

    private static String binToString(BIN bin) {
        return ofNullable(bin)
                .flatMap(value ->
                        value.getContent()
                                .stream()
                                .reduce((serializable, serializable2) -> {
                                    throw new IllegalStateException("Multiple elements for string content");
                                })
                                .map(Objects::toString)
                )
                .orElse(null);
    }

    private static Optional<LanguageTypedInV3> mapLanguageTypedInV3(RegisterData registerData) {
        List<Language> languageTypedInV3List =
                getSKREntryElementsOfType(Language.class, registerData.getClinicalDocument());
        if(languageTypedInV3List.size() > 1) { FaultUtil.throwFSKFault(FaultCode.MAPPING_ERROR_INVALID_DATA); }
        return languageTypedInV3List.stream().findFirst().map(PdcDataMapperV3::mapLanguageTypedInV3Element);
    }

    private static LanguageTypedInV3 mapLanguageTypedInV3Element(Language language) {
        DataEnterer dataEnterer = toDataEnterer(language.getDataEnterer());
        LanguageTypedInV3 languageTypedInV3 =
                new LanguageTypedInV3(new IdHolder(language.getId().getRoot(),
                                                   language.getId().getExtension(),
                                                   language.getId().getAssigningAuthorityName()),
                                      new AssignedAuthor(dataEnterer,
                                                         toRepresentedOrg(language.getDataEnterer())));
        String languageCode = language.getLanguageCode().getValue();
        languageTypedInV3.setLanguageCode(languageCode);
        languageTypedInV3.setLanguageName(toDanishLanguageName(new Locale(languageCode)));
        POCDMT000040Author skrDataEnterer = language.getDataEnterer();
        POCDMT000040AssignedAuthor assignedAuthor = skrDataEnterer.getAssignedAuthor();
        languageTypedInV3.setAuthorTime(getDataEntererTime(skrDataEnterer));
        languageTypedInV3.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));
        POCDMT000040Person assignedPerson = assignedAuthor.getAssignedPerson();
        PdcDataMapperCommon.getPersonGivens(assignedPerson).forEach(languageTypedInV3::addAuthorGiven);
        languageTypedInV3.setAuthorFamily(PdcDataMapperCommon.getPersonFamily(assignedPerson));
        return languageTypedInV3;
    }

    private static <T> List<T> getJAXBElementsFromName(Class<T> clazz, String ns, String localPart, List<Serializable> elements) {
        return elements.stream()
                .filter(o -> o instanceof JAXBElement)
                .map(JAXBElement.class::cast)
                .filter(jaxbElement ->
                        jaxbElement.getName().getNamespaceURI().equals(ns)
                                && jaxbElement.getName().getLocalPart().equals(localPart))
                .map(JAXBElement::getValue)
                .filter(clazz::isInstance)
                .map(clazz::cast)
                .collect(Collectors.toList());
    }

    private static <T> T getSingleJAXBElementFromName(Class<T> clazz, String ns, String localPart, List<Serializable> elements) {
        return elements.stream()
                .filter(o -> o instanceof JAXBElement)
                .map(JAXBElement.class::cast)
                .filter(jaxbElement ->
                        jaxbElement.getName().getNamespaceURI().equals(ns)
                                && jaxbElement.getName().getLocalPart().equals(localPart))
                .reduce((jaxbElement, jaxbElement2) -> {
                    throw new IllegalStateException("Multiple element found with the name " + localPart);
                })
                .map(JAXBElement::getValue)
                .filter(clazz::isInstance)
                .map(clazz::cast)
                .orElse(null);
    }

    private static PhoneHolder mapPhoneHolder(TEL tel) {
        PhoneHolder phoneHolder = new PhoneHolder();
        tel.getUses().forEach(phoneHolder::addUse);
        phoneHolder.setValue(tel.getValue());
        return phoneHolder;
    }

    public static String getAssignedAuthorCpr(POCDMT000040AssignedAuthor assignedAuthor) {
        return getIdElementByRoot("1.2.208.176.1.2", assignedAuthor)
                .map(II::getExtension)
                .orElse(null);
    }

    public static ZonedDateTime getDataEntererTime(POCDMT000040Author dataEnterer) {
        String timeValue = dataEnterer
                .getTime()
                .getValue();
        return ofNullable(timeValue)
                .map(value ->
                        ZonedDateTime.parse(
                                value, dateTimeFormatter))
                .orElse(null);
    }

    public static Optional<II> getIdElementByRoot(String root, POCDMT000040AssignedAuthor assignedAuthor) {
        return assignedAuthor.getIds() // schema requires 1 or more
                .stream()
                .filter(ii -> ii.getRoot().equals(root))
                .reduce((ignored, ignored2) -> {
                    throw new IllegalStateException("Multiple id elements with same root");
                });
    }

    private static <T> List<T> getSKREntryElementsOfType(Class<T> clazz,
                                                         POCDMT000040ClinicalDocument clinicalDocument) {
        return ofNullable(clinicalDocument)
                .map(POCDMT000040ClinicalDocument::getComponent)
                .map(POCDMT000040Component2::getStructuredBody)
                .map(POCDMT000040StructuredBody::getComponents)
                .flatMap(components -> ofNullable(components.get(0)))
                .<List<T>>map(component3 -> {
                    List<T> returnEntries = new ArrayList<>();
                    List<POCDMT000040Entry> entries = component3.getSection().getEntries();

                    for (POCDMT000040Entry entry : entries) {
                        if (clazz.equals(Language.class)) {
                            if (entry.getLanguage() != null) {
                                returnEntries.add((T) entry.getLanguage());
                            }
                        } else if (clazz.equals(TemporaryAddress.class)) {
                            if (entry.getTemporaryAddress() != null) {
                                returnEntries.add((T) entry.getTemporaryAddress());
                            }
                        } else if (clazz.equals(HealthProvider.class)) {
                            if (entry.getHealthProvider() != null) {
                                returnEntries.add((T) entry.getHealthProvider());
                            }
                        } else if (clazz.equals(PatientContact.class)) {
                            if (entry.getPatientContact() != null) {
                                returnEntries.add((T) entry.getPatientContact());
                            }
                        } else if (clazz.equals(RelatedPerson.class)) {
                            if (entry.getRelatedPerson() != null) {
                                returnEntries.add((T) entry.getRelatedPerson());
                            }
                        } else {
                            return emptyList();
                        }
                    }
                    return returnEntries;
                }).orElse(emptyList());
    }

    private static LivingWillRegistrationHolder mapLivingWillRegistrationHolder(RegisterData registerData) {
        LivingWillRegistrationHolder livingWillRegistrationHolder = new LivingWillRegistrationHolder();

        // hasRegistration
        livingWillRegistrationHolder.setHasRegistration(registerData.getHasLivingWill());

        return livingWillRegistrationHolder;
    }

    private static TreatmentWillRegistrationHolder mapTreatmentWillRegistrationHolder(RegisterData registerData) {
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = new TreatmentWillRegistrationHolder();

        // hasRegistration
        treatmentWillRegistrationHolder.setHasRegistration(registerData.getHasTreatmentWill());

        return treatmentWillRegistrationHolder;
    }

    private static OrganDonorRegistrationHolder mapOrganDonorRegistrationHolder(RegisterData registerData) {
        OrganDonorRegistrationHolder organDonorRegistrationHolder = new OrganDonorRegistrationHolder();

        // hasRegistration
        organDonorRegistrationHolder.setHasRegistration(registerData.getHasOrganDonorRegistration());

        return organDonorRegistrationHolder;
    }

    private static ProviderOrganizationHolder mapProviderOrganizationHolder(GeneralPractitioner gp) {
        ProviderOrganizationHolder providerOrganizationHolder = new ProviderOrganizationHolder();
        providerOrganizationHolder.setYderNr(gp.getYdernr());
        providerOrganizationHolder.setName(gp.getOrganizationName());
        providerOrganizationHolder.addPhone(mapPhoneHolder(gp.getPhone()));
        providerOrganizationHolder.addEmail(gp.getEmail());
        providerOrganizationHolder.setAddress(mapAddressHolderSimple("H", Collections.singletonList(gp.getAddress()), gp.getCity(), gp.getPostcode(), null));
        return providerOrganizationHolder;
    }

    @SuppressWarnings("SameParameterValue")
    private static AddressHolder mapAddressHolderSimple(String use, List<String> addressLines, String city, String postcode, String country) {
        AddressHolder addressHolder = new AddressHolder();

        // use
        addressHolder.setUse(use);

        // address line
        addressLines.forEach(addressHolder::addAddressLine);

        // postal code
        addressHolder.setPostalCode(postcode);

        // city
        addressHolder.setCity(city);

        addressHolder.setCountry(country);

        return addressHolder;
    }

    private static PhoneHolder mapPhoneHolder(String phone) {
        PhoneHolder phoneHolder = new PhoneHolder();

        // use
        phoneHolder.addUse("WP");

        // value
        phoneHolder.setValue("tel:" + phone);

        return phoneHolder;
    }

    public static PrimaryPersonInformationHolder mapPersonHolder(RegisterData registerData) {
        String cpr = registerData.getCprnr();

        PrimaryPersonInformationHolder personHolder = new PrimaryPersonInformationHolder();

        // cpr
        personHolder.setCpr(cpr);

        RegularCPRPersonType regularCPRPerson = registerData.getScesData()
                .getPersonWithHealthCareInformationStructure() 
                .getPersonInformationStructure() 
                .getRegularCPRPerson(); 

        PersonNameStructureType personNameStructure = regularCPRPerson
                .getSimpleCPRPerson() 
                .getPersonNameStructure(); 

        if (personNameStructure != null) {
            // givens
            // Split given names by whitespace
            String givenNames = Optional.ofNullable(personNameStructure.getPersonGivenName())
                    .filter(s -> !s.isEmpty())
                    .orElse(null);
            Arrays.stream(ArrayUtils.nullToEmpty(StringUtils.split(givenNames)))
                    .forEach(personHolder::addGiven);

            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonMiddleName()))) 
                    .forEach(personHolder::addGiven); // add middle name to given names

            // family
            String familyName = Optional.ofNullable(personNameStructure.getPersonSurnameName())
                    .filter(s -> !s.isEmpty())
                    .orElse(null);
            personHolder.setFamily(familyName);
        }

        if (cpr != null && cpr.length() == 10) {
            boolean male = "13579".contains(cpr.substring(9)); // is 10th char 1, 3, 5, 7 or 9?

            // gender
            personHolder.setGender(male ? "M" : "F");
        }

        XMLGregorianCalendar birthDate = regularCPRPerson
                .getPersonBirthDateStructure() 
                .getBirthDate(); 

        // birthTime
        personHolder.setBirthDate(birthDate.toGregorianCalendar().toZonedDateTime());

        // address
        personHolder.setAddress(mapAddressHolder(registerData));

        return personHolder;
    }

    public static AddressHolder mapAddressHolder(RegisterData registerData) {
        AddressHolder addressHolder = new AddressHolder();

        Optional<PersonInformationStructureType> personInformationStructureType = ofNullable(registerData)
                .map(RegisterData::getScesData)
                .map(GetPersonWithHealthCareInformationOut::getPersonWithHealthCareInformationStructure)
                .map(PersonWithHealthCareInformationStructureType::getPersonInformationStructure);
        AddressCompleteType addressComplete = personInformationStructureType
                .map(PersonInformationStructureType::getPersonAddressStructure).get().getAddressComplete();


        if (addressComplete != null) {
            AddressPostalType addressPostal = addressComplete.getAddressPostal(); 

            // use
            if (personInformationStructureType
                    .map(PersonInformationStructureType::getRegularCPRPerson)
                    .map(RegularCPRPersonType::isPersonInformationProtectionIndicator).get()) {
                addressHolder.setUse("CONF");
            } else {
                addressHolder.setUse("H");
            }

            if (addressPostal != null) {
                // address line 1
                addressHolder.addAddressLine(
                        PdcDataMapperCommon.buildAddressLine1(
                                addressPostal.getStreetName(), 
                                addressPostal.getStreetBuildingIdentifier(), 
                                addressPostal.getFloorIdentifier(), 
                                addressPostal.getSuiteIdentifier())); 

                // address line 2
                ofNullable(
                        addressPostal.getPostOfficeBoxIdentifier()) 
                        .ifPresent(integer ->
                                addressHolder.addAddressLine(
                                        String.valueOf(integer)));

                // address line 3
                ofNullable(addressPostal.getDistrictSubdivisionIdentifier()) 
                        .ifPresent(addressHolder::addAddressLine);

                // address line 4
                ofNullable(addressPostal.getMailDeliverySublocationIdentifier())
                        .ifPresent(addressHolder::addAddressLine);

                addressHolder.setPostalCode(addressPostal.getPostCodeIdentifier());

                addressHolder.setCity(addressPostal.getDistrictName());

                of(addressPostal)
                        .map(AddressPostalType::getCountryIdentificationCode)
                        .map(CountryIdentificationCodeType::getValue)
                        .ifPresent(addressHolder::setCountry);
            }
        }

        return addressHolder;
    }

    private static Optional<RepresentedOrganization> toRepresentedOrg(POCDMT000040Author author) {
        return ofNullable(author.getAssignedAuthor().getRepresentedOrganization())
                .flatMap(org -> org.getNames().stream().findFirst()
                        .flatMap(on -> on.getContent().stream().findFirst().map(Object::toString))
                        .map(RepresentedOrganization::new));
    }

    private static DataEnterer toDataEnterer(POCDMT000040Author author) {
        String code = author.getAssignedAuthor().getCode().getCode();
        String displayName = author.getAssignedAuthor().getCode().getDisplayName();
        boolean isCitizen = code.equals("EnteredCitizen") && displayName.equals("Indtastet af borger");
        boolean isHealthcareProfessional = code.equals("EnteredHealthcareprofessional") &&
                    displayName.equals("Indtastet af sundhedsprofessionel");
        boolean isRelatedToCitizen = code.equals("EnteredCitizenRelative") &&
                    displayName.equals("Indtastet af borgers pårørende");
        DataEnterer de = isCitizen ? EnteredCitizen :
                         isHealthcareProfessional ? EnteredHealthcareprofessional :
                         isRelatedToCitizen ? EnteredCitizenRelative :
                         null;
        if (de == null) {
            throw new RuntimeException(
                    "Response from SKR contains invalid DataEnterer with code '" + code + "' and displayName '" +
                            displayName + "'");
        }
        return de;
    }
}