/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler;

import dk.stamkort.util.AuthorProperties;
import dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_2_0ElementsMapper;
import dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_2_0NarrativeTextMapper;
import dk.stamkort.webservice.hl7cda.model.*;
import org.hl7.v3.*;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.stream.Collectors;

import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfNull;
import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfWrongSize;
import static dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_2_0ElementsMapper.*;

public class PDC_2_0ClinicalDocumentAssembler {

    public static final ObjectFactory o = new ObjectFactory();

    public static POCDMT000040ClinicalDocument assembleClinicalDocument(PdcV2Data dataHolder, AuthorProperties properties) {
        PrimaryPersonInformationHolder primaryPersonInformation = dataHolder.getPrimaryPersonInformation();
        ProviderOrganizationHolder providerOrganizationHolder = dataHolder.getProviderOrganizationHolder();
        List<CustodyHolder> custodyHolderList = dataHolder.getCustodyHolderList();
        CoverageGroupHolder coverageGroupHolder = dataHolder.getCoverageGroupHolder();
        OrganDonorRegistrationHolder organDonorRegistrationHolder = dataHolder.getOrganDonorRegistrationHolder();
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = dataHolder.getTreatmentWillRegistrationHolder();
        LivingWillRegistrationHolder livingWillRegistrationHolder = dataHolder.getLivingWillRegistrationHolder();
        List<LanguageTypedInHolder> languageTypedInHolderList = dataHolder.getLanguageTypedInHolder();
        List<TempAddrTypedInHolder> tempAddrTypedInHolderList = dataHolder.getTempAddrTypedInHolder();
        Optional<DentistTypedInHolder> dentistTypedInHolder = dataHolder.getDentistTypedInHolder();
        List<PatientContactTypedInHolder> patientContactTypedInHolderList = dataHolder.getPatientContactTypedInHolder();
        List<RelativeTypedInHolder> relativeTypedInHolderList = dataHolder.getRelativeTypedInHolder();

        failIfNull(primaryPersonInformation);
        //providerOrganizationHolder  // no validation needed
        //custodyHolderList // no validation needed
        failIfNull(coverageGroupHolder);
        failIfNull(organDonorRegistrationHolder);
        failIfNull(treatmentWillRegistrationHolder);
        failIfNull(livingWillRegistrationHolder);
        failIfWrongSize(languageTypedInHolderList, 0, 1);
        failIfWrongSize(tempAddrTypedInHolderList, 0, 1);
        failIfWrongSize(patientContactTypedInHolderList, 0, 1);
        // relativeTypedInHolderList // no validation needed

        final ZonedDateTime now = ZonedDateTime.now(); // now

        // ClinicalDocument skeleton
        POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();
        doc.getMoodCodes().add("EVN");
        addSimpleTopLevelElements(
                doc,
                primaryPersonInformation,
                now);

        // recordTarget
        doc.getRecordTargets().add(
                createRecordTarget(
                        primaryPersonInformation,
                        providerOrganizationHolder));

        // author
        doc.getAuthors().add(
                createAuthor(
                        null,
                        createTS(now),
                        createAssignedAuthor(
                                createII(properties.getAuthorInstitutionRoot(), properties.getAuthorInstitutionExtension(), properties.getAuthorInstitutionAssigningAuthorityName()),
                                createOrganization(createON(properties.getAuthorInstitutionName())),
                                null)));

        // custodian
        doc.setCustodian(createCustodian(createAssignedCustodian(
                createCustodianOrganization(
                        createII(properties.getAuthorInstitutionRoot(), properties.getAuthorInstitutionExtension(), properties.getAuthorInstitutionAssigningAuthorityName()),
                        createON(properties.getAuthorInstitutionName())
                ))));

        // documentationOf (creation time)
        doc.getDocumentationOves().add(
                createDocumentationOf(
                        createServiceEvent(
                                createII("1.2.208.184.10.1.11", null, null),
                                null,
                                createIVLTS(now, null))));

        // documentationOf (CDA-profile version)
        doc.getDocumentationOves().add(
                createDocumentationOf(
                        createServiceEvent(
                                createII("1.2.208.184.10.1.10", "2019-09-10", null),
                                createII("1.2.208.184.100.10", "pdc-v2.0", "MedCom"),
                                null)));

        // component
        II ii = createII("1.2.208.184.16.1.10.20.1", "2019-08-14", null);
        StrucDocText strucDocText = createStrucDocText(
                PDC_2_0NarrativeTextMapper.createTextContentElements(
                        primaryPersonInformation,
                        custodyHolderList,
                        coverageGroupHolder,
                        organDonorRegistrationHolder,
                        treatmentWillRegistrationHolder,
                        livingWillRegistrationHolder,
                        languageTypedInHolderList,
                        tempAddrTypedInHolderList,
                        dentistTypedInHolder,
                        patientContactTypedInHolderList,
                        relativeTypedInHolderList));
        List<POCDMT000040Entry> entries = getDocumentEntries(
                primaryPersonInformation,
                custodyHolderList,
                coverageGroupHolder,
                organDonorRegistrationHolder,
                treatmentWillRegistrationHolder,
                livingWillRegistrationHolder,
                languageTypedInHolderList,
                tempAddrTypedInHolderList,
                dentistTypedInHolder,
                patientContactTypedInHolderList,
                relativeTypedInHolderList);

        doc.setComponent(
                createComponent2(
                        createStructuredBody(
                                createComponent3(
                                        createSection(ii, strucDocText, entries)))));

        return doc;
    }

    private static void addSimpleTopLevelElements(POCDMT000040ClinicalDocument doc, PrimaryPersonInformationHolder personHolder, ZonedDateTime effectiveTime) {
        doc.getRealmCodes().add(createCS("DK"));
        doc.setTypeId(createTypeId("2.16.840.1.113883.1.3", "POCD_HD000040"));
        doc.getTemplateIds().add(createII("1.2.208.184.16.1", null, null));
        doc.setId(createII("1.2.208.184", UUID.randomUUID().toString(), "MedCom"));
        String pdcTypeCode = "PDC";
        doc.setCode(createCE(pdcTypeCode, "1.2.208.184.100.1", "Stamkort", "MedCom Message Codes"));
        doc.setTitle(createST("Personal Data Card for " + personHolder.getCpr()));
        doc.setEffectiveTime(createTS(effectiveTime));
        doc.setConfidentialityCode(createCE("N", "2.16.840.1.113883.5.25", null, null));
        doc.setLanguageCode(createCS("da-DK"));
    }

    private static List<POCDMT000040Entry> getDocumentEntries(PrimaryPersonInformationHolder primaryPersonInformation,
                                                              List<CustodyHolder> custodyHolder,
                                                              CoverageGroupHolder coverageGroupHolder,
                                                              OrganDonorRegistrationHolder organDonorRegistrationHolder,
                                                              TreatmentWillRegistrationHolder treatmentWillRegistrationHolder,
                                                              LivingWillRegistrationHolder livingWillRegistrationHolder,
                                                              List<LanguageTypedInHolder> languageTypedInHolder,
                                                              List<TempAddrTypedInHolder> tempAddrTypedInHolder,
                                                              Optional<DentistTypedInHolder> dentistTypedInHolder,
                                                              List<PatientContactTypedInHolder> patientContactTypedInHolder,
                                                              List<RelativeTypedInHolder> relativeTypedInHolder) {
        List<POCDMT000040Entry> entries = new ArrayList<>();

        // "Forældremyndighed over" observation
        custodyHolder.forEach(value ->
                entries.add(
                        createEntry(
                                createCustodyObservation(value))));

        // "Borgerens navn og adresse" observation
        entries.add(
                createEntry(
                        createCitizenNameAddrObservation(primaryPersonInformation)));

        // "Sygesikringsgruppe" observation
        entries.add(
                createEntry(
                        createCoverageGroupObservation(coverageGroupHolder)));

        // "Registreret organdonor" observation
        entries.add(
                createEntry(
                        createOrganDonorRegistrationObservation(organDonorRegistrationHolder)));

        // "Registreret behandlingstestamente" observation
        entries.add(
                createEntry(
                        createTreatmentWillRegistrationObservation(treatmentWillRegistrationHolder)));

        // "Registreret livstestamente" observation
        entries.add(
                createEntry(
                        createLivingWillRegistrationObservation(livingWillRegistrationHolder)));

        // "Talt sprog, indtastet" observation
        languageTypedInHolder.forEach(value ->
                entries.add(
                        createEntry(
                                createLanguageTypedInObservation(value))));

        // "Midlertidig adresse, indtastet" observation
        tempAddrTypedInHolder.forEach(value ->
                entries.add(
                        createEntry(
                                createTempAddrTypedInObservation(value))));

        // "Tandlæge indtastet" observation
        dentistTypedInHolder.ifPresent(value ->
                entries.add(
                        createEntry(
                                createDentistTypedInObservation(value))));

        // "Kontaktoplysninger, indtastet" observation
        patientContactTypedInHolder.forEach(value ->
                entries.add(
                        createEntry(
                                createPatientContactTypedInObservation(value))));

        // "Pårørende, indtastet" observation
        relativeTypedInHolder.forEach(value ->
                entries.add(
                        createEntry(
                                createRelativeTypedInObservation(value))));

        return entries;
    }

    public static POCDMT000040RecordTarget createRecordTarget(PrimaryPersonInformationHolder personHolder, ProviderOrganizationHolder providerOrganizationHolder) {
        POCDMT000040RecordTarget recordTarget = o.createPOCDMT000040RecordTarget();
        recordTarget.setContextControlCode("OP");
        recordTarget.getTypeCodes().add("RCT");
        recordTarget.setPatientRole(
                createPatientRole(personHolder, providerOrganizationHolder));
        return recordTarget;
    }

    private static POCDMT000040PatientRole createPatientRole(PrimaryPersonInformationHolder personHolder, ProviderOrganizationHolder providerOrganizationHolder) {
        POCDMT000040PatientRole patientRole = o.createPOCDMT000040PatientRole();
        patientRole.getClassCodes().add("PAT"); // attribute

        // id
        patientRole.getIds().add(createII("1.2.208.176.1.2", personHolder.getCpr(), "CPR"));

        // addr
        AddressHolder address = personHolder.getAddress();
        patientRole.getAddrs().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));

        // patient
        PN personName = createPN(personHolder.getGivens(), personHolder.getFamily());
        CE administrativeGenderCode = Optional.ofNullable(personHolder.getGender())
                .map(s -> createCE(personHolder.getGender(), "2.16.840.1.113883.5.1", null, null))
                .orElse(null);
        TS ts = Optional.ofNullable(personHolder.getBirthDateAsLocalDate())
                .map(PDC_2_0ElementsMapper::createDateTS)
                .orElse(null);

        patientRole.setPatient(
                createPatient(personName, administrativeGenderCode, ts));

        // ProviderOrganization
        if (providerOrganizationHolder != null) {
            II id = null;
            if (providerOrganizationHolder.getYderNr() != null && providerOrganizationHolder.getYderNr().value != null) {
                // yder
                id = createII("1.2.208.176.1.4", providerOrganizationHolder.getYderNr().value, "Yderregisteret");
            }

            ON name = Optional.ofNullable(providerOrganizationHolder.getName())
                    .map(PDC_2_0ElementsMapper::createON)
                    .orElse(null);

            List<TEL> workPhones = providerOrganizationHolder.getPhones()
                    .stream()
                    .map(value -> PDC_2_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                    .collect(Collectors.toList());

            AD addrs = Optional.ofNullable(providerOrganizationHolder.getAddress())
                    .map(providerOrganizationAddress ->
                            createAD(
                                    providerOrganizationAddress.getUse(),
                                    providerOrganizationAddress.getAddressLines(),
                                    providerOrganizationAddress.getPostalCode(),
                                    providerOrganizationAddress.getCity(),
                                    providerOrganizationAddress.getCountry()))
                    .orElse(null);
            patientRole.setProviderOrganization(createProviderOrganization(id, name, workPhones, addrs));
        }

        return patientRole;
    }

    public static POCDMT000040Observation createCustodyObservation(CustodyHolder custodyHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.23", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD(custodyHolder.getCode().getCode(), "1.2.208.184.100.1", custodyHolder.getCode().getDisplayName(), "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.2", custodyHolder.getCpr(), "CPR"));
        observation.getValues().add(createPN(custodyHolder.getGivens(), custodyHolder.getFamily()));
        observation.getValues().add(createCD(custodyHolder.getRelation().getCode(), "1.2.208.184.100.2", custodyHolder.getRelation().getDisplayName(), "MedCom Relation Codes"));
        observation.getAuthors().add(
                createAuthor(createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "NI", "CPR"),
                                createOrganization(createON("CPR")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createCitizenNameAddrObservation(PrimaryPersonInformationHolder primaryPersonInformation) {
        POCDMT000040Observation observation = createObservation();


        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.26", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("CitizenNameAddr", "1.2.208.184.100.1", "Borgerens navn og adresse", "MedCom Message Codes"));
        observation.getValues().add(createPN(primaryPersonInformation.getGivens(), primaryPersonInformation.getFamily()));
        AddressHolder address = primaryPersonInformation.getAddress();
        observation.getValues().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "NI", "CPR"),
                                createOrganization(createON("CPR")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createObservation() {
        POCDMT000040Observation observation = new POCDMT000040Observation();
        observation.getClassCodes().add("OBS");
        observation.setMoodCode(XActMoodDocumentObservation.EVN);

        return observation;
    }

    public static POCDMT000040Observation createCoverageGroupObservation(CoverageGroupHolder coverageGroupHolder) {
        String group = coverageGroupHolder.getGroup();

        failIfNull(group);

        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.27", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("CoverageGroup", "1.2.208.184.100.1", "Sygesikringsgruppe", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.99.99.99", group, "Sygesikringen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.2.7", "NI", "Sygesikringen"),
                                createOrganization(createON("Sygesikringen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createOrganDonorRegistrationObservation(OrganDonorRegistrationHolder organDonorRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.28", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("OrganDonorRegistration", "1.2.208.184.100.1", "Registreret organdonor", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.10", Boolean.toString(organDonorRegistrationHolder.isHasRegistration()), "Dansk Center For Organdonation"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.10", "NI", "CPR"),
                                createOrganization(createON("Dansk Center For Organdonation")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createTreatmentWillRegistrationObservation(TreatmentWillRegistrationHolder treatmentWillRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.29", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("TreatmentWillRegistration", "1.2.208.184.100.1", "Registreret behandlingstestamente", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.9", Boolean.toString(treatmentWillRegistrationHolder.isHasRegistration()), "Sundhedsdatastyrelsen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.9", "NI", "Sundhedsdatastyrelsen"),
                                createOrganization(createON("Sundhedsdatastyrelsen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createLivingWillRegistrationObservation(LivingWillRegistrationHolder livingWillRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.30", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("LivingWillRegistration", "1.2.208.184.100.1", "Registreret livstestamente", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.8", Boolean.toString(livingWillRegistrationHolder.isHasRegistration()), "Sundhedsdatastyrelsen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.8", "NI", "Sundhedsdatastyrelsen"),
                                createOrganization(createON("Sundhedsdatastyrelsen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createLanguageTypedInObservation(LanguageTypedInHolder languageTypedInHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.20", "2019-08-14", null));
        observation.getIds().add(createII(languageTypedInHolder.getId().getRoot(), languageTypedInHolder.getId().getExtension(), languageTypedInHolder.getId().getAssigningAuthorityName()));
        observation.setCode(createCD("LanguageTypedIn", "1.2.208.184.100.1", "Talt sprog, indtastet", "MedCom Message Codes"));
        observation.getValues().add(createCD(languageTypedInHolder.getLanguageCode(), "1.0.639.1", languageTypedInHolder.getLanguageName(), "ISO-639-1"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(languageTypedInHolder.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                null,
                                createAssignedPerson(
                                        createPN(languageTypedInHolder.getAuthorGivens(), languageTypedInHolder.getAuthorFamily())))));

        return observation;
    }

    public static POCDMT000040Observation createTempAddrTypedInObservation(TempAddrTypedInHolder tempAddrTypedInHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.21", "2019-08-14", null));
        observation.getIds().add(createII(tempAddrTypedInHolder.getId().getRoot(), tempAddrTypedInHolder.getId().getExtension(), tempAddrTypedInHolder.getId().getAssigningAuthorityName()));
        observation.setCode(createCD("TempAddrTypedIn", "1.2.208.184.100.1", "Midlertidig adresse, indtastet", "MedCom Message Codes"));
        // PeriodFrom is mandatory. PeriodTo is optional
        if(tempAddrTypedInHolder.getPeriodTo() != null) {
            observation.setEffectiveTime(createIVLTS(
                    tempAddrTypedInHolder.getPeriodFrom().atStartOfDay(ZoneId.systemDefault()), // system's default timezone... what else?
                    tempAddrTypedInHolder.getPeriodTo().atStartOfDay(ZoneId.systemDefault()))); // system's default timezone... what else?
        }
        else {
            observation.setEffectiveTime(createIVLTS(
                    tempAddrTypedInHolder.getPeriodFrom().atStartOfDay(ZoneId.systemDefault()), null)); // system's default timezone... what else?
        }
        AddressHolder address = tempAddrTypedInHolder.getAddress();
        observation.getValues().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(tempAddrTypedInHolder.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                null,
                                createAssignedPerson(
                                        createPN(tempAddrTypedInHolder.getAuthorGivens(), tempAddrTypedInHolder.getAuthorFamily())))));

        return observation;
    }

    public static POCDMT000040Observation createDentistTypedInObservation(DentistTypedInHolder dentistTypedInHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.22", "2019-08-14", null));
        observation.getIds().add(createII(dentistTypedInHolder.getId().getRoot(), dentistTypedInHolder.getId().getExtension(), dentistTypedInHolder.getId().getAssigningAuthorityName()));
        observation.setCode(createCD("DentistTypedIn", "1.2.208.184.100.1", "Tandlæge indtastet", "MedCom Message Codes"));

        if (dentistTypedInHolder.getYder() != null) {
            // yder
            observation.getValues().add(createII("1.2.208.176.1.4", dentistTypedInHolder.getYder(), "Yderregisteret"));
        } else if (dentistTypedInHolder.getSor() != null) {
            // sor
            observation.getValues().add(createII("1.2.208.176.1.1", dentistTypedInHolder.getSor(), "SOR"));
        }

        observation.getValues().add(createEN(dentistTypedInHolder.getPrefix(), dentistTypedInHolder.getGivens(), dentistTypedInHolder.getFamily()));
        AddressHolder address = dentistTypedInHolder.getAddress();
        observation.getValues().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));
        String phoneValue = dentistTypedInHolder.getPhone().map(PhoneHolder::getValue).orElse(null);
        observation.getValues().add(PDC_2_0ElementsMapper.createTELHandleNull(Collections.singletonList("WP"), phoneValue));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(dentistTypedInHolder.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                null,
                                createAssignedPerson(
                                        createPN(dentistTypedInHolder.getAuthorGivens(), dentistTypedInHolder.getAuthorFamily())))));

        return observation;
    }

    public static POCDMT000040Observation createPatientContactTypedInObservation(PatientContactTypedInHolder patientContactTypedInHolder) {
        List<PhoneHolder> phones = patientContactTypedInHolder.getPhones();
        List<PhoneHolder> emailAddresses = patientContactTypedInHolder.getEmailAddresses();

        failIfWrongSize(phones, 0, 3);
        failIfWrongSize(emailAddresses, 0, 1);

        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.24", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("PatientContactTypedIn", "1.2.208.184.100.1", "Kontaktoplysninger, indtastet", "MedCom Message Codes"));
        observation.getValues().addAll(
                phones
                        .stream()
                        .map(value -> PDC_2_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                        .collect(Collectors.toList()));
        observation.getValues().addAll(
                emailAddresses
                        .stream()
                        .map(value -> PDC_2_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                        .collect(Collectors.toList()));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(patientContactTypedInHolder.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                null,
                                createAssignedPerson(
                                        createPN(patientContactTypedInHolder.getAuthorGivens(), patientContactTypedInHolder.getAuthorFamily())))));
        return observation;
    }

    public static POCDMT000040Observation createRelativeTypedInObservation(RelativeTypedInHolder relativeTypedInHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.25", "2019-08-14", null));
        observation.getIds().add(createII(relativeTypedInHolder.getId().getRoot(), relativeTypedInHolder.getId().getExtension(), relativeTypedInHolder.getId().getAssigningAuthorityName()));
        observation.setCode(createCD("RelativeTypedIn", "1.2.208.184.100.1", "Pårørende, indtastet", "MedCom Message Codes"));
        // Skal acceptere manglende efternavn, selvom det afviger fra MedCom standarden, iht. [SDS-5182]
        observation.getValues().add(createPN(relativeTypedInHolder.getGivens(), relativeTypedInHolder.getFamily() != null ? relativeTypedInHolder.getFamily() : ""));
        observation.getValues().add(createCD(relativeTypedInHolder.getRelation().getCode(), "1.2.208.184.100.2", relativeTypedInHolder.getRelation().getDisplayName(), "MedCom Relation Codes"));
        List<TEL> workPhones = relativeTypedInHolder.getPhoneHolders()
                .stream()
                .map(value -> PDC_2_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                .collect(Collectors.toList());
        observation.getValues().addAll(workPhones);
        if (relativeTypedInHolder.getNote() != null) {
            observation.getValues().add(createST(relativeTypedInHolder.getNote()));
        }
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(relativeTypedInHolder.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                null,
                                createAssignedPerson(
                                        createPN(relativeTypedInHolder.getAuthorGivens(), relativeTypedInHolder.getAuthorFamily())))));

        return observation;
    }
}

