/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler;

import dk.stamkort.util.AuthorProperties;
import dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_3_0ElementsMapper;
import dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_3_0NarrativeTextMapper;
import dk.stamkort.webservice.hl7cda.model.*;
import dk.stamkort.webservice.hl7cda.model.v3.*;
import org.hl7.v3.*;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfNull;
import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfWrongSize;
import static dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_3_0ElementsMapper.*;
import static java.util.Collections.*;
import static org.hl7.v3.XActMoodDocumentObservation.EVN;

public class PDC_3_0ClinicalDocumentAssembler {

    public final static String
            spokenLanguageRootId = "1.2.208.184.16.1.10.20.1.20",
            temporaryAddressRootId = "1.2.208.184.16.1.10.20.1.21",
            dentistRootId = "1.2.208.184.16.1.10.20.1.22",
            contactInfoRootId = "1.2.208.184.16.1.10.20.1.24",
            relativeRootId = "1.2.208.184.16.1.10.20.1.25";

    private PDC_3_0ClinicalDocumentAssembler() { }

    public static POCDMT000040ClinicalDocument assembleClinicalDocument(PdcV3Data data, AuthorProperties properties) {
        PrimaryPersonInformationHolder primaryPersonInformation = data.getPrimaryPersonInformation();
        Optional<ProviderOrganizationHolder> providerOrganizationHolder = data.getProviderOrganizationHolder();
        List<CustodyHolder> custodyHolderList = data.getCustodyHolderList();
        CoverageGroupHolder coverageGroupHolder = data.getCoverageGroupHolder();
        OrganDonorRegistrationHolder organDonorRegistrationHolder = data.getOrganDonorRegistrationHolder();
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = data.getTreatmentWillRegistrationHolder();
        LivingWillRegistrationHolder livingWillRegistrationHolder = data.getLivingWillRegistrationHolder();

        failIfNull(primaryPersonInformation);
        failIfNull(coverageGroupHolder);
        failIfNull(organDonorRegistrationHolder);
        failIfNull(treatmentWillRegistrationHolder);
        failIfNull(livingWillRegistrationHolder);

        final ZonedDateTime now = ZonedDateTime.now();

        POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();
        doc.getMoodCodes().add("EVN");
        addSimpleTopLevelElements(
                doc,
                primaryPersonInformation,
                now);

        // recordTarget
        doc.getRecordTargets().add(
                createRecordTarget(
                        primaryPersonInformation,
                        providerOrganizationHolder));

        // author
        doc.getAuthors().add(
                createAuthor(
                        null,
                        createTS(now),
                        createAssignedAuthor(
                                createII(properties.getAuthorInstitutionRoot(), properties.getAuthorInstitutionExtension(), properties.getAuthorInstitutionAssigningAuthorityName()),
                                createOrganization(createON(properties.getAuthorInstitutionName())),
                                null)));

        // custodian
        doc.setCustodian(createCustodian(createAssignedCustodian(
                createCustodianOrganization(
                        createII(properties.getAuthorInstitutionRoot(), properties.getAuthorInstitutionExtension(), properties.getAuthorInstitutionAssigningAuthorityName()),
                        createON(properties.getAuthorInstitutionName())))));

        // documentationOf (creation time)
        doc.getDocumentationOves().add(
                createDocumentationOf(
                        createServiceEvent(
                                createII("1.2.208.184.10.1.11", null, null),
                                null,
                                createIVLTS(now, null))));

        // documentationOf (CDA-profile version)
        doc.getDocumentationOves().add(
                createDocumentationOf(
                        createServiceEvent(
                                createII("1.2.208.184.10.1.10", "2019-09-10", null),
                                createII("1.2.208.184.100.10", "pdc-v3.0", "MedCom"),
                                null)));

        // component
        II ii = createII("1.2.208.184.16.1.10.20.1", "2019-08-14", null);
        StrucDocText strucDocText = createStrucDocText(
                PDC_3_0NarrativeTextMapper.createTextContentElements(
                        primaryPersonInformation,
                        custodyHolderList,
                        coverageGroupHolder,
                        organDonorRegistrationHolder,
                        treatmentWillRegistrationHolder,
                        livingWillRegistrationHolder,
                        data.getLanguageTypedInV3(),
                        data.getTempAddrTypedInV3(),
                        data.getDentistTypedInV3(),
                        data.getPatientContactTypedInV3(),
                        data.getRelativeTypedInV3(),
                        data.hasNoResuscitation
                ));
        List<POCDMT000040Entry> entries = getDocumentEntries(
                primaryPersonInformation,
                custodyHolderList,
                coverageGroupHolder,
                organDonorRegistrationHolder,
                treatmentWillRegistrationHolder,
                livingWillRegistrationHolder,
                data.getLanguageTypedInV3(),
                data.getTempAddrTypedInV3(),
                data.getDentistTypedInV3(),
                data.getPatientContactTypedInV3(),
                data.getRelativeTypedInV3(),
                data.hasNoResuscitation);

        doc.setComponent(createComponent2(createStructuredBody(createComponent3(
                createSection(ii, strucDocText, entries)))));
        return doc;
    }

    public static POCDMT000040Observation createDentistTypedInObservation(DentistTypedInV3 dth) {
        POCDMT000040Observation observation = createObservation();
        observation.getTemplateIds().add(createII(dentistRootId, "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", dth.getIdExtension(), null));
        observation.setCode(createCD("DentistTypedIn", "1.2.208.184.100.1", "Tandlæge indtastet",
                                     "MedCom Message Codes"));

        List<ANY> values = observation.getValues();
        values.add(createII("1.2.208.176.1.4", dth.getYder(), "Yderregisteret"));

        dth.getName().ifPresent(n -> values.add(createPN(singletonList(n))));
        values.add(createTN(dth.getOrganizationName()));
        AddressHolder address = dth.getAddress();
        values.add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(),
                            address.getCountry()));
        dth.getPhone().map(PhoneHolder::getValue)
                .ifPresent(pn -> values.add(PDC_3_0ElementsMapper.createTELHandleNull(singletonList("WP"), pn)));

        AssignedAuthor aa = dth.getAssignedAuthor();
        DataEnterer de = aa.dataEnterer;
        boolean isHealthcareProfessional = de == DataEnterer.EnteredHealthcareprofessional;
        POCDMT000040Organization representedOrg = isHealthcareProfessional
                ? createOrganization(createON(aa.representedOrganization.get().value))
                : null;
        POCDMT000040Person assignedPerson = buildAssignedPerson(dth.getAuthorGivens(), dth.getAuthorFamily(), isHealthcareProfessional);

        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(dth.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                createCE(de.code, "1.2.208.184.100.1", de.displayName, "MedCom Message Codes"),
                                representedOrg,
                                assignedPerson)));
        return observation;
    }

    public static Stream<POCDMT000040Observation> findObservation(POCDMT000040ClinicalDocument doc,
                                                                  String templateRootId) {
        return doc.getComponent().getStructuredBody().getComponents().stream()
                .flatMap(c -> c.getSection().getEntries().stream())
                .map(POCDMT000040Entry::getObservation)
                .filter(o -> o.getTemplateIds().stream().anyMatch(tid -> tid.getRoot().equals(templateRootId)));
    }

    private static POCDMT000040Observation noResuscitationRegistration(Optional<Boolean> hasNoResuscitation) {
        POCDMT000040Observation o = new POCDMT000040Observation();
        o.getClassCodes().add("OBS");
        o.setMoodCode(EVN);
        o.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.31", "2023-07-01", null));
        o.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        o.setCode(createCD("NoResuscitationRegistration",
                           "1.2.208.184.100.1",
                           "Registreret fravalg af genoplivningsforsøg v. hjertestop",
                           "MedCom Message Codes"));
        hasNoResuscitation.map(r -> Boolean.toString(r))
                .ifPresent(v -> o.getValues().add(createII("1.2.208.176.1.11", v, "Sundhedsdatastyrelsen")));
        if(!hasNoResuscitation.isPresent()) {
            II ii = new II() { }; ii.getNullFlavors().add("NAV");
            o.getValues().add(ii);
        }
        o.getAuthors().add(createAuthor(createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                                        createTS(null),
                                        createAssignedAuthor(createII("1.2.208.176.1.8", "NI", "Sundhedsdatastyrelsen"),
                                                             createOrganization(createON("Sundhedsdatastyrelsen")),
                                                             null)));
        return o;
    }

    private static void addSimpleTopLevelElements(POCDMT000040ClinicalDocument doc, PrimaryPersonInformationHolder personHolder, ZonedDateTime effectiveTime) {
        doc.getRealmCodes().add(createCS("DK"));
        doc.setTypeId(createTypeId("2.16.840.1.113883.1.3", "POCD_HD000040"));
        doc.getTemplateIds().add(createII("1.2.208.184.16.1", "3.0", null));
        doc.setId(createII("1.2.208.184", UUID.randomUUID().toString(), "MedCom"));
        String pdcTypeCode = "PDC";
        doc.setCode(createCE(pdcTypeCode, "1.2.208.184.100.1", "Stamkort", "MedCom Message Codes"));
        doc.setTitle(createST("Personal Data Card for " + personHolder.getCpr()));
        doc.setEffectiveTime(createTS(effectiveTime));
        doc.setConfidentialityCode(createCE("N", "2.16.840.1.113883.5.25", null, null));
        doc.setLanguageCode(createCS("da-DK"));
    }

    private static List<POCDMT000040Entry> getDocumentEntries(PrimaryPersonInformationHolder primaryPersonInformation,
                                                              List<CustodyHolder> custodyHolder,
                                                              CoverageGroupHolder coverageGroupHolder,
                                                              OrganDonorRegistrationHolder organDonorRegistrationHolder,
                                                              TreatmentWillRegistrationHolder treatmentWillRegistrationHolder,
                                                              LivingWillRegistrationHolder livingWillRegistrationHolder,
                                                              Optional<LanguageTypedInV3> languageTypedIn,
                                                              Optional<TempAddrTypedInV3> tempAddrTypedIn,
                                                              Optional<DentistTypedInV3> dentistTypedIn,
                                                              Optional<PatientContactTypedInV3> patientContactTypedIn,
                                                              List<RelativeTypedInV3> relativeTypedIn,
                                                              Optional<Boolean> hasNoResuscitation) {
        List<POCDMT000040Entry> entries = new ArrayList<>();
        Consumer<POCDMT000040Observation> addToEntries = o -> entries.add(createEntry(o));
        custodyHolder.forEach(value -> addToEntries.accept(createCustodyObservation(value)));
        addToEntries.accept(createCitizenNameAddrObservation(primaryPersonInformation));
        addToEntries.accept(createCoverageGroupObservation(coverageGroupHolder));
        addToEntries.accept(createOrganDonorRegistrationObservation(organDonorRegistrationHolder));
        addToEntries.accept(createTreatmentWillRegistrationObservation(treatmentWillRegistrationHolder));
        addToEntries.accept(createLivingWillRegistrationObservation(livingWillRegistrationHolder));
        languageTypedIn.ifPresent(value -> addToEntries.accept(createLanguageTypedInObservation(value)));
        tempAddrTypedIn.ifPresent(value -> addToEntries.accept(createTempAddrTypedInObservation(value)));
        dentistTypedIn.ifPresent(value -> addToEntries.accept(createDentistTypedInObservation(value)));
        patientContactTypedIn.ifPresent(value -> addToEntries.accept(createPatientContactTypedInObservation(value)));
        relativeTypedIn.forEach(value -> addToEntries.accept(createRelativeTypedInObservation(value)));
        addToEntries.accept(noResuscitationRegistration(hasNoResuscitation));
        return entries;
    }

    public static POCDMT000040RecordTarget createRecordTarget(PrimaryPersonInformationHolder personHolder,
                                                              Optional<ProviderOrganizationHolder> providerOrganizationHolder) {
        POCDMT000040RecordTarget recordTarget = o.createPOCDMT000040RecordTarget();
        recordTarget.setContextControlCode("OP");
        recordTarget.getTypeCodes().add("RCT");
        recordTarget.setPatientRole(createPatientRole(personHolder, providerOrganizationHolder));
        return recordTarget;
    }

    private static POCDMT000040PatientRole createPatientRole(PrimaryPersonInformationHolder personHolder,
                                                             Optional<ProviderOrganizationHolder> providerOrganizationHolder) {
        POCDMT000040PatientRole patientRole = o.createPOCDMT000040PatientRole();
        patientRole.getClassCodes().add("PAT"); // attribute

        // id
        patientRole.getIds().add(createII("1.2.208.176.1.2", personHolder.getCpr(), "CPR"));

        // addr
        AddressHolder address = personHolder.getAddress();
        patientRole.getAddrs().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));

        // patient
        PN personName = createPN(personHolder.getGivens(), personHolder.getFamily());
        CE administrativeGenderCode = Optional.ofNullable(personHolder.getGender())
                .map(s -> createCE(personHolder.getGender(), "2.16.840.1.113883.5.1", null, null))
                .orElse(null);
        TS ts = Optional.ofNullable(personHolder.getBirthDate())
                .map(PDC_3_0ElementsMapper::createTS)
                .orElse(null);

        patientRole.setPatient(
                createPatient(personName, administrativeGenderCode, ts));

        // ProviderOrganization
        providerOrganizationHolder.ifPresent(org -> {
            II id = null;
            if (org.getYderNr() != null && org.getYderNr().value != null) {
                // yder
                id = createII("1.2.208.176.1.4", org.getYderNr().value, "Yderregisteret");
            }

            Optional<ON> name = Optional.ofNullable(org.getName()).map(PDC_3_0ElementsMapper::createON);

            List<TEL> workPhones = org.getPhones()
                    .stream()
                    .map(value -> PDC_3_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                    .collect(Collectors.toList());

            Optional<AD> addrs = Optional.ofNullable(org.getAddress())
                    .map(providerOrganizationAddress ->
                                 createAD(
                                         providerOrganizationAddress.getUse(),
                                         providerOrganizationAddress.getAddressLines(),
                                         providerOrganizationAddress.getPostalCode(),
                                         providerOrganizationAddress.getCity(),
                                         providerOrganizationAddress.getCountry()));
            patientRole.setProviderOrganization(createProviderOrganization(id, name, workPhones, addrs));
        });

        return patientRole;
    }

    public static POCDMT000040Observation createCustodyObservation(CustodyHolder custodyHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.23", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD(custodyHolder.getCode().getCode(), "1.2.208.184.100.1", custodyHolder.getCode().getDisplayName(), "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.2", custodyHolder.getCpr(), "CPR"));
        observation.getValues().add(createPN(custodyHolder.getGivens(), custodyHolder.getFamily()));
        observation.getValues().add(createCD(custodyHolder.getRelation().getCode(), "1.2.208.184.100.2", custodyHolder.getRelation().getDisplayName(), "MedCom Relation Codes"));
        observation.getAuthors().add(
                createAuthor(createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                             createTS(null),
                             createAssignedAuthor(
                                     createII("1.2.208.176.1.2", "NI", "CPR"),
                                     createOrganization(createON("CPR")),
                                     null)));

        return observation;
    }

    public static POCDMT000040Observation createCitizenNameAddrObservation(PrimaryPersonInformationHolder primaryPersonInformation) {
        POCDMT000040Observation observation = createObservation();


        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.26", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("CitizenNameAddr", "1.2.208.184.100.1", "Borgerens navn og adresse", "MedCom Message Codes"));
        observation.getValues().add(createPN(primaryPersonInformation.getGivens(), primaryPersonInformation.getFamily()));
        AddressHolder address = primaryPersonInformation.getAddress();
        observation.getValues().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "NI", "CPR"),
                                createOrganization(createON("CPR")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createObservation() {
        POCDMT000040Observation observation = new POCDMT000040Observation();
        observation.getClassCodes().add("OBS");
        observation.setMoodCode(XActMoodDocumentObservation.EVN);

        return observation;
    }

    public static POCDMT000040Observation createCoverageGroupObservation(CoverageGroupHolder coverageGroupHolder) {
        String group = coverageGroupHolder.getGroup();

        failIfNull(group);

        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.27", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("CoverageGroup", "1.2.208.184.100.1", "Sygesikringsgruppe", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.2.7", group, "Sygesikringen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.2.7", "NI", "Sygesikringen"),
                                createOrganization(createON("Sygesikringen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createOrganDonorRegistrationObservation(OrganDonorRegistrationHolder organDonorRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.28", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("OrganDonorRegistration", "1.2.208.184.100.1", "Registreret organdonor", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.10", Boolean.toString(organDonorRegistrationHolder.isHasRegistration()), "Dansk Center For Organdonation"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.10", "NI", "CPR"),
                                createOrganization(createON("Dansk Center For Organdonation")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createTreatmentWillRegistrationObservation(TreatmentWillRegistrationHolder treatmentWillRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.29", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("TreatmentWillRegistration", "1.2.208.184.100.1", "Registreret behandlingstestamente", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.9", Boolean.toString(treatmentWillRegistrationHolder.isHasRegistration()), "Sundhedsdatastyrelsen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.9", "NI", "Sundhedsdatastyrelsen"),
                                createOrganization(createON("Sundhedsdatastyrelsen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createLivingWillRegistrationObservation(LivingWillRegistrationHolder livingWillRegistrationHolder) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.30", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("LivingWillRegistration", "1.2.208.184.100.1", "Registreret livstestamente", "MedCom Message Codes"));
        observation.getValues().add(createII("1.2.208.176.1.8", Boolean.toString(livingWillRegistrationHolder.isHasRegistration()), "Sundhedsdatastyrelsen"));
        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.31", "2019-08-14", null),
                        createTS(null),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.8", "NI", "Sundhedsdatastyrelsen"),
                                createOrganization(createON("Sundhedsdatastyrelsen")),
                                null)));

        return observation;
    }

    public static POCDMT000040Observation createLanguageTypedInObservation(LanguageTypedInV3 languageTypedIn) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII(spokenLanguageRootId, "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", languageTypedIn.getId().getExtension(), null));
        observation.setCode(createCD("LanguageTypedIn", "1.2.208.184.100.1", "Talt sprog, indtastet", "MedCom Message Codes"));
        observation.getValues().add(createCD(languageTypedIn.getLanguageCode(), "1.0.639.1", languageTypedIn.getLanguageName(), "ISO-639-1"));

        AssignedAuthor aa = languageTypedIn.getAssignedAuthor();
        DataEnterer de = aa.dataEnterer;
        boolean isHealthcareProfessional = de == DataEnterer.EnteredHealthcareprofessional;
        POCDMT000040Organization representedOrg = isHealthcareProfessional
                ? createOrganization(createON(aa.representedOrganization.get().value))
                : null;
        POCDMT000040Person assignedPerson = buildAssignedPerson(languageTypedIn.getAuthorGivens(), languageTypedIn.getAuthorFamily(), isHealthcareProfessional);

        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(languageTypedIn.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                createCE(de.code, "1.2.208.184.100.1", de.displayName, "MedCom Message Codes"),
                                representedOrg,
                                assignedPerson)));
        return observation;
    }

    public static POCDMT000040Observation createTempAddrTypedInObservation(TempAddrTypedInV3 tempAddrTypedIn) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII(temporaryAddressRootId, "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", tempAddrTypedIn.getId().getExtension(), null));
        observation.setCode(createCD("TempAddrTypedIn", "1.2.208.184.100.1", "Midlertidig adresse, indtastet", "MedCom Message Codes"));
        // PeriodFrom is mandatory. PeriodTo is optional
        if(tempAddrTypedIn.getPeriodTo() != null) {
            observation.setEffectiveTime(createIVLTSForAddress(
                    tempAddrTypedIn.getPeriodFrom().atStartOfDay(ZoneId.systemDefault()),
                    tempAddrTypedIn.getPeriodTo().atStartOfDay(ZoneId.systemDefault())));
        }
        else {
            observation.setEffectiveTime(createIVLTSForAddress(
                    tempAddrTypedIn.getPeriodFrom().atStartOfDay(ZoneId.systemDefault()), null));
        }
        AddressHolder address = tempAddrTypedIn.getAddress();
        observation.getValues().add(createAD(address.getUse(), address.getAddressLines(), address.getPostalCode(), address.getCity(), address.getCountry()));

        AssignedAuthor aa = tempAddrTypedIn.getAssignedAuthor();
        DataEnterer de = aa.dataEnterer;
        boolean isHealthcareProfessional = de == DataEnterer.EnteredHealthcareprofessional;
        POCDMT000040Organization representedOrg = isHealthcareProfessional
                ? createOrganization(createON(aa.representedOrganization.get().value))
                : null;
        POCDMT000040Person assignedPerson = buildAssignedPerson(tempAddrTypedIn.getAuthorGivens(), tempAddrTypedIn.getAuthorFamily(), isHealthcareProfessional);

        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(tempAddrTypedIn.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                createCE(de.code, "1.2.208.184.100.1", de.displayName, "MedCom Message Codes"),
                                representedOrg,
                                assignedPerson)));

        return observation;
    }

    public static POCDMT000040Observation createPatientContactTypedInObservation(PatientContactTypedInV3 patientContactTypedIn) {
        List<PhoneHolder> phones = patientContactTypedIn.getPhones();
        List<PhoneHolder> emailAddresses = patientContactTypedIn.getEmailAddresses();

        failIfWrongSize(phones, 0, 3);
        failIfWrongSize(emailAddresses, 0, 1);

        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.24", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", UUID.randomUUID().toString(), null));
        observation.setCode(createCD("PatientContactTypedIn", "1.2.208.184.100.1", "Kontaktoplysninger, indtastet", "MedCom Message Codes"));
        observation.getValues().addAll(
                phones
                        .stream()
                        .map(value -> PDC_3_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                        .collect(Collectors.toList()));
        observation.getValues().addAll(
                emailAddresses
                        .stream()
                        .map(value -> PDC_3_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                        .collect(Collectors.toList()));

        AssignedAuthor aa = patientContactTypedIn.getAssignedAuthor();
        DataEnterer de = aa.dataEnterer;
        boolean isHealthcareProfessional = de == DataEnterer.EnteredHealthcareprofessional;
        POCDMT000040Organization representedOrg = isHealthcareProfessional
                ? createOrganization(createON(aa.representedOrganization.get().value))
                : null;
        POCDMT000040Person assignedPerson = buildAssignedPerson(patientContactTypedIn.getAuthorGivens(), patientContactTypedIn.getAuthorFamily(), isHealthcareProfessional);

        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(patientContactTypedIn.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                createCE(de.code, "1.2.208.184.100.1", de.displayName, "MedCom Message Codes"),
                                representedOrg,
                                assignedPerson)));
        return observation;
    }

    public static POCDMT000040Observation createRelativeTypedInObservation(RelativeTypedInV3 relativeTypedIn) {
        POCDMT000040Observation observation = createObservation();

        observation.getTemplateIds().add(createII("1.2.208.184.16.1.10.20.1.25", "2019-08-14", null));
        observation.getIds().add(createII("1.2.208.184", relativeTypedIn.getId().getExtension(), null));
        observation.setCode(createCD("RelativeTypedIn", "1.2.208.184.100.1", "Pårørende, indtastet", "MedCom Message Codes"));
        // Skal acceptere manglende efternavn, selvom det afviger fra MedCom standarden, iht. [SDS-5182]
        observation.getValues().add(createPN(relativeTypedIn.getGivens(), relativeTypedIn.getFamily() != null ? relativeTypedIn.getFamily() : ""));
        observation.getValues().add(createCD(relativeTypedIn.getRelation().getCode(), "1.2.208.184.100.2", relativeTypedIn.getRelation().getDisplayName(), "MedCom Relation Codes"));
        List<TEL> workPhones = relativeTypedIn.getPhoneHolders()
                .stream()
                .map(value -> PDC_3_0ElementsMapper.createTEL(value.getUses(), value.getValue()))
                .collect(Collectors.toList());
        observation.getValues().addAll(workPhones);
        if (relativeTypedIn.getNote() != null) {
            observation.getValues().add(createST(relativeTypedIn.getNote()));
        }

        AssignedAuthor aa = relativeTypedIn.getAssignedAuthor();
        DataEnterer de = aa.dataEnterer;
        boolean isHealthcareProfessional = de == DataEnterer.EnteredHealthcareprofessional;
        POCDMT000040Organization representedOrg = isHealthcareProfessional
                ? createOrganization(createON(aa.representedOrganization.get().value))
                : null;
        POCDMT000040Person assignedPerson = buildAssignedPerson(relativeTypedIn.getAuthorGivens(), relativeTypedIn.getAuthorFamily(), isHealthcareProfessional);

        observation.getAuthors().add(
                createAuthor(
                        createII("1.2.208.184.16.1.10.20.30", "2019-08-14", null),
                        createTS(relativeTypedIn.getAuthorTime()),
                        createAssignedAuthor(
                                createII("1.2.208.176.1.2", "ANONYM", "CPR"), // Use "ANONYM" to avoid exposing author CPR
                                createCE(de.code, "1.2.208.184.100.1", de.displayName, "MedCom Message Codes"),
                                representedOrg,
                                assignedPerson)));

        return observation;
    }

    private static POCDMT000040Person buildAssignedPerson(List<String> givens, String family, boolean isHealthcareProfessional) {
        POCDMT000040Person assignedPerson = null;
        boolean hasAuthorName = !(givens == null || family == null || givens.isEmpty() || family.isEmpty());
        if (isHealthcareProfessional) {
            // Det er optional at angive navn ved sundhedsfaglig
            assignedPerson = hasAuthorName
                    ? createAssignedPerson(createPN(givens, family))
                    : null;
        } else {
            assignedPerson = createAssignedPerson(createPN(givens, family));
        }
        return assignedPerson;
    }
}