/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler.mapper;

import org.hl7.v3.*;

import javax.xml.bind.JAXBElement;
import java.io.Serializable;
import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Optional;

import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfEmpty;
import static dk.stamkort.webservice.hl7cda.assembler.SourceDataValidator.failIfNull;

public class PDC_2_0ElementsMapper {

    public static final ObjectFactory o = new ObjectFactory();

    public static final DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ"); // Output sample: 20171019144200+0200
    public static final DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd"); // Output sample: 20171019

    public static StrucDocText createStrucDocText(List<Serializable> textContentElements) {
        StrucDocText text = o.createStrucDocText();
        text.getContent().addAll(textContentElements);
        return text;
    }

    public static POCDMT000040Patient createPatient(PN personName, CE administrativeGenderCode, TS birthTime) {
        POCDMT000040Patient patient = o.createPOCDMT000040Patient();

        // classCode attribute
        patient.getClassCodes().add("PSN");

        // name
        patient.getNames().add(personName);

        // administrativeGenderCode
        Optional.ofNullable(administrativeGenderCode).ifPresent(patient::setAdministrativeGenderCode);

        // birthTime
        Optional.ofNullable(birthTime).ifPresent(patient::setBirthTime);

        return patient;
    }

    public static POCDMT000040Section createSection(II templateId, StrucDocText text, List<POCDMT000040Entry> entries) {
        POCDMT000040Section section = o.createPOCDMT000040Section();
        section.getClassCodes().add("DOCSECT"); // attribute
        section.getMoodCodes().add("EVN"); // attribute
        section.getTemplateIds().add(templateId);
        section.setText(text);
        entries.forEach(value ->
                section.getEntries().add(value));
        return section;
    }

    public static POCDMT000040Component2 createComponent2(POCDMT000040StructuredBody structuredBody) {
        POCDMT000040Component2 component2 = o.createPOCDMT000040Component2();
        component2.setTypeCode(ActRelationshipHasComponent.COMP); // attribute
        component2.setContextConductionInd(Boolean.TRUE); // attribute
        component2.setStructuredBody(structuredBody);
        return component2;
    }

    public static POCDMT000040Component3 createComponent3(POCDMT000040Section section) {
        POCDMT000040Component3 component3 = o.createPOCDMT000040Component3();
        component3.setTypeCode(ActRelationshipHasComponent.COMP); // attribute
        component3.setContextConductionInd(Boolean.TRUE); // attribute
        component3.setSection(section);
        return component3;
    }

    public static POCDMT000040StructuredBody createStructuredBody(POCDMT000040Component3 component3) {
        POCDMT000040StructuredBody structuredBody = o.createPOCDMT000040StructuredBody();
        structuredBody.getClassCodes().add("DOCBODY"); // attribute
        structuredBody.getMoodCodes().add("EVN"); // attribute
        structuredBody.getComponents().add(component3);
        return structuredBody;
    }

    public static POCDMT000040Entry createEntry(POCDMT000040Observation observation) {
        POCDMT000040Entry entry = o.createPOCDMT000040Entry();
        entry.setTypeCode(XActRelationshipEntry.COMP);
        entry.setObservation(observation);
        return entry;
    }

    public static II createII(String root, String extension, String assigningAuthorityName) {
        failIfNull(root);

        II id = o.createII();
        id.setRoot(root);
        id.setExtension(extension);
        id.setAssigningAuthorityName(assigningAuthorityName);
        return id;
    }

    public static AD createAD(String use, List<String> addressLines, String postalCode, String city, String country) {
        AD address = o.createAD();

        // use attribute
        address.getUses().add(use);

        // streetAddressLine
        if(addressLines != null) {
            addressLines.forEach(v -> {
                AdxpStreetAddressLine streetAddressLineType = o.createAdxpStreetAddressLine();
                streetAddressLineType.getContent().add(v);
                address.getContent().add(o.createADStreetAddressLine(streetAddressLineType));
            });
        }

        // postalCode
        if(postalCode != null) {
            AdxpPostalCode postalCodeType = o.createAdxpPostalCode();
            postalCodeType.getContent().add(postalCode);
            address.getContent().add(o.createADPostalCode(postalCodeType));
        }

        // city
        if(city != null) {
            AdxpCity cityType = o.createAdxpCity();
            cityType.getContent().add(city);
            address.getContent().add(o.createADCity(cityType));
        }

        // country
        if (country != null) {
            AdxpCountry countryType = o.createAdxpCountry();
            countryType.getContent().add(country);
            address.getContent().add(o.createADCountry(countryType));
        }

        return address;
    }

    public static POCDMT000040Organization createProviderOrganization(II id, ON name, List<TEL> telecoms, AD addrs) {
        POCDMT000040Organization providerOrganization = o.createPOCDMT000040Organization();

        // id
        Optional.ofNullable(id).ifPresent(value ->
                providerOrganization.getIds().add(value));

        // name
        Optional.ofNullable(name).ifPresent(value ->
                providerOrganization.getNames().add(value));

        // telecom
        providerOrganization.getTelecoms().addAll(telecoms);

        // addr
        Optional.ofNullable(addrs).ifPresent(value ->
                providerOrganization.getAddrs().add(value));

        return providerOrganization;
    }

    public static CS createCS(String code) {
        failIfNull(code);

        CS cs = o.createCS();
        cs.setCode(code);
        return cs;
    }

    public static CE createCE(String code, String codeSystem, String displayName, String codeSystemName) {
        failIfNull(codeSystem);

        CE ce = o.createCE();
        ce.setDisplayName(displayName);
        ce.setCodeSystem(codeSystem);
        ce.setCode(code);
        ce.setCodeSystemName(codeSystemName);
        return ce;
    }

    public static CD createCD(String code, String codeSystem, String displayName, String codeSystemName) {
        failIfNull(codeSystem);

        CD cd = o.createCD();
        cd.setDisplayName(displayName);
        cd.setCodeSystem(codeSystem);
        cd.setCode(code);
        cd.setCodeSystemName(codeSystemName);
        return cd;
    }

    public static EN createEN(String prefix, List<String> givens, String family) {
        failIfEmpty(givens);
        failIfNull(family);

        EN name = o.createEN();

        // prefix
        Optional.ofNullable(prefix).ifPresent(value -> {
            EnPrefix enPrefix = o.createEnPrefix();
            enPrefix.getContent().add(value);
            JAXBElement<EnPrefix> enPrefixType = o.createENPrefix(enPrefix);
            name.getContent().add(enPrefixType);
        });

        // given
        givens.forEach(value -> {
            EnGiven enGiven = o.createEnGiven();
            enGiven.getContent().add(value);
            JAXBElement<EnGiven> enGivenType = o.createENGiven(enGiven);
            name.getContent().add(enGivenType);
        });

        // family
        EnFamily enFamily = o.createEnFamily();
        enFamily.getContent().add(family);
        JAXBElement<EnFamily> enFamilyType = o.createENFamily(enFamily);
        name.getContent().add(enFamilyType);

        return name;
    }

    public static POCDMT000040InfrastructureRootTypeId createTypeId(String root, String extension) {
        failIfNull(root);
        failIfNull(extension);

        POCDMT000040InfrastructureRootTypeId typeId = o.createPOCDMT000040InfrastructureRootTypeId();
        typeId.setRoot(root);
        typeId.setExtension(extension);
        return typeId;
    }

    public static ST createST(String content) {
        failIfNull(content);

        ST st = o.createST();
        st.getContent().add(content);
        return st;
    }

    public static TS createTS(ZonedDateTime zonedDateTime) {
        TS ts = o.createTS();

        if (zonedDateTime == null) {
            ts.getNullFlavors().add("NI");
        } else {
            ts.setValue(dateTimeFormatter.format(zonedDateTime));
        }

        return ts;
    }

    public static TS createDateTS(LocalDate zonedDateTime) {
        TS ts = o.createTS();

        if (zonedDateTime == null) {
            ts.getNullFlavors().add("NI");
        } else {
            ts.setValue(dateFormatter.format(zonedDateTime));
        }

        return ts;
    }

    public static TEL createTEL(List<String> uses, String value) {
        failIfNull(uses);
        failIfNull(value);

        TEL tel = o.createTEL();
        tel.getUses().addAll(uses);
        tel.setValue(FormatHelper.formatPhoneNumber(value));
        return tel;
    }

    public static TEL createTELHandleNull(List<String> uses, String value) {
        failIfNull(uses);

        TEL tel = o.createTEL();
        tel.getUses().addAll(uses);
        if (value != null && !value.trim().isEmpty()) {
            tel.setValue(FormatHelper.formatPhoneNumber(value));
        } else {
            tel.getNullFlavors().add("NI");
        }
        return tel;
    }

    public static POCDMT000040Author createAuthor(II templateId, TS time, POCDMT000040AssignedAuthor assignedAuthor) {
        POCDMT000040Author author = o.createPOCDMT000040Author();

        Optional.ofNullable(templateId).ifPresent(value ->
                author.getTemplateIds().add(value));
        author.setTime(time);
        author.setAssignedAuthor(assignedAuthor);
        return author;
    }

    public static POCDMT000040AssignedAuthor createAssignedAuthor(II id, POCDMT000040Organization representedOrganization, POCDMT000040Person assignedPerson) {
        POCDMT000040AssignedAuthor assignedAuthor = o.createPOCDMT000040AssignedAuthor();

        assignedAuthor.getIds().add(id);
        assignedAuthor.setRepresentedOrganization(representedOrganization);
        assignedAuthor.setAssignedPerson(assignedPerson);

        return assignedAuthor;
    }

    public static POCDMT000040Organization createOrganization(ON name) {
        POCDMT000040Organization organization = o.createPOCDMT000040Organization();

        organization.getNames().add(name);

        return organization;
    }

    public static ON createON(String content) {
        failIfNull(content);

        ON on = o.createON();
        on.getContent().add(content);

        return on;
    }

    public static POCDMT000040Person createAssignedPerson(PN name) {
        POCDMT000040Person assignedPerson = o.createPOCDMT000040Person();

        assignedPerson.getNames().add(name);

        return assignedPerson;
    }

    public static POCDMT000040Custodian createCustodian(POCDMT000040AssignedCustodian assignedCustodian) {
        POCDMT000040Custodian custodian = o.createPOCDMT000040Custodian();

        custodian.setAssignedCustodian(assignedCustodian);

        return custodian;
    }

    public static POCDMT000040AssignedCustodian createAssignedCustodian(POCDMT000040CustodianOrganization representedCustodianOrganization) {
        POCDMT000040AssignedCustodian assignedCustodian = o.createPOCDMT000040AssignedCustodian();

        assignedCustodian.setRepresentedCustodianOrganization(representedCustodianOrganization);

        return assignedCustodian;
    }

    public static POCDMT000040CustodianOrganization createCustodianOrganization(II id, ON name) {
        POCDMT000040CustodianOrganization custodianOrganization = o.createPOCDMT000040CustodianOrganization();

        custodianOrganization.getIds().add(id);
        custodianOrganization.setName(name);

        return custodianOrganization;
    }

    public static POCDMT000040DocumentationOf createDocumentationOf(POCDMT000040ServiceEvent serviceEvent) {
        POCDMT000040DocumentationOf documentationOf = o.createPOCDMT000040DocumentationOf();
        documentationOf.getTypeCodes().add("DOC");
        documentationOf.setServiceEvent(serviceEvent);

        return documentationOf;
    }

    public static POCDMT000040ServiceEvent createServiceEvent(II templateId, II id, IVLTS effectiveTime) {
        POCDMT000040ServiceEvent serviceEvent = o.createPOCDMT000040ServiceEvent();
        serviceEvent.getClassCodes().add("ACT");
        serviceEvent.getMoodCodes().add("EVN");

        // templateId
        Optional.ofNullable(templateId).ifPresent(value ->
                serviceEvent.getTemplateIds().add(value));

        // id
        Optional.ofNullable(id).ifPresent(value ->
                serviceEvent.getIds().add(value));

        // effectiveTime
        serviceEvent.setEffectiveTime(effectiveTime);

        return serviceEvent;
    }

    public static IVLTS createIVLTS(ZonedDateTime low, ZonedDateTime high) {
        IVXBTS lowType = o.createIVXBTS();
        if (low == null) {
            lowType.getNullFlavors().add("NA");
        } else {
            lowType.setValue(dateTimeFormatter.format(low));
        }

        IVXBTS highType = o.createIVXBTS();
        if (high == null) {
            highType.getNullFlavors().add("NA");
        } else {
            highType.setValue(dateTimeFormatter.format(high));
        }

        IVLTS ivlts = o.createIVLTS();

        ivlts.setLow(lowType);
        ivlts.setHigh(highType);

        return ivlts;
    }

    public static PN createPN(List<String> givens, String family) {
        failIfEmpty(givens);
        failIfNull(family);

        PN name = o.createPN();

        // given
        givens.forEach(value -> {
            EnGiven enGiven = o.createEnGiven();
            enGiven.getContent().add(value);
            JAXBElement<EnGiven> enGivenType = o.createENGiven(enGiven);
            name.getContent().add(enGivenType);
        });

        // family
        EnFamily enFamily = o.createEnFamily();
        enFamily.getContent().add(family);
        JAXBElement<EnFamily> enFamilyType = o.createENFamily(enFamily);
        name.getContent().add(enFamilyType);

        return name;
    }

    public static TN createTN(String orgName) {
        TN tn = new TN(); tn.getContent().add(orgName); return tn;
    }
}
