/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.sds.nsp.security.SecurityContext.Organisation;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.OrganizationTypeVO;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.integrations.minlog.MinLogDataWrapper;
import dk.stamkort.integrations.minlog.MinLogException;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.*;
import dk.stamkort.webservice.hl7cda.ClinicalDocumentProvider;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType;
import org.apache.commons.io.Charsets;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.xml.bind.JAXBException;
import java.util.Optional;

@Component
public class DocumentProvider {
    private static final Logger log = LogManager.getLogger(DocumentProvider.class);
    private static final String SERVICE_IDENTIFIER = "Det Fælles Stamkort";

    @Autowired
    public PersonalDataCardServiceImpl personalDataCardService;

    @Autowired
    public RegistryIndexDAO registryIndexDAO;

    @Autowired
    private MinLogDataWrapper minLogDataWrapper;

    @Autowired
    private Properties properties;

    @Autowired
    private ClinicalDocumentProvider clinicalDocumentProvider;

    public Optional<DocumentResponse> getDocument(RequestContextVO requestContextVO,
                                                  RetrieveDocumentSetRequestType.DocumentRequest documentRequest)
            throws JAXBException, DAOException, FSKException {

        RetrieveDocumentSetResponseType.DocumentResponse documentResponse = new RetrieveDocumentSetResponseType.DocumentResponse();
        documentResponse.setRepositoryUniqueId(documentRequest.getRepositoryUniqueId());
        documentResponse.setDocumentUniqueId(documentRequest.getDocumentUniqueId());
        documentResponse.setHomeCommunityId(documentRequest.getHomeCommunityId());
        documentResponse.setMimeType("text/xml");

        String documentUniqueId = documentRequest.getDocumentUniqueId();
        RegistryIndexVO registryIndex = registryIndexDAO.getRegistryIndexByDocumentId(documentUniqueId);

        if (registryIndex == null) { return Optional.empty(); }

        if(!matchPatientId(requestContextVO, registryIndex.getPersonIdentifier())) {
            log.info("PatientContext does not match personId on document.");
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, "PatientId i sikkerhedsbilletten matcher ikke patientId i requestet.");
        }

        String xml = clinicalDocumentProvider.getClinicalDocument(registryIndex.getPersonIdentifier(),
                                                                  requestContextVO,
                                                                  registryIndex.getFormatCode());
        documentResponse.setDocument(xml.getBytes(Charsets.UTF_8));

        if (validateSkipMinLog(requestContextVO, registryIndex.getPersonIdentifier())) {
            logReadActivityToMinLog(requestContextVO, registryIndex.getPersonIdentifier());
        }

        return Optional.of(new DocumentResponse(documentResponse, registryIndex.getPersonIdentifier()));
    }

    private boolean matchPatientId(RequestContextVO requestContextVO, String personIdentifier) {
        String patientInContext = requestContextVO.getPatientInContext();
        if(patientInContext != null) {
            return patientInContext.equals(personIdentifier);
        }
        return true;
    }

    private void logReadActivityToMinLog(RequestContextVO context, String cprnr) {
        try {
            if (context != null) {
                minLogDataWrapper.logReadActivity(context, properties.getMinLogReadActivityText(), cprnr,
                        (Optional.ofNullable(context.getOrganisation()).map(Organisation::getIdentifier).orElse(null)),
                        (Optional.ofNullable(context.getOrganisation()).flatMap(Organisation::getName).orElse(null)),
                        getOrgType(context.getOrganisation()),
                        context.getUserCprIdentifier(), null, SERVICE_IDENTIFIER);
            }
        } catch (MinLogException e) {
            log.error("Error writing to MinLog.", e);
        }
    }

    private OrganizationTypeVO getOrgType(Organisation organisation) {
        if (organisation != null) {
            switch (organisation.getIdentifierFormat()) {
                case CVR:
                    return OrganizationTypeVO.CVRNUMBER;
                default:
                    break;
            }
        }
        return null;
    }

    private boolean validateSkipMinLog(RequestContextVO requestContextVO, String cprCitizen) {
        if (requestContextVO.getUserType().equals(RequestContextVO.UserType.Citizen) && requestContextVO.getUserCprIdentifier().equals(cprCitizen)) {
            return false;
        }

        if (requestContextVO.getUserType().equals(RequestContextVO.UserType.SystemUser)) {
            return false;
        }

        return true;
    }
}
