/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.auditlog;

import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.security.SecurityContext;
import dk.stamkort.util.RequestContextVO;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Optional;

public class AuditLoggerTest {
    private static final String COMPONENT = "FSK";

    private AuditBuilder auditBuilder;
    private AuditLogger auditLogger;

    @Before
    public void setUp() throws Exception {
        auditBuilder = Mockito.mock(AuditBuilder.class);
        auditLogger = new AuditLogger(() -> auditBuilder);
    }

    @Test
    public void logRetrieveDocumentSet_WithDocumentsAndPatientIds_LogsDocumentIdsPatientIdsAndServiceActor() {
        String document1 = "doc1";
        String document2 = "doc2";
        HashSet<String> documents = new HashSet<>(Arrays.asList(document1, document2));
        String patientId1 = "0000000000";
        String patientId2 = "1111111111";
        HashSet<String> patientIds = new HashSet<>(Arrays.asList(patientId1, patientId2));
        RequestContextVO requestContext = createRequestContext();

        auditLogger.logRetrieveDocumentSet(documents, patientIds, requestContext);

        String context = "retrieveDocumentSet";
        Mockito.verify(auditBuilder).addAuditInformation(COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "document-id", document1);
        Mockito.verify(auditBuilder).addAuditInformation(COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "document-id", document2);
        Mockito.verify(auditBuilder).addAuditInformation(COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr", patientId1);
        Mockito.verify(auditBuilder).addAuditInformation(COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr", patientId2);
        verifyServiceActorAuditLogged(requestContext, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logRetrieveDocumentSet_WithoutDocuments_DoesNotLog() {
        HashSet<String> documents = new HashSet<>();
        HashSet<String> patientIds = new HashSet<>(Collections.singletonList("0000000000"));
        RequestContextVO requestContext = createRequestContext();

        auditLogger.logRetrieveDocumentSet(documents, patientIds, requestContext);

        Mockito.verifyZeroInteractions(auditBuilder);
    }

    private void verifyServiceActorAuditLogged(RequestContextVO requestContext, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "actor-cpr", requestContext.getUserCprIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-usertype", requestContext.getUserType());
    }

    private RequestContextVO createRequestContext() {
        RequestContextVO requestContext = new RequestContextVO();
        requestContext.setUserCprIdentifier("1234567890");
        requestContext.setGivenName("Jens");
        requestContext.setSurName("Jensen");
        requestContext.setUserType(RequestContextVO.UserType.Citizen);
        requestContext.setAuthorizationCode("QWERTY");
        requestContext.setEducationCode("1234");
        requestContext.setOrganisation(createOrganisation());
        return requestContext;
    }

    private static SecurityContext.Organisation createOrganisation() {
        return new SecurityContext.Organisation() {
            @Override
            public OrganisationIdentifierFormat getIdentifierFormat() {
                return OrganisationIdentifierFormat.CVR;
            }

            @Override
            public String getIdentifier() {
                return "9999999999";
            }

            @Override
            public Optional<String> getName() {
                return Optional.of("Test Org");
            }
        };
    }
}