/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.conf;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Optional;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.sql.DataSource;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import org.flywaydb.core.Flyway;
import org.junit.ClassRule;
import org.junit.rules.TemporaryFolder;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.support.GenericApplicationContext;
import org.springframework.jdbc.datasource.DataSourceTransactionManager;
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabase;
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseBuilder;
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseType;
import org.springframework.transaction.PlatformTransactionManager;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.ws.SLALoggingServletFilter;
import dk.stamkort.util.RequestContextEnricher;

@Configuration
@PropertySource("application.properties")
public class TestDAOConfig {

	@Autowired
	GenericApplicationContext applicationContext;
	
	@Bean
	public WebServiceContext createMockWebServiceContext() {

		WebServiceContext mock = Mockito.mock(WebServiceContext.class);
		MessageContext messageContext = Mockito.mock(MessageContext.class);
		ServletContext servletContext = Mockito.mock(ServletContext.class);
		Mockito.when(servletContext.getAttribute(Mockito.anyString())).thenReturn(applicationContext);
		
		HttpServletRequest httpServletRequest = Mockito.mock(HttpServletRequest.class);
		SLALogItem slaLogItem = Mockito.mock(SLALogItem.class);
		Mockito.when(httpServletRequest.getAttribute(Mockito.eq(SLALoggingServletFilter.CURRENT_LOGITEM_REQUEST_KEY))).thenReturn(slaLogItem);
		
		Mockito.when(messageContext.get(Mockito.eq(MessageContext.SERVLET_REQUEST))).thenReturn(httpServletRequest);
		Mockito.when(messageContext.get(Mockito.eq(MessageContext.SERVLET_CONTEXT))).thenReturn(servletContext);
		
		Mockito.when(mock.getMessageContext()).thenReturn(messageContext);

		return mock;
	}
	
    @ClassRule
    private TemporaryFolder tempDBFolder = new TemporaryFolder();
    TestDBScriptPreprocessor scriptPreprocessor = new TestDBScriptPreprocessor();

    private final Flyway flywayFSK = new Flyway();
    private final Flyway flywayStamdata = new Flyway();

    @Bean
    @Primary // make primary as we do not need transactions for other datasources
    public PlatformTransactionManager fskDataSourceTxManager(DataSource fskDataSource) {
        return new DataSourceTransactionManager(fskDataSource);
    }

	@Bean(name="fsk")
    @Primary
    public DataSource fskDataSource() {
        DataSource datasource = createEmbeddedDatabase("fsk");
        File scriptsLocation = prepareScripts("../compose/database/db/migration/fsk", "../compose/database/db/test/fsk");

        flywayFSK.setDataSource(datasource);
        flywayFSK.setLocations("filesystem:/" + scriptsLocation.getPath());
        flywayFSK.migrate();
        return datasource;
    }

	@Bean(name="stm")
    public DataSource stamdataDataSource() {
        DataSource datasource = createEmbeddedDatabase("stm");
        File scriptsLocation = prepareScripts("../compose/database/db/migration/stm", "../compose/database/db/test/stm");

        flywayStamdata.setDataSource(datasource);
        flywayStamdata.setLocations("filesystem:/" + scriptsLocation.getPath());
        flywayStamdata.migrate();
        return datasource;
    }

    @Bean
    public PlatformTransactionManager stamdataDataSourceTxManager(@Qualifier("stm") DataSource stamdataDataSource) {
        return new DataSourceTransactionManager(stamdataDataSource);
    }


    
    private EmbeddedDatabase createEmbeddedDatabase(String name) {
        return new EmbeddedDatabaseBuilder()
                .setType(EmbeddedDatabaseType.H2)
                .setName(name + ";DB_CLOSE_DELAY=-1;MODE=MySQL")
                .build();
    }

    
    
    public Flyway getFlywayFSK() {
        return flywayFSK;
    }

    public Flyway getFlywayStamdata() {
        return flywayStamdata;
    }


    private File prepareScripts(String... paths) {
        try {
            tempDBFolder.create();
            File subFolder = tempDBFolder.newFolder();
            for (String path : paths) {
                scriptPreprocessor.preprocessScripts(path, subFolder);
            }
            return subFolder;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }
    
    @Bean
    public RequestContextEnricher requestContextEnricher() {
    	return new RequestContextEnricher() {
			
			@Override
			public Optional<String> createRequestWithContext(Element request) {
				try {
			        MessageFactory messageFactory = MessageFactory.newInstance();
			        SOAPMessage soapMessage = messageFactory.createMessage();
			        SOAPEnvelope soapEnvelope = soapMessage.getSOAPPart().getEnvelope();
			        SOAPBody soapBody = soapEnvelope.getBody();
			        Node imported = soapBody.getOwnerDocument().importNode(request, true);
			        soapBody.appendChild(imported);

			        // get xml as string with correct encoding
			        StringWriter sw = new StringWriter();
			        TransformerFactory.newInstance().newTransformer().transform(new DOMSource(soapMessage.getSOAPPart()), new StreamResult(sw));
			        String requestXml = sw.toString();
			        
			    	return (Optional.of(requestXml));
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
			}
		};
    }
}
