/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.context;

import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityProvider;
import org.w3c.dom.Element;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicReference;

public class FskDevelopmentSecurityProvider extends SecurityProvider implements SecurityContext {

    private static AtomicReference<Ticket> globalTicket = new AtomicReference(new DevelopmentTicket());
    private static AtomicReference<Message> globalMessage = new AtomicReference(new DevelopmentMessage());
    private static AtomicReference<User> globalActingUser = new AtomicReference(new DevelopmentUser());
    private static AtomicReference<User> globalPrincipalUser = new AtomicReference(new DevelopmentUser());
    private static AtomicReference<Organisation> globalOrganisation = new AtomicReference(new DevelopmentOrganisation());
    private static AtomicReference<Client> globalClient = new AtomicReference(new DevelopmentClient());
    private static ThreadLocal<Ticket> threadTicket = new ThreadLocal();
    private static ThreadLocal<Message> threadMessage = new ThreadLocal();
    private static ThreadLocal<User> threadActingUser = new ThreadLocal();
    private static ThreadLocal<User> threadPrincipalUser = new ThreadLocal();
    private static ThreadLocal<Organisation> threadOrganisation = new ThreadLocal();
    private static ThreadLocal<Client> threadClient = new ThreadLocal();

    public SecurityContext getSecurityContext() {
        System.err.println("WARNING: This NSP Security Provider must not be used in test or production environments. It is only for development");
        return this;
    }

    public Optional<Ticket> getTicket() {
        Ticket ticket = (Ticket)threadTicket.get();
        if (ticket == null) {
            ticket = (Ticket)globalTicket.get();
        }

        return Optional.of(ticket);
    }

    public void setGlobalTicket(Ticket ticket) {
        globalTicket.set(ticket);
    }

    public void setThreadTicket(Ticket ticket) {
        threadTicket.set(ticket);
    }

    public Optional<Message> getMessage() {
        Message message = (Message)threadMessage.get();
        if (message == null) {
            message = (Message)globalMessage.get();
        }

        return Optional.of(message);
    }

    public void setGlobalMessage(Message message) {
        globalMessage.set(message);
    }

    public void setThreadMessage(Message message) {
        threadMessage.set(message);
    }

    public Optional<User> getActingUser() {
        return Optional.empty();
    }

    public void setGlobalActingUser(User user) {
        globalActingUser.set(user);
    }

    public void setThreadActingUser(User user) {
        threadActingUser.set(user);
    }

    public Optional<User> getPrincipalUser() { return Optional.empty(); }

    public void setGlobalPrincipalUser(User user) {
        globalPrincipalUser.set(user);
    }

    public void setThreadPrincipalUser(User user) {
        threadPrincipalUser.set(user);
    }

    public Optional<Organisation> getOrganisation() {
        Organisation organisation = (Organisation)threadOrganisation.get();
        if (organisation == null) {
            organisation = (Organisation)globalOrganisation.get();
        }

        return Optional.of(organisation);
    }

    public void setGlobalOrganisation(Organisation organisation) {
        globalOrganisation.set(organisation);
    }

    public void setThreadOrganisation(Organisation organisation) {
        threadOrganisation.set(organisation);
    }

    public Optional<Client> getClient() {
        Client client = (Client)threadClient.get();
        if (client == null) {
            client = (Client)globalClient.get();
        }

        return Optional.of(client);
    }

    public void setGlobalClient(Client client) {
        globalClient.set(client);
    }

    public void setThreadClient(Client client) {
        threadClient.set(client);
    }

    private static final class DevelopmentClient implements Client {
        private DevelopmentClient() {
        }

        public Optional<String> getName() {
            return Optional.of("SYSTEM A");
        }

        public Optional<PersistentUniqueKey> getPersistentUniqueKey() {
            return Optional.of("CVR:46837428-UID:27910135"::toString);
        }
    }

    private static final class DevelopmentOrganisation implements Organisation {
        private DevelopmentOrganisation() {
        }

        public OrganisationIdentifierFormat getIdentifierFormat() {
            return OrganisationIdentifierFormat.CVR;
        }

        public String getIdentifier() {
            return "12345678";
        }

        public Optional<String> getName() {
            return Optional.of("Lægehuset");
        }
    }

    private static final class DevelopmentUser implements User {
        private DevelopmentUser() {
        }

        public UserType getUserType() {
            return UserType.HealthcareProfessional;
        }

        public PersonIdentifierFormat getIdentifierFormat() {
            return PersonIdentifierFormat.CPR;
        }

        public String getIdentifier() {
            return "0102732379";
        }

        public Optional<String> getGivenName() {
            return Optional.of("Karl Hoffmann");
        }

        public Optional<String> getSurname() {
            return Optional.of("Svendsen");
        }

        public Optional<Credentials> getCredentials() {
            return Optional.of(new DevelopmentCredentials());
        }

        public Optional<PersistentUniqueKey> getPersistentUniqueKey() {
            return Optional.of("CVR:25450442-RID:40252666"::toString);
        }

        private static final class DevelopmentCredentials implements Credentials {
            private DevelopmentCredentials() {
            }

            public Optional<String> getAuthorizationCode() {
                return Optional.of("NS362");
            }

            public Optional<String> getEducationCode() {
                return Optional.of("7170");
            }

            public Optional<String> getNationalRole() {
                return null;
            }

            public Optional<String> getUnverifiedRole() {
                return null;
            }

            public Set<String> getPowerOfAttorneyPrivileges() {
                return Collections.emptySet();
            }
        }
    }

    private static final class DevelopmentMessage implements Message {
        private DevelopmentMessage() {
        }

        public Optional<String> getIdentifier() {
            return Optional.of("bar");
        }

        public Optional<String> getConversationIdentifier() {
            return Optional.of("foo");
        }

        public Optional<String> getAction() {
            return Optional.of(UUID.randomUUID().toString());
        }
    }

    private static final class DevelopmentTicket implements Ticket {
        private final ZonedDateTime now;

        private DevelopmentTicket() {
            this.now = ZonedDateTime.now();
        }

        public boolean isValid() {
            return true;
        }

        public Federation getFederation() {
            return Federation.Development;
        }

        public Optional<String> getAudience() {
            return Optional.of("Http://nsp");
        }

        public Optional<ZonedDateTime> getCreated() {
            return Optional.of(this.now.minus(1L, ChronoUnit.MINUTES));
        }

        public Optional<ZonedDateTime> getValidFrom() {
            return Optional.of(this.now.minus(1L, ChronoUnit.MINUTES));
        }

        public Optional<ZonedDateTime> getValidTo() {
            return Optional.of(this.now.plus(1L, ChronoUnit.MINUTES));
        }

        public Optional<String> createRequestWithContext(Element requestBody, Element... headers) {
            return Optional.empty();
        }
    }
}
