/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.dao;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.time.Instant;
import java.util.List;
import java.util.UUID;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import dk.stamkort.conf.FskConfig;
import dk.stamkort.conf.TestDAOConfig;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.RegistryIndexVO;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { TestDAOConfig.class, FskConfig.class })
@Transactional
public class RegistryIndexDAOTests {
    private static final String TYPE_CODE = "testTypeCode";
    private static final String FORMAT_CODE = "testFormatCode";
    private static final String CPRNR = "testIdent";
    private static final String FIRST_NAME = "testFirstName";
    private static final String LAST_NAME = "testLastName";
    private static final String DOCUMENT_ID = UUID.randomUUID().toString();

    @Autowired
    private RegistryIndexDAO dao;

    @Test
    public void testCreateAndDeleteRegistryIndex() throws DAOException {
        Integer id = dao.createRegistryIndexVO(createVO());
        assertNotNull(id);

        RegistryIndexVO actual = dao.getRegistryIndexVO(CPRNR);
        assertNotNull(actual);
        assertNotNull(actual.getRegistryIndexPid());
        assertEquals(CPRNR, actual.getPersonIdentifier());
        assertEquals(FIRST_NAME, actual.getFirstName());
        assertEquals(LAST_NAME, actual.getLastName());
        assertEquals(DOCUMENT_ID, actual.getFSKDocumentId());
        assertNull(actual.getDeleteAfter());

        dao.deleteByDocumentId(actual.getFSKDocumentId());
        actual = dao.getRegistryIndexVO(CPRNR);
        assertNull(actual);
    }
 
    @Test
    public void testDeleteAfterRegistryIndex() throws DAOException {
    	
        Integer id = dao.createRegistryIndexVO(createVO());
        assertNotNull(id);
      
  	    Instant aFewMomentsAgo = Instant.ofEpochMilli(System.currentTimeMillis() - 100000L);

        dao.markForDeletion(CPRNR, aFewMomentsAgo); // a lille while ago

        RegistryIndexVO actual = dao.getRegistryIndexVO(CPRNR);
        assertNotNull(actual);
        assertNotNull(actual.getDeleteAfter());
      
        List<RegistryIndexVO> list = dao.getFSKDocumentIdsToBeDeleted(Instant.now());
        assertNotNull(list);
        assertEquals(1, list.size());
        assertEquals(id, list.get(0).getRegistryIndexPid());
        
    }

    private RegistryIndexVO createVO() {
        RegistryIndexVO data = new RegistryIndexVO();
        data.setTypeCode(TYPE_CODE);
        data.setFormatCode(FORMAT_CODE);
        data.setFirstName(FIRST_NAME);
        data.setLastName(LAST_NAME);
        data.setPersonIdentifier(CPRNR);
        data.setFSKDocumentId(DOCUMENT_ID);
        return data;
    }


}
