/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.vault.CredentialPair;
import dk.sosi.seal.vault.CredentialVault;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.w3c.dom.Document;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.sql.Date;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Properties;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

public class DGWSServiceTest {

    @Mock
    private TransportUtil httpClientService;

    @Mock
    private CredentialVault credentialVault;

    @Mock
    private SOSIFactory sosiFactory;

    @Mock
    private CredentialVault vault;

    @Mock
    private CredentialPair credentialPair;

    // This one is real
    private DGWSService dgwsService;

    @Before
    public void setUp() {
        MockitoAnnotations.initMocks(this);

        String endpoint = "testEndpoint";
        String organizationName = "testOrg";
        String organizationCode = "testCode";
        String keystorePath = "testKeystorePath";
        String keystorePassword = "testKeystorePassword";
        String systemName = "testSystemName";

        // Mock all the SoSi stuff
        when(credentialVault.getSystemCredentialPair()).thenReturn(mock(CredentialPair.class));
        when(sosiFactory.createNewSecurityTokenRequest()).thenReturn(mock(SecurityTokenRequest.class));
        when(sosiFactory.getCredentialVault()).thenReturn(vault);
        when(vault.getSystemCredentialPair()).thenReturn(credentialPair);
        when(credentialPair.getCertificate()).thenReturn(mock(X509Certificate.class));

        dgwsService = new DGWSService(endpoint, organizationName, organizationCode, keystorePath, keystorePassword, systemName, httpClientService) {
            @Override
            protected CredentialVault getVault(Properties properties, String ksPath, String ksPassword) {
                return credentialVault;
            }

            @Override
            protected SOSIFactory getSosiFactoryTestFederation(CredentialVault vault, Properties properties) {
                return sosiFactory;
            }

            @Override
            protected SOSIFactory getSosiFactoryFederation(CredentialVault vault, Properties properties) {
                return sosiFactory;
            }
        };
    }

    @Test
    public void buildSecurityTokenRequest() {
        // Act
        dgwsService.buildSecurityTokenRequest();

        // Assert
        verify(sosiFactory).createNewSecurityTokenRequest();
        verify(sosiFactory).createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any());
        verify(sosiFactory).getCredentialVault();
    }

    @Test
    public void buildSecurityTokenRequest_returnsSystemIdCard_noErrors() {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(false);
        when(securityTokenResponse.getIDCard()).thenReturn(mock(SystemIDCard.class));

        // Act
        dgwsService.buildSystemIDCard("testResponseXml");

    }

    @Test
    public void buildSecurityTokenRequest_returnsFault_throwsException() {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(true);

        // Act
        try  {
            dgwsService.buildSystemIDCard("testResponseXml");
        } catch (IllegalStateException e) {
            // Assert
            verify(securityTokenResponse).isFault();
            verify(securityTokenResponse).getFaultString();
            assertEquals("SOAP Fault from the STS", e.getMessage());
        }

    }

    @Test
    public void buildSecurityTokenRequest_returnsNull_throwsException() {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.getIDCard()).thenReturn(null);

        // Act
        try  {
            dgwsService.buildSystemIDCard("testResponseXml");
        } catch (IllegalStateException e) {
            // Assert
            verify(securityTokenResponse).isFault();
            verify(securityTokenResponse, Mockito.times(0)).getFaultString();
            assertEquals("SOAP Fault from the STS", e.getMessage());
        }
    }

    @Test
    public void createRequest() {
        // Arrange
        Request request = mock(Request.class);
        when(sosiFactory.createNewRequest(anyBoolean(), anyString())).thenReturn(request);
        when(request.serialize2DOMDocument()).thenReturn(mock(Document.class));

        // Act
        dgwsService.createRequest();

        // Assert
        verify(sosiFactory).createNewRequest(anyBoolean(), anyString());
        verify(request).setIDCard(any());
        verify(request).setDGWSVersion(anyString());
        verify(request).setBody(any());
    }

    @Test
    public void ensureValidIdCard_noPreviousIDCard_getsNewIDCard() throws IOException {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);

        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(mock(SecurityTokenResponse.class));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any())).thenReturn(mock(SystemIDCard.class));
        when(httpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn("testResponseXml");
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(false);

        when(securityTokenResponse.getIDCard()).thenReturn(new SystemIDCard("1.0.1", AuthenticationLevel.VOCES_TRUSTED_SYSTEM,
                "testIssuer", new SystemInfo(mock(CareProvider.class), "testSystemVersion"),
                "testCertHash", "testAlternativeIdentifier", "testUsername", "testPassword"));

        // Act
        dgwsService.ensureValidIdCard();

        // Verify
        verify(httpClientService).sendRequest(anyString(), anyString(), anyString());
        verify(sosiFactory).deserializeSecurityTokenResponse(anyString());
        verify(sosiFactory).createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any());
        verify(sosiFactory).getCredentialVault();
    }

    @Test
    public void ensureValidIdCard_IDCardExpired_renews() throws IOException {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);

        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(mock(SecurityTokenResponse.class));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any())).thenReturn(mock(SystemIDCard.class));
        when(httpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn("testResponseXml");
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(false);

        SystemIDCard idCard = mock(SystemIDCard.class);
        when(idCard.getExpiryDate()).thenReturn(Date.from(Instant.now().minusSeconds(10000)));
        when(securityTokenResponse.getIDCard()).thenReturn(idCard);

        // Act
        dgwsService.ensureValidIdCard();

        // Verify
        verify(httpClientService).sendRequest(anyString(), anyString(), anyString());
        verify(sosiFactory).deserializeSecurityTokenResponse(anyString());
        verify(sosiFactory).createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any());
        verify(sosiFactory).getCredentialVault();
    }

    @Test
    public void ensureValidIdCard_IDCardStillValid_doesNothing() throws IOException {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);

        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(mock(SecurityTokenResponse.class));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any())).thenReturn(mock(SystemIDCard.class));
        when(httpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn("testResponseXml");
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(false);

        SystemIDCard idCard = mock(SystemIDCard.class);
        when(idCard.getExpiryDate()).thenReturn(Date.from(Instant.now().plus(10, ChronoUnit.DAYS)));
        when(securityTokenResponse.getIDCard()).thenReturn(idCard);

        dgwsService.ensureValidIdCard();
        Mockito.clearInvocations(httpClientService);

        // Act
        dgwsService.ensureValidIdCard(); // IDCard already there

        // Verify
        // This should not have been called if IDCard is still valid
        verify(httpClientService, Mockito.times(0)).sendRequest(anyString(), anyString(), anyString());
    }

    @Test
    public void ensureValidIdCard_IDCardExpiresInFourHours_renews() throws IOException {
        // Arrange
        SecurityTokenResponse securityTokenResponse = mock(SecurityTokenResponse.class);

        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(mock(SecurityTokenResponse.class));
        when(sosiFactory.createNewSystemIDCard(anyString(), any(), any(), any(), any(), any(), any())).thenReturn(mock(SystemIDCard.class));
        when(httpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn("testResponseXml");
        when(sosiFactory.deserializeSecurityTokenResponse(anyString())).thenReturn(securityTokenResponse);
        when(securityTokenResponse.isFault()).thenReturn(false);

        SystemIDCard idCard = mock(SystemIDCard.class);
        when(idCard.getExpiryDate()).thenReturn(Date.from(Instant.now().plus(4, ChronoUnit.HOURS)));
        when(securityTokenResponse.getIDCard()).thenReturn(idCard);

        dgwsService.ensureValidIdCard();
        Mockito.clearInvocations(httpClientService);

        // Act
        dgwsService.ensureValidIdCard(); // IDCard already there

        // Verify
        verify(httpClientService, Mockito.times(1)).sendRequest(anyString(), anyString(), anyString());
    }

}
