/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.fgvhr;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

import dk.sosi.seal.model.Request;
import dk.stamkort.integrations.DGWSService;
import dk.stamkort.integrations.TransportUtil;
import dk.sundhedsdatastyrelsen.fravalg_af_genoplivning._2023._06._01.CPRIdentifierType;
import dk.sundhedsdatastyrelsen.fravalg_af_genoplivning._2023._06._01.CPRValueType;
import dk.sundhedsdatastyrelsen.fravalg_af_genoplivning._2023._06._01.InputIdentifierType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;

public class FgvhrClientTest {

    private FgvhrClient fgvhrClient;
    private DGWSService mockDgwsService;
    private TransportUtil mockHttpClientService;
    private FgvhrRequestBuilder mockRequestBuilder;

    private final String responseXmlTrue =
            "<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
            "    <soap:Body>\n" +
            "        <HasNoResuscitationResponse xmlns=\"http://sundhedsdatastyrelsen.dk/Fravalg-Af-Genoplivning/2023/06/01/\">true\n" +
            "        </HasNoResuscitationResponse>\n" +
            "    </soap:Body>\n" +
            "</soap:Envelope>";
    private final String responseXmlFalse =
            "<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                    "    <soap:Body>\n" +
                    "        <HasNoResuscitationResponse xmlns=\"http://sundhedsdatastyrelsen.dk/Fravalg-Af-Genoplivning/2023/06/01/\">false\n" +
                    "        </HasNoResuscitationResponse>\n" +
                    "    </soap:Body>\n" +
                    "</soap:Envelope>";
    private final String endpoint = "myEndpointUrl";

    @Before
    public void setUp() {
        mockDgwsService = mock(DGWSService.class);
        mockHttpClientService = mock(TransportUtil.class);
        mockRequestBuilder = mock(FgvhrRequestBuilder.class);
        fgvhrClient = new FgvhrClient(endpoint, mockDgwsService, mockHttpClientService, mockRequestBuilder);
    }

    @Test
    public void hasNoResuscitation_returnsValues() throws Exception {
        // Arrange
        String patientId = "testPatientId";
        CPRValueType CPRValueType = new CPRValueType();
        CPRValueType.setValue(patientId);
        CPRIdentifierType cprIdentifierType = new CPRIdentifierType();
        cprIdentifierType.setValue(CPRValueType);
        InputIdentifierType inputIdentifierType = new InputIdentifierType();
        inputIdentifierType.setIdentifier(cprIdentifierType);

        when(mockRequestBuilder.createHasNoResuscitationRequest(patientId)).thenReturn(inputIdentifierType);
        when(mockHttpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn(responseXmlTrue);
        when(mockDgwsService.createRequest()).thenReturn(mock(Request.class));

        // Act
        fgvhrClient.hasNoResuscitation(patientId);

        // Assert
        ArgumentCaptor<String> endpointCaptor = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> actionCaptor = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<String> requestCaptor = ArgumentCaptor.forClass(String.class);

        verify(mockHttpClientService).sendRequest(endpointCaptor.capture(), actionCaptor.capture(), requestCaptor.capture());

        assertEquals(endpoint, endpointCaptor.getValue());
        assertNotNull(actionCaptor.getValue());
        assertNotNull(requestCaptor.getValue());
    }

    @Test
    public void hasNoResuscitation_responseContainsTrue_returnsTrue() throws Exception {
        // Arrange
        String patientId = "testPatientId";
        CPRValueType CPRValueType = new CPRValueType();
        CPRValueType.setValue(patientId);
        CPRIdentifierType cprIdentifierType = new CPRIdentifierType();
        cprIdentifierType.setValue(CPRValueType);
        InputIdentifierType inputIdentifierType = new InputIdentifierType();
        inputIdentifierType.setIdentifier(cprIdentifierType);

        when(mockRequestBuilder.createHasNoResuscitationRequest(patientId)).thenReturn(inputIdentifierType);
        when(mockHttpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn(responseXmlTrue);
        when(mockDgwsService.createRequest()).thenReturn(mock(Request.class));

        // Act
        boolean result = fgvhrClient.hasNoResuscitation(patientId);

        // Assert
        assertTrue(result);
    }

    @Test
    public void hasNoResuscitation_responseContainsFalse_returnsFalse() throws Exception {
        // Arrange
        String patientId = "testPatientId";
        CPRValueType CPRValueType = new CPRValueType();
        CPRValueType.setValue(patientId);
        CPRIdentifierType cprIdentifierType = new CPRIdentifierType();
        cprIdentifierType.setValue(CPRValueType);
        InputIdentifierType inputIdentifierType = new InputIdentifierType();
        inputIdentifierType.setIdentifier(cprIdentifierType);

        when(mockRequestBuilder.createHasNoResuscitationRequest(patientId)).thenReturn(inputIdentifierType);
        when(mockHttpClientService.sendRequest(anyString(), anyString(), anyString())).thenReturn(responseXmlFalse);
        when(mockDgwsService.createRequest()).thenReturn(mock(Request.class));

        // Act
        boolean result = fgvhrClient.hasNoResuscitation(patientId);

        // Assert
        assertFalse(result);
    }
}
