/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util;

import dk.nsi.hsuid.ActingUserCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.Attribute;
import dk.nsi.hsuid.CitizenCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.HealthcareServiceUserIdentificationHeaderUtil;
import dk.nsi.hsuid.OperationsOrganisationNameAttribute;
import dk.nsi.hsuid.OrganisationIdentifierAttribute;
import dk.nsi.hsuid.ResponsibleUserAuthorizationCodeAttribute;
import dk.nsi.hsuid.ResponsibleUserCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.SystemNameAttribute;
import dk.nsi.hsuid.SystemVendorNameAttribute;
import dk.nsi.hsuid.SystemVersionAttribute;
import dk.nsi.hsuid.UserTypeAttribute;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.sds.nsp.security.SecurityContext.Organisation;
import dk.stamkort.context.*;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.test.context.junit4.SpringRunner;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
public class RequestContextProviderTest {
    private static final String WHITELISTED_CVR = "12345678";

    @InjectMocks
    RequestContextProvider requestContextProvider;

    @Mock
    Properties properties;

    @Before
    public void setup() {
        Mockito.when(properties.getWhitelistedLevel3CVRs()).thenReturn(WHITELISTED_CVR);
    }

    @Test
    public void getSecurityData_flowId() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals("foo", result.getConversationid());
    }

    @Test
    public void getSecurityData_messageId() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals("bar", result.getMessageId());
    }

    @Test(expected = FSKException.class)
    public void getSecurityData_dgwsAuthenticationLevel3NoOrganisation() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskNoOrgDevelopmentSecurityProvider.class.getName());

        // Act
        try {
            RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);
        } catch (FSKException e) {
            // Assert
        	Assert.assertEquals("Organisation og organisationsidentifikator skal være tilstede i sikkehedsbilletten", e.getMessage());
        	throw e;
        }
    }

    @Test
    public void getSecurityData_dgwsAuthenticationLevel3WithOrganisation() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals(RequestContextVO.UserType.SystemUser, result.getUserType());
        assertNotNull(result.getOrganisation());
        assertEquals("12345678", result.getOrganisation().getIdentifier());
        Assert.assertNull(result.getGivenName());
        Assert.assertNull(result.getSurName());
        Assert.assertNull(result.getUserCprIdentifier());
        Assert.assertNull(result.getAuthorizationCode());
        Assert.assertNull(result.getEducationCode());
    }

    @Test(expected = FSKException.class)
    public void getSecurityData_citizenTypeRejected() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskCitizenDevelopmentSecurityProvider.class.getName());

        // Act
        try {
            RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);
        } catch (FSKException e) {
            // Assert
        	Assert.assertEquals("500: Borgerbillet ikke tilladt", e.getMessage());
        	throw e;
        }
    }

    @Test
    public void getSecurityData_authenticationLevel4() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskHealthcareProfDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals(RequestContextVO.UserType.HealthCareProfessional, result.getUserType());
    }

    @Test
    public void getSecurityData_healthcareProfessionalWithHsuidHeader() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskHealthcareProfDevelopmentSecurityProvider.class.getName());
        String hsuidIdentifier = "0909990909";

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(createHsuidHeader(hsuidIdentifier));

        // Assert
        assertEquals(RequestContextVO.UserType.HealthCareProfessional, result.getUserType());
        assertEquals(hsuidIdentifier, result.getPatientInContext());
        assertEquals(hsuidIdentifier, result.getUserCprIdentifier());
    }

    @Test
    public void getSecurityData_healthcareProfessionalWithUnauthorizedHsuidHsuidHeader() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskHealthcareProfDevelopmentSecurityProvider.class.getName());
        String hsuidIdentifier = "0909990909";

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(createUnauthorizedHsuidHeader(hsuidIdentifier));

        // Assert
        assertEquals(RequestContextVO.UserType.HealthCareProfessional, result.getUserType());
        assertEquals(hsuidIdentifier, result.getPatientInContext());
        assertEquals(hsuidIdentifier, result.getUserCprIdentifier());
    }

    @Test
    public void getSecurityData_cprnr() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskHealthcareProfDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals("0101010101", result.getUserCprIdentifier());
    }

    @Test
    public void getSecurityData_careProviderName() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskNoMessageDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals("Lægehuset", result.getOrganisation().getName().get());
    }

    @Test
    public void getSecurityData_careProviderId() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskNoMessageDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals("12345678", result.getOrganisation().getIdentifier());
    }

    @Test
    public void getSecurityData_careProviderType() {
        // Arrange
        System.setProperty("dk.sds.nsp.security.provider", FskNoMessageDevelopmentSecurityProvider.class.getName());

        // Act
        RequestContextVO result = requestContextProvider.getAndValidateSecuityData(null);

        // Assert
        assertEquals(Organisation.OrganisationIdentifierFormat.CVR, result.getOrganisation().getIdentifierFormat());
    }

    @Test
    public void testWhitelisting() {
        assertFalse(requestContextProvider.cvrIsWhiteListed(null));
        assertFalse(requestContextProvider.cvrIsWhiteListed(""));
        assertFalse(requestContextProvider.cvrIsWhiteListed("12341234"));
        assertTrue(requestContextProvider.cvrIsWhiteListed(WHITELISTED_CVR));
    }

    private HsuidHeader createHsuidHeader(String cpr) {
        List<Attribute> attributes = new ArrayList<Attribute>();
        attributes.add(UserTypeAttribute.HEALTHCAREPROFESSIONAL);
        attributes.add(new CitizenCivilRegistrationNumberAttribute(cpr));
        attributes.add(new ActingUserCivilRegistrationNumberAttribute(cpr));
        attributes.add(new SystemVendorNameAttribute("system Owner"));
        attributes.add(new SystemNameAttribute("system Name"));
        attributes.add(new SystemVersionAttribute("system Version"));
        attributes.add(new OperationsOrganisationNameAttribute("responsible Org"));
        attributes.add(new ResponsibleUserCivilRegistrationNumberAttribute("0101010101"));
        attributes.add(new ResponsibleUserAuthorizationCodeAttribute("ZXY71"));
        attributes.add(new OrganisationIdentifierAttribute("293591000016003", "NSI_SORCODE"));

        return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader("system owner", attributes);
    }

    private HsuidHeader createUnauthorizedHsuidHeader(String cpr) {
        List<Attribute> attributes = new ArrayList<Attribute>();
        attributes.add(UserTypeAttribute.UNAUTHORIZEDPROFESSIONAL);
        attributes.add(new CitizenCivilRegistrationNumberAttribute(cpr));
        attributes.add(new ActingUserCivilRegistrationNumberAttribute(cpr));
        attributes.add(new SystemVendorNameAttribute("system Owner"));
        attributes.add(new SystemNameAttribute("system Name"));
        attributes.add(new SystemVersionAttribute("system Version"));
        attributes.add(new OperationsOrganisationNameAttribute("responsible Org"));
        attributes.add(new ResponsibleUserCivilRegistrationNumberAttribute("0101010101"));
        attributes.add(new ResponsibleUserAuthorizationCodeAttribute("ZXY71"));
        attributes.add(new OrganisationIdentifierAttribute("293591000016003", "NSI_SORCODE"));

        return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader("system owner", attributes);
    }
}