/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler;

import static java.lang.Boolean.*;
import static java.util.Optional.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.InputStream;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.namespace.QName;

import dk.stamkort.domain.YderNr;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import org.hl7.v3.AD;
import org.hl7.v3.ANY;
import org.hl7.v3.AdxpCity;
import org.hl7.v3.AdxpCountry;
import org.hl7.v3.AdxpPostalCode;
import org.hl7.v3.AdxpStreetAddressLine;
import org.hl7.v3.CD;
import org.hl7.v3.EN;
import org.hl7.v3.EnFamily;
import org.hl7.v3.EnGiven;
import org.hl7.v3.EnPrefix;
import org.hl7.v3.II;
import org.hl7.v3.PN;
import org.hl7.v3.POCDMT000040Author;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.hl7.v3.POCDMT000040Component2;
import org.hl7.v3.POCDMT000040Custodian;
import org.hl7.v3.POCDMT000040Observation;
import org.hl7.v3.POCDMT000040Organization;
import org.hl7.v3.POCDMT000040Patient;
import org.hl7.v3.POCDMT000040PatientRole;
import org.hl7.v3.POCDMT000040RecordTarget;
import org.hl7.v3.POCDMT000040Section;
import org.hl7.v3.POCDMT000040StructuredBody;
import org.hl7.v3.ST;
import org.hl7.v3.TEL;
import org.hl7.v3.XActMoodDocumentObservation;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.xml.sax.ContentHandler;

import dk.oio.rep.cpr_dk.xml.schemas.core._2005._11._24.PersonBirthDateStructureType;
import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.CountryIdentificationCodeType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.stamkort.conf.FskConfig;
import dk.stamkort.conf.TestDAOConfig;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.util.AuthorProperties;
import dk.stamkort.util.Properties;
import dk.stamkort.webservice.hl7cda.PdcDataMapperV2;
import dk.stamkort.webservice.hl7cda.assembler.mapper.PDC_2_0ElementsMapper;
import dk.stamkort.webservice.hl7cda.model.AddressHolder;
import dk.stamkort.webservice.hl7cda.model.CodeHolder;
import dk.stamkort.webservice.hl7cda.model.CoverageGroupHolder;
import dk.stamkort.webservice.hl7cda.model.CustodyHolder;
import dk.stamkort.webservice.hl7cda.model.DentistTypedInHolder;
import dk.stamkort.webservice.hl7cda.model.IdHolder;
import dk.stamkort.webservice.hl7cda.model.LanguageTypedInHolder;
import dk.stamkort.webservice.hl7cda.model.LivingWillRegistrationHolder;
import dk.stamkort.webservice.hl7cda.model.OrganDonorRegistrationHolder;
import dk.stamkort.webservice.hl7cda.model.PdcV2Data;
import dk.stamkort.webservice.hl7cda.model.PatientContactTypedInHolder;
import dk.stamkort.webservice.hl7cda.model.PhoneHolder;
import dk.stamkort.webservice.hl7cda.model.PrimaryPersonInformationHolder;
import dk.stamkort.webservice.hl7cda.model.ProviderOrganizationHolder;
import dk.stamkort.webservice.hl7cda.model.RelativeTypedInHolder;
import dk.stamkort.webservice.hl7cda.model.TempAddrTypedInHolder;
import dk.stamkort.webservice.hl7cda.model.TreatmentWillRegistrationHolder;
import oio.medcom.cprservice._1_0.AddressCompleteType;
import oio.medcom.cprservice._1_0.AddressPostalType;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import oio.medcom.cprservice._1_0.PersonAddressStructureType;
import oio.medcom.cprservice._1_0.PersonHealthCareInformationStructureType;
import oio.medcom.cprservice._1_0.PersonInformationStructureType;
import oio.medcom.cprservice._1_0.PersonPublicHealthInsuranceType;
import oio.medcom.cprservice._1_0.PersonWithHealthCareInformationStructureType;
import oio.medcom.cprservice._1_0.PublicHealthInsuranceGroupIdentifierType;
import oio.medcom.cprservice._1_0.PublicHealthInsuranceStatusType;
import oio.medcom.cprservice._1_0.RegularCPRPersonType;
import oio.medcom.cprservice._1_0.SimpleCPRPersonType;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { TestDAOConfig.class, FskConfig.class })
public class PDC_2_0ClinicalDocumentAssemblerTest {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private ContentHandler handler;

    private static Marshaller marshaller;

    @Autowired
    Properties properties;

    @BeforeClass
    public static void beforeClass() throws Exception {
        JAXBContext context = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        marshaller = context.createMarshaller();
    }

    @Test
    public void assembleClinicalDocument() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response_Clinical.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setClinicalDocument(personalDataCardResponse.getClinicalDocument());

        YderVO yder = new YderVO();
        yder.setYdernr(new YderNr("123123"));
        yder.setPhone("22334455");
        yder.setAddress("Adresse 1");
        yder.setPostcode("1234");
        yder.setCity("City");
        PdcV2Data dataHolder = PdcDataMapperV2.mapDataHolder(registerData, yder);
        AuthorProperties authorProperties = properties.createAuthorProperties();
        POCDMT000040ClinicalDocument result = PDC_2_0ClinicalDocumentAssembler.assembleClinicalDocument(dataHolder, authorProperties);

        assertEquals(1, result.getMoodCodes().size());
        assertEquals("EVN", result.getMoodCodes().get(0));
        assertEquals(1, result.getRecordTargets().size());
        assertEquals(1, result.getAuthors().size());
        assertEquals("1.2.208.176.1.1", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getRoot());
        assertEquals("1126211000016009", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getExtension());
        assertEquals("SOR", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getAssigningAuthorityName());
        assertEquals("Sundhedsdatastyrelsen", result.getAuthors().get(0).getAssignedAuthor().getRepresentedOrganization().getNames().get(0).getContent().get(0));
        assertEquals("Sundhedsdatastyrelsen", result.getCustodian().getAssignedCustodian().getRepresentedCustodianOrganization().getName().getContent().get(0));
        assertEquals(POCDMT000040Custodian.class, result.getCustodian().getClass());
        assertEquals(2, result.getDocumentationOves().size());
        assertEquals(POCDMT000040Component2.class, result.getComponent().getClass());
        assertEquals(POCDMT000040StructuredBody.class, result.getComponent().getStructuredBody().getClass());
        assertEquals(1, result.getComponent().getStructuredBody().getComponents().size());
        assertEquals(POCDMT000040Section.class, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getClass());
        assertEquals(10, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().size());
        assertEquals(1, result.getMoodCodes().size());
        assertEquals("EVN", result.getMoodCodes().get(0));

        assertEquals(2, result.getDocumentationOves().size());
        assertEquals("DOC", result.getDocumentationOves().get(1).getTypeCodes().get(0));
        assertEquals("pdc-v2.0", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getExtension());
        assertEquals("1.2.208.184.100.10", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getRoot());
        assertEquals("MedCom", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getAssigningAuthorityName());
        assertEquals("EVN", result.getDocumentationOves().get(1).getServiceEvent().getMoodCodes().get(0));
        assertEquals("ACT", result.getDocumentationOves().get(1).getServiceEvent().getClassCodes().get(0));

        assertEquals(1, result.getTemplateIds().size());
        assertEquals("1.2.208.184.16.1", result.getTemplateIds().get(0).getRoot());

        assertEquals("1.2.208.184.10.1.11", result.getDocumentationOves().get(0).getServiceEvent().getTemplateIds().get(0).getRoot());
        assertEquals("1.2.208.184.10.1.10", result.getDocumentationOves().get(1).getServiceEvent().getTemplateIds().get(0).getRoot());

        assertEquals("20190101", result.getRecordTargets().get(0).getPatientRole().getPatient().getBirthTime().getValue());

        dataHolder.getRelativeTypedInHolder().forEach(c -> {
            assertEquals(1, c.getAuthorGivens().size());
            assertEquals("Fissirul", c.getAuthorGivens().get(0));
            assertEquals("Lohman", c.getAuthorFamily());
        });
        dataHolder.getLanguageTypedInHolder().forEach(c -> {
            assertEquals(1, c.getAuthorGivens().size());
            assertEquals("Fissirul", c.getAuthorGivens().get(0));
            assertEquals("Lohman", c.getAuthorFamily());
        });
        dataHolder.getTempAddrTypedInHolder().forEach(c -> {
            assertEquals(1, c.getAuthorGivens().size());
            assertEquals("Fissirul", c.getAuthorGivens().get(0));
            assertEquals("Lohman", c.getAuthorFamily());
        });
        dataHolder.getPatientContactTypedInHolder().forEach(c -> {
            assertEquals(1, c.getAuthorGivens().size());
            assertEquals("Fissirul", c.getAuthorGivens().get(0));
            assertEquals("Lohman", c.getAuthorFamily());
        });
    }

    @Test
    public void assembleClinicalDocumentNoAutherName() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response_Clinical_Missing_AssignedPerson.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setClinicalDocument(personalDataCardResponse.getClinicalDocument());

        YderVO yder = new YderVO();
        yder.setYdernr(new YderNr("123123"));
        yder.setPhone("22334455");
        yder.setAddress("Adresse 1");
        yder.setPostcode("1234");
        yder.setCity("City");
        PdcV2Data dataHolder = PdcDataMapperV2.mapDataHolder(registerData, yder);
        AuthorProperties authorProperties = properties.createAuthorProperties();
        POCDMT000040ClinicalDocument result = PDC_2_0ClinicalDocumentAssembler.assembleClinicalDocument(dataHolder, authorProperties);

        assertEquals(1, result.getMoodCodes().size());
        assertEquals("EVN", result.getMoodCodes().get(0));
        assertEquals(1, result.getRecordTargets().size());
        assertEquals(1, result.getAuthors().size());
        assertEquals("1.2.208.176.1.1", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getRoot());
        assertEquals("1126211000016009", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getExtension());
        assertEquals("SOR", result.getAuthors().get(0).getAssignedAuthor().getIds().get(0).getAssigningAuthorityName());
        assertEquals("Sundhedsdatastyrelsen", result.getAuthors().get(0).getAssignedAuthor().getRepresentedOrganization().getNames().get(0).getContent().get(0));
        assertEquals("Sundhedsdatastyrelsen", result.getCustodian().getAssignedCustodian().getRepresentedCustodianOrganization().getName().getContent().get(0));
        assertEquals(POCDMT000040Custodian.class, result.getCustodian().getClass());
        assertEquals(2, result.getDocumentationOves().size());
        assertEquals(POCDMT000040Component2.class, result.getComponent().getClass());
        assertEquals(POCDMT000040StructuredBody.class, result.getComponent().getStructuredBody().getClass());
        assertEquals(1, result.getComponent().getStructuredBody().getComponents().size());
        assertEquals(POCDMT000040Section.class, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getClass());
        assertEquals(10, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().size());
        assertEquals(1, result.getMoodCodes().size());
        assertEquals("EVN", result.getMoodCodes().get(0));

        assertEquals(2, result.getDocumentationOves().size());
        assertEquals("DOC", result.getDocumentationOves().get(1).getTypeCodes().get(0));
        assertEquals("pdc-v2.0", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getExtension());
        assertEquals("1.2.208.184.100.10", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getRoot());
        assertEquals("MedCom", result.getDocumentationOves().get(1).getServiceEvent().getIds().get(0).getAssigningAuthorityName());
        assertEquals("EVN", result.getDocumentationOves().get(1).getServiceEvent().getMoodCodes().get(0));
        assertEquals("ACT", result.getDocumentationOves().get(1).getServiceEvent().getClassCodes().get(0));

        assertEquals(1, result.getTemplateIds().size());
        assertEquals("1.2.208.184.16.1", result.getTemplateIds().get(0).getRoot());

        assertEquals("1.2.208.184.10.1.11", result.getDocumentationOves().get(0).getServiceEvent().getTemplateIds().get(0).getRoot());
        assertEquals("1.2.208.184.10.1.10", result.getDocumentationOves().get(1).getServiceEvent().getTemplateIds().get(0).getRoot());

        assertEquals("20190101", result.getRecordTargets().get(0).getPatientRole().getPatient().getBirthTime().getValue());

        // Check assignedPerson
        dataHolder.getRelativeTypedInHolder().forEach(c ->
                checkAssignedPersonNotFilledOutPlaceHolder(c.getAuthorGivens(), c.getAuthorFamily())
        );
        dataHolder.getLanguageTypedInHolder().forEach(c ->
                checkAssignedPersonNotFilledOutPlaceHolder(c.getAuthorGivens(), c.getAuthorFamily())
        );
        dataHolder.getTempAddrTypedInHolder().forEach(c ->
                checkAssignedPersonNotFilledOutPlaceHolder(c.getAuthorGivens(), c.getAuthorFamily())
        );
        dataHolder.getPatientContactTypedInHolder().forEach(c ->
                checkAssignedPersonNotFilledOutPlaceHolder(c.getAuthorGivens(), c.getAuthorFamily())
        );
    }

    void checkAssignedPersonNotFilledOutPlaceHolder(List<String> authorGivens, String authorFamily) {
        String expectValue = "Ikke udfyldt";
        for (String authorGiven : authorGivens) {
            assertEquals(expectValue, authorGiven);
        }
        assertEquals(expectValue, authorFamily);
    }

    @Test
    public void assembleClinicalDocumentNoEndDateOnTemporaryAddress() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response_Clinical_No_End_Date_On_Temporary_Address.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setClinicalDocument(personalDataCardResponse.getClinicalDocument());

        YderVO yder = new YderVO();
        yder.setYdernr(new YderNr("123123"));
        yder.setPhone("22334455");
        yder.setAddress("Adresse 1");
        yder.setPostcode("1234");
        yder.setCity("City");


        PdcV2Data dataHolder = PdcDataMapperV2.mapDataHolder(registerData, yder);
        AuthorProperties authorProperties = properties.createAuthorProperties();
        POCDMT000040ClinicalDocument result = PDC_2_0ClinicalDocumentAssembler.assembleClinicalDocument(dataHolder, authorProperties);

        assertEquals(1, result.getRecordTargets().size());
        assertEquals(1, result.getAuthors().size());
        assertEquals(POCDMT000040Custodian.class, result.getCustodian().getClass());
        assertEquals(2, result.getDocumentationOves().size());
        assertEquals(POCDMT000040Component2.class, result.getComponent().getClass());
        assertEquals(POCDMT000040StructuredBody.class, result.getComponent().getStructuredBody().getClass());
        assertEquals(1, result.getComponent().getStructuredBody().getComponents().size());
        assertEquals(POCDMT000040Section.class, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getClass());
        assertEquals(10, result.getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().size());
    }

    @Test
    public void createRecordTarget() {
        // Arrange
        PrimaryPersonInformationHolder person = createPrimaryPersonInformationHolderMock();
        ProviderOrganizationHolder providerOrganizationHolder = createProviderOrganizationHolderMock();

        // Act
        POCDMT000040RecordTarget result = PDC_2_0ClinicalDocumentAssembler.createRecordTarget(person, providerOrganizationHolder);

        // Assert
        assertEquals("OP", result.getContextControlCode());
        assertEquals("RCT", result.getTypeCodes().get(0));

        POCDMT000040PatientRole patientRole = result.getPatientRole();

        assertEquals(POCDMT000040PatientRole.class, patientRole.getClass());

        assertEquals("PAT", patientRole.getClassCodes().get(0));

        II id = patientRole.getIds().get(0);
        assertEquals("1.2.208.176.1.2", id.getRoot());
        assertNotNull(id.getExtension());
        assertEquals("CPR", id.getAssigningAuthorityName());

        assertEquals(AD.class, patientRole.getAddrs().get(0).getClass());
        assertEquals(POCDMT000040Patient.class, patientRole.getPatient().getClass());
        assertEquals(POCDMT000040Organization.class, patientRole.getProviderOrganization().getClass());

        testMarshalling(result);
    }

    @Test
    public void createCustodyObservation() {
        CustodyHolder relativeTypedInHolder = createCustodyHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createCustodyObservation(relativeTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(3, values.size());

        // assert cpr
        assertEquals(relativeTypedInHolder.getCpr(), ((II) values.get(0)).getExtension());

        // assert given
        //noinspection unchecked
        assertEquals(
                relativeTypedInHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((PN) values.get(1)).getContent().get(0)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                relativeTypedInHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((PN) values.get(1)).getContent().get(1)).getValue().getContent().get(0));

        // assert relation
        assertEquals(relativeTypedInHolder.getRelation().getCode(), ((CD) values.get(2)).getCode());
        assertEquals(relativeTypedInHolder.getRelation().getDisplayName(), ((CD) values.get(2)).getDisplayName());

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createCustodyObservation_sameIdIsUuid() {
        String id =
                PDC_2_0ClinicalDocumentAssembler.createCustodyObservation(createCustodyHolderMock())
                        .getIds().get(0).getExtension();
        UUID.fromString(id);
    }

    @Test
    public void createCitizenNameAddrObservation() {
        PrimaryPersonInformationHolder primaryPersonInformationHolder = createPrimaryPersonInformationHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createCitizenNameAddrObservation(primaryPersonInformationHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(2, values.size());

        // assert given
        //noinspection unchecked
        assertEquals(
                primaryPersonInformationHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((PN) values.get(0)).getContent().get(0)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                primaryPersonInformationHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((PN) values.get(0)).getContent().get(1)).getValue().getContent().get(0));

        // assert address use
        assertEquals(primaryPersonInformationHolder.getAddress().getUse(), ((AD) values.get(1)).getUses().get(0));

        // assert address addressLines
        //noinspection unchecked
        assertTrue(((JAXBElement<AdxpStreetAddressLine>) ((AD) values.get(1)).getContent().get(0)).getValue().getContent().containsAll(
                primaryPersonInformationHolder.getAddress().getAddressLines()));

        // assert address postalCode
        //noinspection unchecked
        assertEquals(
                primaryPersonInformationHolder.getAddress().getPostalCode(),
                ((JAXBElement<AdxpPostalCode>) ((AD) values.get(1)).getContent().get(1)).getValue().getContent().get(0));

        // assert address city
        //noinspection unchecked
        assertEquals(
                primaryPersonInformationHolder.getAddress().getCity(),
                ((JAXBElement<AdxpCity>) ((AD) values.get(1)).getContent().get(2)).getValue().getContent().get(0));

        // assert address country
        //noinspection unchecked
        assertEquals(
                primaryPersonInformationHolder.getAddress().getCountry(),
                ((JAXBElement<AdxpCountry>) ((AD) values.get(1)).getContent().get(3)).getValue().getContent().get(0));

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createCoverageGroupObservation() {
        CoverageGroupHolder coverageGroupHolder = createCoverageGroupHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createCoverageGroupObservation(coverageGroupHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());
        assertEquals(coverageGroupHolder.getGroup(), ((II) values.get(0)).getExtension());

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createOrganDonorRegistrationObservation() {
        OrganDonorRegistrationHolder organDonorRegistrationHolder = createOrganDonorRegistrationHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createOrganDonorRegistrationObservation(organDonorRegistrationHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());
        assertEquals(Boolean.toString(organDonorRegistrationHolder.isHasRegistration()), ((II) values.get(0)).getExtension());

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createTreatmentWillRegistrationObservation() {
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = createTreatmentWillRegistrationHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createTreatmentWillRegistrationObservation(treatmentWillRegistrationHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());
        assertEquals(Boolean.toString(treatmentWillRegistrationHolder.isHasRegistration()), ((II) values.get(0)).getExtension());

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createLivingWillRegistrationObservation() {
        LivingWillRegistrationHolder livingWillRegistrationHolder = createLivingWillRegistrationHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createLivingWillRegistrationObservation(livingWillRegistrationHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());
        assertEquals(Boolean.toString(livingWillRegistrationHolder.isHasRegistration()), ((II) values.get(0)).getExtension());

        assertEquals(1, result.getAuthors().size());

        testMarshalling(result);
    }

    @Test
    public void createLanguageTypedInObservation() {
        LanguageTypedInHolder languageTypedInHolder = createLanguageTypedInHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createLanguageTypedInObservation(languageTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());
        assertEquals(languageTypedInHolder.getLanguageCode(), ((CD) values.get(0)).getCode());
        assertEquals(languageTypedInHolder.getLanguageName(), ((CD) values.get(0)).getDisplayName());

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(languageTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(languageTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                languageTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createTempAddrTypedInObservation() {
        TempAddrTypedInHolder tempAddrTypedInHolder = createTempAddrTypedInHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createTempAddrTypedInObservation(tempAddrTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(tempAddrTypedInHolder.getPeriodFrom().atStartOfDay(ZoneId.systemDefault())),
                result.getEffectiveTime().getLow().getValue());

        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(tempAddrTypedInHolder.getPeriodTo().atStartOfDay(ZoneId.systemDefault())),
                result.getEffectiveTime().getHigh().getValue());

        List<ANY> values = result.getValues();
        assertEquals(1, values.size());

        // assert address use
        assertEquals(tempAddrTypedInHolder.getAddress().getUse(), ((AD) values.get(0)).getUses().get(0));

        // assert address addressLines
        //noinspection unchecked
        assertTrue(((JAXBElement<AdxpStreetAddressLine>) ((AD) values.get(0)).getContent().get(0)).getValue().getContent().containsAll(
                tempAddrTypedInHolder.getAddress().getAddressLines()));

        // assert address postalCode
        //noinspection unchecked
        assertEquals(
                tempAddrTypedInHolder.getAddress().getPostalCode(),
                ((JAXBElement<AdxpPostalCode>) ((AD) values.get(0)).getContent().get(1)).getValue().getContent().get(0));

        // assert address city
        //noinspection unchecked
        assertEquals(
                tempAddrTypedInHolder.getAddress().getCity(),
                ((JAXBElement<AdxpCity>) ((AD) values.get(0)).getContent().get(2)).getValue().getContent().get(0));

        // assert address country
        //noinspection unchecked
        assertEquals(
                tempAddrTypedInHolder.getAddress().getCountry(),
                ((JAXBElement<AdxpCountry>) ((AD) values.get(0)).getContent().get(3)).getValue().getContent().get(0));

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(tempAddrTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(tempAddrTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                tempAddrTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createDentistTypedInObservation() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(4, values.size());

        // Yder
        assertEquals(dentistTypedInHolder.getYder(), (((II) values.get(0)).getExtension()));

        // assert prefix
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getPrefix(),
                ((JAXBElement<EnPrefix>) ((EN) values.get(1)).getContent().get(0)).getValue().getContent().get(0));

        // assert given
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((EN) values.get(1)).getContent().get(1)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((EN) values.get(1)).getContent().get(2)).getValue().getContent().get(0));

        // assert address use
        assertEquals(dentistTypedInHolder.getAddress().getUse(), ((AD) values.get(2)).getUses().get(0));

        // assert address addressLines
        //noinspection unchecked
        assertTrue(((JAXBElement<AdxpStreetAddressLine>) ((AD) values.get(2)).getContent().get(0)).getValue().getContent().containsAll(
                dentistTypedInHolder.getAddress().getAddressLines()));

        // assert address postalCode
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getPostalCode(),
                ((JAXBElement<AdxpPostalCode>) ((AD) values.get(2)).getContent().get(1)).getValue().getContent().get(0));

        // assert address city
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCity(),
                ((JAXBElement<AdxpCity>) ((AD) values.get(2)).getContent().get(2)).getValue().getContent().get(0));

        // assert address country
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCountry(),
                ((JAXBElement<AdxpCountry>) ((AD) values.get(2)).getContent().get(3)).getValue().getContent().get(0));

        // assert phone numbeer
        //noinspection unchecked
        assertEquals(dentistTypedInHolder.getPhone().get().getValue(),((TEL) values.get(3)).getValue());
        assertEquals(dentistTypedInHolder.getPhone().get().getUses().get(0),((TEL) values.get(3)).getUses().get(0));

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(dentistTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(dentistTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createDentistTypedInObservationNoAddress() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getAddress().getAddressLines().clear();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(4, values.size());

        // Yder
        assertEquals(dentistTypedInHolder.getYder(), (((II) values.get(0)).getExtension()));

        // assert prefix
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getPrefix(),
                ((JAXBElement<EnPrefix>) ((EN) values.get(1)).getContent().get(0)).getValue().getContent().get(0));

        // assert given
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((EN) values.get(1)).getContent().get(1)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((EN) values.get(1)).getContent().get(2)).getValue().getContent().get(0));

        // assert address use
        assertEquals(dentistTypedInHolder.getAddress().getUse(), ((AD) values.get(2)).getUses().get(0));

        // assert address postalCode
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getPostalCode(),
                ((JAXBElement<AdxpPostalCode>) ((AD) values.get(2)).getContent().get(0)).getValue().getContent().get(0));

        // assert address city
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCity(),
                ((JAXBElement<AdxpCity>) ((AD) values.get(2)).getContent().get(1)).getValue().getContent().get(0));

        // assert address country
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCountry(),
                ((JAXBElement<AdxpCountry>) ((AD) values.get(2)).getContent().get(2)).getValue().getContent().get(0));

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(dentistTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(dentistTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createDentistTypedInObservationNoPhone() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.setPhone(empty());

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(4, values.size());

        // Yder
        assertEquals(dentistTypedInHolder.getYder(), (((II) values.get(0)).getExtension()));

        // assert prefix
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getPrefix(),
                ((JAXBElement<EnPrefix>) ((EN) values.get(1)).getContent().get(0)).getValue().getContent().get(0));

        // assert given
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((EN) values.get(1)).getContent().get(1)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((EN) values.get(1)).getContent().get(2)).getValue().getContent().get(0));

        // assert address use
        assertEquals(dentistTypedInHolder.getAddress().getUse(), ((AD) values.get(2)).getUses().get(0));

        // assert address addressLines
        //noinspection unchecked
        assertTrue(((JAXBElement<AdxpStreetAddressLine>) ((AD) values.get(2)).getContent().get(0)).getValue().getContent().containsAll(
                dentistTypedInHolder.getAddress().getAddressLines()));

        // assert address postalCode
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getPostalCode(),
                ((JAXBElement<AdxpPostalCode>) ((AD) values.get(2)).getContent().get(1)).getValue().getContent().get(0));

        // assert address city
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCity(),
                ((JAXBElement<AdxpCity>) ((AD) values.get(2)).getContent().get(2)).getValue().getContent().get(0));

        // assert address country
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAddress().getCountry(),
                ((JAXBElement<AdxpCountry>) ((AD) values.get(2)).getContent().get(3)).getValue().getContent().get(0));

        // assert phone numbeer
        //noinspection unchecked
        assertEquals("NI", ((TEL) values.get(3)).getNullFlavors().get(0));
        assertEquals("WP",((TEL) values.get(3)).getUses().get(0));

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(dentistTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(dentistTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                dentistTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createPatientContactTypedInObservation() {
        PatientContactTypedInHolder patientContactTypedInHolder = createPatientContactTypedInHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createPatientContactTypedInObservation(patientContactTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(2, values.size());
        assertEquals(patientContactTypedInHolder.getPhones().get(0).getValue(), ((TEL) values.get(0)).getValue());
        assertEquals(patientContactTypedInHolder.getEmailAddresses().get(0).getValue(), ((TEL) values.get(1)).getValue());

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(patientContactTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(patientContactTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                patientContactTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));

        testMarshalling(result);
    }

    @Test
    public void createRelativeTypedInObservation() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        // assert observation attributes
        assertEquals(1, result.getClassCodes().size());
        assertEquals("OBS", result.getClassCodes().get(0));
        assertEquals(XActMoodDocumentObservation.EVN, result.getMoodCode());

        assertEquals(1, result.getTemplateIds().size());

        assertEquals(1, result.getIds().size());

        assertEquals(CD.class, result.getCode().getClass());

        List<ANY> values = result.getValues();
        assertEquals(4, values.size());

        // assert given
        //noinspection unchecked
        assertEquals(
                relativeTypedInHolder.getGivens().get(0),
                ((JAXBElement<EnGiven>) ((PN) values.get(0)).getContent().get(0)).getValue().getContent().get(0));

        // assert family
        //noinspection unchecked
        assertEquals(
                relativeTypedInHolder.getFamily(),
                ((JAXBElement<EnFamily>) ((PN) values.get(0)).getContent().get(1)).getValue().getContent().get(0));

        // assert relation
        assertEquals(relativeTypedInHolder.getRelation().getCode(), ((CD) values.get(1)).getCode());
        assertEquals(relativeTypedInHolder.getRelation().getDisplayName(), ((CD) values.get(1)).getDisplayName());

        // Phone
        assertEquals("tel:911", ((TEL)values.get(2)).getValue());
        assertEquals("WP", ((TEL)values.get(2)).getUses().get(0));

        // assert note
        assertEquals(relativeTypedInHolder.getNote(), ((ST) values.get(3)).getContent().get(0));

        assertEquals(1, result.getAuthors().size());

        POCDMT000040Author author = result.getAuthors().get(0);

        // assert author time
        assertEquals(
                PDC_2_0ElementsMapper.dateTimeFormatter.format(relativeTypedInHolder.getAuthorTime()),
                author.getTime().getValue());

        // assert author cpr
        assertEquals(
                "ANONYM",
                author.getAssignedAuthor().getIds().get(0).getExtension());

        // assert author givens
        //noinspection unchecked
        assertTrue(
                ((JAXBElement<EnGiven>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(0)).getValue().getContent()
                        .containsAll(relativeTypedInHolder.getAuthorGivens()));

        // assert author family
        //noinspection unchecked
        assertEquals(
                relativeTypedInHolder.getAuthorFamily(),
                ((JAXBElement<EnFamily>) author.getAssignedAuthor().getAssignedPerson().getNames().get(0).getContent().get(1)).getValue().getContent().get(0));



        testMarshalling(result);
    }

    @Test
    public void createRelativeTypedInObservationFamilyNameNullAccepted() {
        // Skal acceptere manglende efternavn, selvom det afviger fra MedCom standarden, iht. [SDS-5182]

        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMockNoFamilyName();

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        assertNotNull(result);
        List<ANY> values = result.getValues();
        assertEquals("",((JAXBElement<EnFamily>) ((PN) values.get(0)).getContent().get(1)).getValue().getContent().get(0));
    }

    @Test
    public void createDentistTypedInObservationWithPhoneWithWhitespace() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getPhone().get().setValue("tel:+45 11 22 33 44");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(3)).getValue();
        assertEquals("tel:+4511223344", phoneNumber);
    }

    @Test
    public void createDentistTypedInObservationWithPhoneWithWhitespace2() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getPhone().get().setValue("tel:11 22 33 44");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(3)).getValue();
        assertEquals("tel:11223344", phoneNumber);
    }

    @Test
    public void createDentistTypedInObservationWithPhoneWithoutPrefixWithWhitespace() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getPhone().get().setValue("11 22 33 44");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(3)).getValue();
        assertEquals("tel:11223344", phoneNumber);
    }

    @Test
    public void createDentistTypedInObservationWithInvalidPhoneWithWhitespace() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getPhone().get().setValue("tel:+4511223344 88 88 88 88");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(3)).getValue();
        assertEquals("tel:+451122334488888888", phoneNumber);
    }

    @Test
    public void createDentistTypedInObservationWithInvalidPhoneWithoutPrefixWithWhitespace() {
        DentistTypedInHolder dentistTypedInHolder = createDentistTypedInHolderMock();
        dentistTypedInHolder.getPhone().get().setValue("+4511223344 88 88 88 88");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createDentistTypedInObservation(dentistTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(3)).getValue();
        assertEquals("tel:+451122334488888888", phoneNumber);
    }

    @Test
    public void createRelativeTypedInObservationWithPhoneWithWhitespace() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();
        relativeTypedInHolder.getPhoneHolders().get(0).setValue("tel:+45 1122 3344");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(2)).getValue();
        assertEquals("tel:+4511223344", phoneNumber);
    }

    @Test
    public void createRelativeTypedInObservationWithPhoneWithWhitespace2() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();
        relativeTypedInHolder.getPhoneHolders().get(0).setValue("tel:11 22 33 44");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(2)).getValue();
        assertEquals("tel:11223344", phoneNumber);
    }

    @Test
    public void createRelativeTypedInObservationWithPhoneWithoutPrefixWithWhitespace() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();
        relativeTypedInHolder.getPhoneHolders().get(0).setValue("11 22 33 44");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(2)).getValue();
        assertEquals("tel:11223344", phoneNumber);
    }

    @Test
    public void createRelativeTypedInObservationWithInvalidPhoneWithWhitespace() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();
        relativeTypedInHolder.getPhoneHolders().get(0).setValue("tel:45 88 88 88 88");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(2)).getValue();
        assertEquals("tel:4588888888", phoneNumber);
    }

    @Test
    public void createRelativeTypedInObservationWithInvalidPhoneWithoutPrefixWithWhitespace() {
        RelativeTypedInHolder relativeTypedInHolder = createRelativeTypedInHolderMock();
        relativeTypedInHolder.getPhoneHolders().get(0).setValue("45 88 88 88 88");

        POCDMT000040Observation result = PDC_2_0ClinicalDocumentAssembler.createRelativeTypedInObservation(relativeTypedInHolder);

        String phoneNumber = ((TEL) result.getValues().get(2)).getValue();
        assertEquals("tel:4588888888", phoneNumber);
    }

    private <T> void testMarshalling(T element) {
        try {
            //noinspection unchecked
            JAXBElement<T> jaxbElement =
                    new JAXBElement<>(new QName(
                            "uri", "local"),
                            (Class<T>) element.getClass(),
                            element);
            marshaller.marshal(jaxbElement, handler);
        } catch (JAXBException e) {
            throw new RuntimeException(e);
        }
    }

    private static CustodyHolder createCustodyHolderMock() {
        CustodyHolder custodyHolder = new CustodyHolder();

        custodyHolder.setCode(createCodeHolderMock());
        custodyHolder.setCpr("test1");
        custodyHolder.addGiven("test2");
        custodyHolder.setFamily("test3");
        custodyHolder.setRelation(createCodeHolderMock());

        return custodyHolder;
    }

    private static PrimaryPersonInformationHolder createPrimaryPersonInformationHolderMock() {
        PrimaryPersonInformationHolder personHolder = new PrimaryPersonInformationHolder();

        personHolder.setCpr("test1");
        personHolder.addGiven("test2");
        personHolder.setFamily("test3");
        personHolder.setGender("test4");
        personHolder.setBirthDate(ZonedDateTime.now());
        personHolder.setAddress(createAddressHolderMock());

        return personHolder;
    }

    private static ProviderOrganizationHolder createProviderOrganizationHolderMock() {
        ProviderOrganizationHolder providerOrganizationHolder = new ProviderOrganizationHolder();

        providerOrganizationHolder.setYderNr(new YderNr("test1"));
        providerOrganizationHolder.setName("test3");
        providerOrganizationHolder.addPhone(createPhoneHolderMock());
        providerOrganizationHolder.addEmail("test4");
        providerOrganizationHolder.setAddress(createAddressHolderMock());

        return providerOrganizationHolder;
    }

    private static CoverageGroupHolder createCoverageGroupHolderMock() {
        CoverageGroupHolder coverageGroupHolder = new CoverageGroupHolder();

        coverageGroupHolder.setGroup("2");

        return coverageGroupHolder;
    }

    private static OrganDonorRegistrationHolder createOrganDonorRegistrationHolderMock() {
        OrganDonorRegistrationHolder organDonorRegistrationHolder = new OrganDonorRegistrationHolder();

        organDonorRegistrationHolder.setHasRegistration(false);

        return organDonorRegistrationHolder;
    }

    private static TreatmentWillRegistrationHolder createTreatmentWillRegistrationHolderMock() {
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = new TreatmentWillRegistrationHolder();

        treatmentWillRegistrationHolder.setHasRegistration(false);

        return treatmentWillRegistrationHolder;
    }

    private static LivingWillRegistrationHolder createLivingWillRegistrationHolderMock() {
        LivingWillRegistrationHolder livingWillRegistrationHolder = new LivingWillRegistrationHolder();

        livingWillRegistrationHolder.setHasRegistration(false);

        return livingWillRegistrationHolder;
    }

    private static LanguageTypedInHolder createLanguageTypedInHolderMock() {
        LanguageTypedInHolder languageTypedInHolder = new LanguageTypedInHolder(new IdHolder("1.2.208.184.16.1.10.20.1.20", "2019-08-14", "FSK"));

        languageTypedInHolder.setLanguageCode("da");
        languageTypedInHolder.setLanguageName("Danish");
        languageTypedInHolder.setAuthorTime(ZonedDateTime.now());
        languageTypedInHolder.setAuthorCpr("abc");
        languageTypedInHolder.addAuthorGiven("def");
        languageTypedInHolder.setAuthorFamily("ghi");

        return languageTypedInHolder;
    }

    private static TempAddrTypedInHolder createTempAddrTypedInHolderMock() {
        TempAddrTypedInHolder tempAddrTypedInHolder = new TempAddrTypedInHolder(new IdHolder("1.2.208.184", "3a96fe45-0074-44e1-a2c0-ae5efc49db69", "FSK"));
        tempAddrTypedInHolder.setPeriodFrom(LocalDate.now().minusDays(5));
        tempAddrTypedInHolder.setPeriodTo(LocalDate.now().plusDays(5));
        tempAddrTypedInHolder.setAddress(createAddressHolderMock());
        tempAddrTypedInHolder.setAuthorTime(ZonedDateTime.now());
        tempAddrTypedInHolder.setAuthorCpr("abc");
        tempAddrTypedInHolder.addAuthorGiven("def");
        tempAddrTypedInHolder.setAuthorFamily("ghi");

        return tempAddrTypedInHolder;
    }

    private static DentistTypedInHolder createDentistTypedInHolderMock() {
        DentistTypedInHolder dentistTypedInHolder = new DentistTypedInHolder(new IdHolder("1.2.208.184.15.13", "168d430f-2b23-4679-a426-37db553926d0", "FSK"));

        dentistTypedInHolder.setPrefix("Tandlæge");
        dentistTypedInHolder.addGiven("test123");
        dentistTypedInHolder.setFamily("test456");
        dentistTypedInHolder.setAddress(createAddressHolderMock());
        dentistTypedInHolder.setPhone(createWorkPhoneHolderMock());
        dentistTypedInHolder.setAuthorTime(ZonedDateTime.now());
        dentistTypedInHolder.addAuthorGiven("def");
        dentistTypedInHolder.setAuthorFamily("ghi");
        dentistTypedInHolder.setYder("YDER");

        return dentistTypedInHolder;
    }

    private static PatientContactTypedInHolder createPatientContactTypedInHolderMock() {
        PatientContactTypedInHolder patientContactTypedInHolder = new PatientContactTypedInHolder();

        patientContactTypedInHolder.addPhone(createPhoneHolderMock());
        patientContactTypedInHolder.addEmailAddress(createPhoneHolderMock());
        patientContactTypedInHolder.setAuthorTime(ZonedDateTime.now());
        patientContactTypedInHolder.setAuthorCpr("abc");
        patientContactTypedInHolder.addAuthorGiven("def");
        patientContactTypedInHolder.setAuthorFamily("ghi");

        return patientContactTypedInHolder;
    }

    private static RelativeTypedInHolder createRelativeTypedInHolderMock() {
        RelativeTypedInHolder relativeTypedInHolder = new RelativeTypedInHolder(new IdHolder("1.2.208.184.15.3", "b9b38b0f-5e00-475b-85b3-e6caf86d270f", "FSK"));
        relativeTypedInHolder.addGiven("some given name");
        relativeTypedInHolder.setFamily("some family name");
        relativeTypedInHolder.setRelation(createCodeHolderMock());
        relativeTypedInHolder.setNote("some note");
        relativeTypedInHolder.setAuthorTime(ZonedDateTime.now());
        relativeTypedInHolder.setAuthorCpr("abc");
        relativeTypedInHolder.addAuthorGiven("def");
        relativeTypedInHolder.setAuthorFamily("ghi");
        PhoneHolder phone = new PhoneHolder();
        phone.addUse("WP");
        phone.setValue("911");
        relativeTypedInHolder.getPhoneHolders().add(phone);

        return relativeTypedInHolder;
    }

    private static RelativeTypedInHolder createRelativeTypedInHolderMockNoFamilyName() {
        RelativeTypedInHolder relativeTypedInHolder = new RelativeTypedInHolder(new IdHolder("1.2.208.184.15.3", "b9b38b0f-5e00-475b-85b3-e6caf86d270f", "FSK"));
        relativeTypedInHolder.addGiven("some given name");
        relativeTypedInHolder.setFamily(null);
        relativeTypedInHolder.setRelation(createCodeHolderMock());
        relativeTypedInHolder.setNote("some note");
        relativeTypedInHolder.setAuthorTime(ZonedDateTime.now());
        relativeTypedInHolder.setAuthorCpr("abc");
        relativeTypedInHolder.addAuthorGiven("def");
        relativeTypedInHolder.setAuthorFamily("ghi");
        PhoneHolder phone = new PhoneHolder();
        phone.addUse("WP");
        phone.setValue("911");
        relativeTypedInHolder.getPhoneHolders().add(phone);

        return relativeTypedInHolder;
    }

    private static AddressHolder createAddressHolderMock() {
        AddressHolder addressHolder = new AddressHolder();

        addressHolder.setUse("test1");
        addressHolder.addAddressLine("test2");
        addressHolder.setPostalCode("test3");
        addressHolder.setCity("test4");
        addressHolder.setCountry("test5");

        return addressHolder;
    }

    private static PhoneHolder createPhoneHolderMock() {
        PhoneHolder phoneHolder = new PhoneHolder();

        phoneHolder.addUse("H");
        phoneHolder.setValue("tel:+4512345678");

        return phoneHolder;
    }

    private static PhoneHolder createWorkPhoneHolderMock() {
        PhoneHolder phoneHolder = new PhoneHolder();

        phoneHolder.addUse("WP");
        phoneHolder.setValue("tel:+4512345678");

        return phoneHolder;
    }
    private static CodeHolder createCodeHolderMock() {
        CodeHolder codeHolder = new CodeHolder();

        codeHolder.setCode("abc");
        codeHolder.setDisplayName("def");

        return codeHolder;
    }

    public static RegisterData createRegisterData(String cpr) throws DatatypeConfigurationException {
        return createRegisterData(cpr, empty());
    }

    public static RegisterData createRegisterData(String cpr, boolean hasNoResuscitation)  throws DatatypeConfigurationException {
        return createRegisterData(cpr, of(hasNoResuscitation));
    }

    public static RegisterData createRegisterData(String cpr, Optional<Boolean> hasNoResuscitation)  throws DatatypeConfigurationException {
        RegisterData registerData = new RegisterData(cpr);

        RegularCPRPersonType regularCPRPersonType = new RegularCPRPersonType();
        regularCPRPersonType.setSimpleCPRPerson(new SimpleCPRPersonType());
        regularCPRPersonType.getSimpleCPRPerson().setPersonNameStructure(new PersonNameStructureType());

        regularCPRPersonType.setPersonBirthDateStructure(new PersonBirthDateStructureType());
        XMLGregorianCalendar xmlCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(new GregorianCalendar());
        xmlCal.setYear(2019);
        xmlCal.setMonth(1);
        xmlCal.setDay(1);
        xmlCal.setHour(0);
        xmlCal.setMinute(0);
        xmlCal.setSecond(0);
        regularCPRPersonType.getPersonBirthDateStructure().setBirthDate(xmlCal);

        PersonAddressStructureType personAddressStructureType = new PersonAddressStructureType();
        personAddressStructureType.setAddressComplete(new AddressCompleteType());
        personAddressStructureType.getAddressComplete().setAddressPostal(new AddressPostalType());
        personAddressStructureType.getAddressComplete().getAddressPostal().setCountryIdentificationCode(new CountryIdentificationCodeType());


        GetPersonWithHealthCareInformationOut scesData = new GetPersonWithHealthCareInformationOut();
        scesData.setPersonWithHealthCareInformationStructure(new PersonWithHealthCareInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().setPersonInformationStructure(new PersonInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().setRegularCPRPerson(regularCPRPersonType);
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().setPersonAddressStructure(personAddressStructureType);
        scesData.getPersonWithHealthCareInformationStructure().setPersonHealthCareInformationStructure(new PersonHealthCareInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure().setPersonPublicHealthInsurance(new PersonPublicHealthInsuranceType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance().setPublicHealthInsuranceStatus(PublicHealthInsuranceStatusType.D);
        scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance().setPublicHealthInsuranceGroupIdentifier(PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_1);
        scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance().setPublicHealthInsuranceGroupStartDate(xmlCal);

        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().getCountryIdentificationCode().setValue("DK");
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().setPostCodeIdentifier("8000");
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().setDistrictName("Aarhus");

        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().getSimpleCPRPerson().setPersonNameStructure(new PersonNameStructureType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().setPersonGivenName("Nielsen");
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().setPersonSurnameName("Nielsen");

        registerData.setScesData(scesData);
        registerData.setHasOrganDonorRegistration(TRUE);
        registerData.setHasTreatmentWill(TRUE);
        registerData.setHasLivingWill(TRUE);
        registerData.setHasNoResuscitation(hasNoResuscitation);

        return registerData;
    }
}