/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler;

import dk.stamkort.conf.FskConfig;
import dk.stamkort.conf.TestDAOConfig;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.util.FSKException;
import dk.stamkort.util.Properties;
import dk.stamkort.webservice.hl7cda.PdcDataMapperV3;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import org.hl7.v3.II;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeConfigurationException;
import java.util.List;
import java.util.stream.Collectors;

import static dk.stamkort.webservice.hl7cda.assembler.PDC_2_0ClinicalDocumentAssemblerTest.createRegisterData;
import static dk.stamkort.webservice.hl7cda.assembler.PDC_3_0ClinicalDocumentAssembler.assembleClinicalDocument;
import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { TestDAOConfig.class, FskConfig.class })
public class PDC_3_0ClinicalDocumentAssemblerTest {
    @Autowired
    Properties properties;

    @Test
    public void givenClinicalDocumentV3WhenCreatedFromSkrResponseThenThereShouldNotBeAnyDuplicateObservations()
            throws JAXBException, DatatypeConfigurationException {
        POCDMT000040ClinicalDocument result =
                getDocument("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3.xml");
        List<String> allTemplateRootIds = result.getComponent().getStructuredBody().getComponents().stream()
                .flatMap(c -> c.getSection().getEntries().stream())
                .flatMap(pocdmt000040Entry -> pocdmt000040Entry.getObservation().getTemplateIds().stream())
                .map(II::getRoot)
                .collect(Collectors.toList());
        List<String> uniqueTemplateRootIds = allTemplateRootIds.stream().distinct().collect(Collectors.toList());
        assertEquals(uniqueTemplateRootIds.size(), allTemplateRootIds.size());
    }

    @Test
    // AssignedPerson is not optional on Citizen
    public void givenClinicalDocumentV3_missingAssignedPersonOnCitizen_FSKException()
            throws JAXBException, DatatypeConfigurationException {
        try {
            POCDMT000040ClinicalDocument result =
                    getDocument("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3_Missing_AssignedPerson.xml");
            List<String> allTemplateRootIds = result.getComponent().getStructuredBody().getComponents().stream()
                    .flatMap(c -> c.getSection().getEntries().stream())
                    .flatMap(pocdmt000040Entry -> pocdmt000040Entry.getObservation().getTemplateIds().stream())
                    .map(II::getRoot)
                    .collect(Collectors.toList());
            List<String> uniqueTemplateRootIds = allTemplateRootIds.stream().distinct().collect(Collectors.toList());
            assertEquals(uniqueTemplateRootIds.size(), allTemplateRootIds.size());
        } catch (FSKException e) {
            assertNotNull(e);
            assertEquals("201: Manglende data i datakilde for stamkortet", e.getMessage());
        }
    }

    @Test
    // AssignedPerson is optional on HealthcareProfessional
    public void givenClinicalDocumentV3_missingAssignedPersonOnHealthcareProfessional_FSKException()
            throws JAXBException, DatatypeConfigurationException {
        POCDMT000040ClinicalDocument result =
                getDocument("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3_Healthcareprofessional_Missing_AssignedPerson.xml");
        List<String> allTemplateRootIds = result.getComponent().getStructuredBody().getComponents().stream()
                .flatMap(c -> c.getSection().getEntries().stream())
                .flatMap(pocdmt000040Entry -> pocdmt000040Entry.getObservation().getTemplateIds().stream())
                .map(II::getRoot)
                .collect(Collectors.toList());
        List<String> uniqueTemplateRootIds = allTemplateRootIds.stream().distinct().collect(Collectors.toList());
        assertEquals(uniqueTemplateRootIds.size(), allTemplateRootIds.size());
    }

    private POCDMT000040ClinicalDocument getDocument(String skrResponsePath)
            throws JAXBException, DatatypeConfigurationException {
        GetPersonalDataCardResponse personalDataCard = (GetPersonalDataCardResponse)
                JAXBContext.newInstance(GetPersonalDataCardResponse.class)
                        .createUnmarshaller()
                        .unmarshal(getClass().getResourceAsStream(skrResponsePath));

        String cprNr = personalDataCard.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0)
                .getExtension();
        RegisterData registerData = createRegisterData(cprNr);
        registerData.setClinicalDocument(personalDataCard.getClinicalDocument());
        return assembleClinicalDocument(PdcDataMapperV3.mapData(registerData), properties.createAuthorProperties());
    }
}