/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.stamkort.conf.FskConfig;
import dk.stamkort.conf.TestDAOConfig;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.*;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.util.FSKException;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.webservice.hl7cda.ClinicalDocumentProvider;
import dk.stamkort.webservice.hl7cda.ClinicalDocumentProviderImpl;
import dk.stamkort.webservice.hl7cda.assembler.PDC_2_0ClinicalDocumentAssemblerTest;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import oio.medcom.cprservice._1_0.CurrentGeneralPractitionerStructureType;
import org.hl7.v3.*;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.math.BigInteger;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static dk.stamkort.webservice.hl7cda.assembler.PDC_3_0ClinicalDocumentAssembler.*;
import static java.util.Optional.*;
import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { TestDAOConfig.class, FskConfig.class })
public class DocumentProviderTest {

    @Autowired
    Properties properties;

    @InjectMocks
    DocumentProvider provider;

    @Mock
    RegistryIndexDAO registryIndexDAO;

    @Mock
    ClinicalDocumentProvider clinicalDocumentProvider;

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenV2ExtensionShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV2Document(
                        "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional" +
                                ".xml",
                        properties);
        NamedNodeMap documentTemplateIdAttrs =
                getNode(document, "ClinicalDocument", "templateId").get().getAttributes();
        assertEquals("1.2.208.184.16.1", documentTemplateIdAttrs.getNamedItem("root").getNodeValue());
        assertNull(documentTemplateIdAttrs.getNamedItem("extension"));
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenV3ExtensionShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV3Document(
                        "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional" +
                                ".xml",
                        properties);
        NamedNodeMap documentTemplateIdAttrs =
                getNode(document, "ClinicalDocument", "templateId").get().getAttributes();
        assertEquals("1.2.208.184.16.1", documentTemplateIdAttrs.getNamedItem("root").getNodeValue());
        assertEquals("3.0", documentTemplateIdAttrs.getNamedItem("extension").getNodeValue());
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenV3ExtensionShouldBeInDocument_missingFamilyName()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV3Document(
                        "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional_missingFamilyName" +
                                ".xml",
                        properties);
        NamedNodeMap documentTemplateIdAttrs =
                getNode(document, "ClinicalDocument", "templateId").get().getAttributes();
        assertEquals("1.2.208.184.16.1", documentTemplateIdAttrs.getNamedItem("root").getNodeValue());
        assertEquals("3.0", documentTemplateIdAttrs.getNamedItem("extension").getNodeValue());
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenCorrectOidsShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document = getV3Document(
                "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional" +
                        ".xml",
                properties);
        NamedNodeMap documentTemplateIdAttrs =
                getNode(document, "ClinicalDocument", "templateId").get().getAttributes();
        assertEquals("1.2.208.184.16.1", documentTemplateIdAttrs.getNamedItem("root").getNodeValue());
        assertEquals("3.0", documentTemplateIdAttrs.getNamedItem("extension").getNodeValue());
    }

    @Test
    public void getDocument_WhereCprDoesNotMatch_ThrowsException() throws DAOException {
        RequestContextVO requestContextVO = mock(RequestContextVO.class);
        when(requestContextVO.getPatientInContext()).thenReturn("1122334488");
        RetrieveDocumentSetRequestType.DocumentRequest request = new RetrieveDocumentSetRequestType.DocumentRequest();
        request.setDocumentUniqueId("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000000");
        RegistryIndexVO registryIndex = new RegistryIndexVO();
        registryIndex.setPersonIdentifier("0000000000");
        when(registryIndexDAO.getRegistryIndexByDocumentId(request.getDocumentUniqueId()))
                .thenReturn(registryIndex);

        FSKException e = assertThrows(FSKException.class, () -> provider.getDocument(requestContextVO, request));
        assertTrue(e.getFaultInfo().contains("PatientId i sikkerhedsbilletten matcher ikke patientId i requestet."));
        assertEquals(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, e.getFaultCode());
    }

    @Test
    public void getDocument_WhereDocumentIsSuccessfullyRetrieved_ReturnsDocumentAndPatientId()
            throws DAOException, JAXBException {
        final String patientId = "1122334488", documentVersion = "urn:ad:dk:medcom:pdc-v2.0:full";
        RequestContextVO requestContextVO = mock(RequestContextVO.class);
        when(requestContextVO.getPatientInContext()).thenReturn(patientId);
        when(requestContextVO.getUserType()).thenReturn(RequestContextVO.UserType.SystemUser);
        RetrieveDocumentSetRequestType.DocumentRequest request = new RetrieveDocumentSetRequestType.DocumentRequest();
        request.setDocumentUniqueId("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000000");
        RegistryIndexVO registryIndex = new RegistryIndexVO();
        registryIndex.setPersonIdentifier(patientId);
        registryIndex.setFormatCode(documentVersion);
        when(registryIndexDAO.getRegistryIndexByDocumentId(request.getDocumentUniqueId()))
                .thenReturn(registryIndex);
        String document = "test";
        when(clinicalDocumentProvider.getClinicalDocument(patientId, requestContextVO, documentVersion))
                .thenReturn(document);

        Optional<DocumentResponse> response = provider.getDocument(requestContextVO, request);

        assertTrue(response.isPresent());
        assertEquals(patientId, response.get().getPatientIdentifier());
        assertEquals(document, new String(response.get().getDocument().getDocument()));
    }

    final static String
            expectedPracticeName = "laege praksis",
            expectedDentistName = "tandlaege navn",
            expectedV3PhoneNumber = "87654321";

    @Test
    public void givenRegisterDataWithSkrDentistResponseWhenCallingGetClinicalDocumentV2ThenCorrectPhoneNumberFromYderShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV2Document("/testfiles/GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                              properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node dentistObservation = findObservationWithTemplateRootId(section, dentistRootId).get();
        List<Node> values = new ArrayList<>();
        for (int i = 0; i < dentistObservation.getChildNodes().getLength(); i++) {
            Node item = dentistObservation.getChildNodes().item(i);
            if (item.getNodeName().equals("value")) { values.add(item); }
        }

        NamedNodeMap telValueAttrs = values.stream()
                .filter(attrs -> attrs.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("TEL"))
                .findFirst().get().getAttributes();
        assertEquals("WP", telValueAttrs.getNamedItem("use").getNodeValue());
        String numberInGetPersonalDataCardResponse = "12345678";
        assertEquals("tel:" + numberInGetPersonalDataCardResponse, telValueAttrs.getNamedItem("value").getNodeValue());
    }

    @Test
    public void givenRegisterDataWithSkrDentistResponseWhenCallingGetClinicalDocumentThenV3HealthProviderShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV3Document("/testfiles/GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                              properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node dentistObservation = findObservationWithTemplateRootId(section, dentistRootId).get();
        List<Node> values = new ArrayList<>();
        for (int i = 0; i < dentistObservation.getChildNodes().getLength(); i++) {
            Node item = dentistObservation.getChildNodes().item(i);
            if (item.getNodeName().equals("value")) { values.add(item); }
        }

        String practiceName = values.stream()
                .filter(attrs -> attrs.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("TN"))
                .findFirst().get().getTextContent();
        assertEquals(expectedPracticeName, practiceName);

        Node dentistName = values.stream()
                .filter(attrs -> attrs.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("PN"))
                .findFirst().get();
        assertEquals(expectedDentistName, getNode(dentistName, "given").get().getTextContent());
        assertEquals("NA",
                     getNode(dentistName, "family").get().getAttributes().getNamedItem("nullFlavor").getNodeValue());

        NamedNodeMap telValueAttrs = values.stream()
                .filter(v -> v.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("TEL"))
                .findFirst().get().getAttributes();
        assertEquals("WP", telValueAttrs.getNamedItem("use").getNodeValue());
        assertEquals("tel:" + expectedV3PhoneNumber, telValueAttrs.getNamedItem("value").getNodeValue());
    }

    @Test
    public void givenRegisterDataWithSkrDentistResponseWithCorrespondingInvalidPhoneNumberInYderRegisterWhenCallingGetClinicalDocumentThenPhoneNumberShouldNotBePresent()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        String invalidPhoneNumber = "a3223";
        Document document =
                getV3Document(createRegisterData(
                                      "/testfiles/GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                                      invalidPhoneNumber),
                              properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node dentistObservation = findObservationWithTemplateRootId(section, dentistRootId).get();
        List<Node> values = new ArrayList<>();
        for (int i = 0; i < dentistObservation.getChildNodes().getLength(); i++) {
            Node item = dentistObservation.getChildNodes().item(i);
            if (item.getNodeName().equals("value")) { values.add(item); }
        }
        long phoneNumbersCount = values.stream()
                .filter(v -> v.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("TEL")).count();
        assertEquals(0, phoneNumbersCount);
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenResuscitationShouldBeInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document =
                getV3Document(
                        "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional" +
                                ".xml",
                        properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        String resuscitationObservationRootId = "1.2.208.184.16.1.10.20.1.31";
        Node resuscitationObservation =
                findObservationWithTemplateRootId(section, resuscitationObservationRootId).get();

        NamedNodeMap observationAttrs = resuscitationObservation.getAttributes();
        assertEquals("OBS", observationAttrs.getNamedItem("classCode").getNodeValue());
        assertEquals("EVN", observationAttrs.getNamedItem("moodCode").getNodeValue());

        String templateIdExtension = getNode(resuscitationObservation, "templateId").get()
                .getAttributes().getNamedItem("extension").getNodeValue();
        assertEquals("2023-07-01", templateIdExtension);

        NamedNodeMap idAttrs = getNode(resuscitationObservation, "id").get().getAttributes();
        UUID shouldNotThrow = UUID.fromString(idAttrs.getNamedItem("extension").getNodeValue());
        assertEquals("1.2.208.184", idAttrs.getNamedItem("root").getNodeValue());

        NamedNodeMap code = getNode(resuscitationObservation, "code").get().getAttributes();
        assertEquals("NoResuscitationRegistration", code.getNamedItem("code").getNodeValue());
        assertEquals("1.2.208.184.100.1", code.getNamedItem("codeSystem").getNodeValue());
        assertEquals("MedCom Message Codes", code.getNamedItem("codeSystemName").getNodeValue());
        assertEquals("Registreret fravalg af genoplivningsforsøg v. hjertestop",
                     code.getNamedItem("displayName").getNodeValue());

        Node author = getNode(resuscitationObservation, "author").get();
        NamedNodeMap authorTemplateIdAttrs = getNode(author, "templateId").get().getAttributes();
        assertEquals("1.2.208.184.16.1.10.20.31", authorTemplateIdAttrs.getNamedItem("root").getNodeValue());
        assertEquals("2019-08-14", authorTemplateIdAttrs.getNamedItem("extension").getNodeValue());

        String authorTime = getNode(author, "time").get().getAttributes().getNamedItem("nullFlavor").getNodeValue();
        assertEquals("NI", authorTime);

        Node assignedAuthor = getNode(author, "assignedAuthor").get();
        NamedNodeMap assignedAuthorIdAttr = getNode(assignedAuthor, "id").get().getAttributes();
        assertEquals("Sundhedsdatastyrelsen",
                     assignedAuthorIdAttr.getNamedItem("assigningAuthorityName").getNodeValue());
        assertEquals("1.2.208.176.1.8", assignedAuthorIdAttr.getNamedItem("root").getNodeValue());
        assertEquals("NI", assignedAuthorIdAttr.getNamedItem("extension").getNodeValue());

        Node representedOrganizationName = getNode(assignedAuthor, "representedOrganization", "name").get();
        assertEquals("Sundhedsdatastyrelsen", representedOrganizationName.getTextContent());
    }

    @Test
    public void givenRegisterDataWithDifferentResuscitationValueWhenCallingGetClinicalDocumentThenCorrectValueShouldBeInDocument() {
        Stream.<Optional<Boolean>>of(of(true), of(false), empty(), null).forEach(expected -> {
            RegisterData registerData = createRegisterData(
                    "/testfiles/GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional.xml");
            registerData.setHasNoResuscitation(expected);
            Document document;
            try {
                document = getV3Document(registerData, properties);
            } catch (JAXBException | SAXException | IOException | ParserConfigurationException e) {
                throw new RuntimeException(e);
            }
            Node section =
                    getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
            String resuscitationObservationRootId = "1.2.208.184.16.1.10.20.1.31";
            Node resuscitationObservation =
                    findObservationWithTemplateRootId(section, resuscitationObservationRootId).get();
            NamedNodeMap valueAttrs = getNode(resuscitationObservation, "value").get().getAttributes();
            if(expected == null || !expected.isPresent()) {
                //SDS-7755: Fælles Stamkort skal returnere "false" i fravalgsoplysningen som default indtil der åbnes for registreringer
                //assertEquals("NAV", valueAttrs.getNamedItem("nullFlavor").getNodeValue());
                assertEquals("Sundhedsdatastyrelsen", valueAttrs.getNamedItem("assigningAuthorityName").getNodeValue());
                assertEquals("false", valueAttrs.getNamedItem("extension").getNodeValue());
                assertEquals("1.2.208.176.1.11", valueAttrs.getNamedItem("root").getNodeValue());
                assertEquals("II", valueAttrs.getNamedItem("xsi:type").getNodeValue());
            } else {
                assertEquals("Sundhedsdatastyrelsen", valueAttrs.getNamedItem("assigningAuthorityName").getNodeValue());
                assertEquals(expected.get() ? "true" : "false", valueAttrs.getNamedItem("extension").getNodeValue());
                assertEquals("1.2.208.176.1.11", valueAttrs.getNamedItem("root").getNodeValue());
                assertEquals("II", valueAttrs.getNamedItem("xsi:type").getNodeValue());
            }
        });
    }

    class ResponseAndRootId {
        public final String responseFilePath, rootId;

        ResponseAndRootId(String responseFilePath, String rootId) {
            this.responseFilePath = responseFilePath; this.rootId = rootId;
        }
    }

    @Test
    public void givenSkrResponsesWithDifferentManuallyEnteredInfoWhenCallingGetClinicalDocumentThenCorrectRepresentedOrgShouldBeInDocument() {
        String contactInfo = "GetPersonalDataCard_response_with_contact_info_enteredHealthcareprofessional.xml",
                dentist = "GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                language = "GetPersonalDataCard_response_with_language_enteredHealthcareprofessional.xml",
                relative = "GetPersonalDataCard_response_with_relative_enteredHealthcareprofessional.xml",
                address = "GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional.xml";
        Stream.of(new ResponseAndRootId(contactInfo, contactInfoRootId),
                  new ResponseAndRootId(dentist, dentistRootId),
                  new ResponseAndRootId(language, spokenLanguageRootId),
                  new ResponseAndRootId(relative, relativeRootId),
                  new ResponseAndRootId(address, temporaryAddressRootId))
                .forEach(skrResponseAndRootId -> {
                    try {
                        Document document =
                                getV3Document("/testfiles/" + skrResponseAndRootId.responseFilePath, properties);
                        Node section = getNode(document,
                                               "ClinicalDocument",
                                               "component",
                                               "structuredBody",
                                               "component",
                                               "section").get();
                        Node observation =
                                findObservationWithTemplateRootId(section, skrResponseAndRootId.rootId).get();
                        String representedOrgName = getNode(observation, "author",
                                                            "assignedAuthor",
                                                            "representedOrganization",
                                                            "name").get().getTextContent();
                        boolean isSundhedsdatastyrelsenInSkrResponse =
                                skrResponseAndRootId.rootId.equals(dentistRootId) ||
                                        skrResponseAndRootId.rootId.equals(temporaryAddressRootId);
                        String expected = isSundhedsdatastyrelsenInSkrResponse
                                ? "Sundhedsdatastyrelsen"
                                : "Organisation";
                        assertEquals(expected, representedOrgName);
                    } catch (JAXBException | IOException | ParserConfigurationException | SAXException e) {
                        throw new RuntimeException(e);
                    }
                });
    }

    @Test
    public void givenSkrResponsesWithDifferentManuallyEnteredInfoWhenCallingGetClinicalDocumentThenCorrectCodeShouldBeInDocument() {
        Stream.of(new ResponseAndRootId("GetPersonalDataCard_response_with_contact_info.xml", contactInfoRootId),
                  new ResponseAndRootId("GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                                        dentistRootId),
                  new ResponseAndRootId("GetPersonalDataCard_response_with_language.xml", spokenLanguageRootId),
                  new ResponseAndRootId("GetPersonalDataCard_response_with_relative.xml", relativeRootId),
                  new ResponseAndRootId(
                          "GetPersonalDataCard_response_with_temporary_address_enteredHealthcareprofessional.xml",
                          temporaryAddressRootId))
                .forEach(skrResponseAndRootId -> {
                    try {
                        Document document =
                                getV3Document("/testfiles/" + skrResponseAndRootId.responseFilePath, properties);
                        Node section = getNode(document,
                                               "ClinicalDocument",
                                               "component",
                                               "structuredBody",
                                               "component",
                                               "section").get();
                        String rootId = skrResponseAndRootId.rootId;
                        Node observation = findObservationWithTemplateRootId(section, rootId).get();
                        NamedNodeMap code =
                                getNode(observation, "author", "assignedAuthor", "code").get().getAttributes();
                        boolean isDentistDataEnterer = rootId.equals(dentistRootId) ||
                                rootId.equals(temporaryAddressRootId);
                        String expectedCode = isDentistDataEnterer ? "EnteredHealthcareprofessional" :
                                rootId.equals(relativeRootId) ? "EnteredCitizenRelative" :
                                        "EnteredCitizen",
                                expectedDisplayName = isDentistDataEnterer ? "Indtastet af sundhedsprofessionel" :
                                        rootId.equals(relativeRootId) ? "Indtastet af borgers pårørende" :
                                                "Indtastet af borger";
                        assertEquals(expectedCode, code.getNamedItem("code").getNodeValue());
                        assertEquals("1.2.208.184.100.1", code.getNamedItem("codeSystem").getNodeValue());
                        assertEquals("MedCom Message Codes", code.getNamedItem("codeSystemName").getNodeValue());
                        assertEquals(expectedDisplayName, code.getNamedItem("displayName").getNodeValue());
                    } catch (JAXBException | IOException | ParserConfigurationException | SAXException e) {
                        throw new RuntimeException(e);
                    }
                });
    }

    @Test
    public void givenSkrResponseWithAllManuallyEnteredInfoWhenCallingGetClinicalDocumentThenCorrectCodeAndRepresentedOrgShouldBeInDocument()
            throws JAXBException, ParserConfigurationException, IOException, SAXException {
        Document document = getV3Document("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3.xml", properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Stream<String> manuallyEnteredRootIds = Stream.of(spokenLanguageRootId,
                                                          temporaryAddressRootId,
                                                          dentistRootId,
                                                          contactInfoRootId,
                                                          relativeRootId);
        manuallyEnteredRootIds.map(id -> findObservationWithTemplateRootId(section, id).get())
                .forEach(observation -> {
                    boolean isDentistDataEnterer =
                            getNode(observation, "templateId").get().getAttributes().getNamedItem("root").getNodeValue()
                                    .equals(dentistRootId);
                    String expectedCode = isDentistDataEnterer
                            ? "EnteredHealthcareprofessional"
                            : "EnteredCitizen",
                            expectedDisplayName = isDentistDataEnterer
                                    ? "Indtastet af sundhedsprofessionel"
                                    : "Indtastet af borger";
                    NamedNodeMap code = getNode(observation, "author", "assignedAuthor", "code").get().getAttributes();
                    assertEquals(expectedCode, code.getNamedItem("code").getNodeValue());
                    assertEquals("1.2.208.184.100.1", code.getNamedItem("codeSystem").getNodeValue());
                    assertEquals("MedCom Message Codes", code.getNamedItem("codeSystemName").getNodeValue());
                    assertEquals(expectedDisplayName, code.getNamedItem("displayName").getNodeValue());

                    Optional<String> expected = of("Organisation").filter(u -> isDentistDataEnterer);
                    Optional<Node> representedOrgName = getNode(observation,
                                                                "author",
                                                                "assignedAuthor",
                                                                "representedOrganization",
                                                                "name");
                    expected.ifPresent(e -> assertEquals(e, representedOrgName.get().getTextContent()));
                });
    }

    @Test
    public void testDentistTypedInV3_allDentistInfoIsFromYder()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        // Arrange
        String phoneNumber = "4242";

        // Act
        Document document = getV3Document(createRegisterData(
                "/testfiles/GetPersonalDataCard_response_with_dentist_enteredHealthcareprofessional.xml",
                phoneNumber), properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node dentistObservation = findObservationWithTemplateRootId(section, dentistRootId).get();

        List<Node> values = new ArrayList<>();
        for (int i = 0; i < dentistObservation.getChildNodes().getLength(); i++) {
            Node item = dentistObservation.getChildNodes().item(i);
            if (item.getNodeName().equals("value")) {
                values.add(item);
            }
        }

        // Extract
        NamedNodeMap telValueAttrs = values.stream()
                .filter(attrs -> attrs.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("TEL"))
                .findFirst().get().getAttributes();
        Node ad = values.stream()
                .filter(attrs -> attrs.getAttributes().getNamedItem("xsi:type").getNodeValue().equals("AD"))
                .findFirst().get();
        String streetAddressLine = getNode(ad, "streetAddressLine").get().getTextContent();
        String postalCode = getNode(ad, "postalCode").get().getTextContent();
        String city = getNode(ad, "city").get().getTextContent();
        String country = getNode(ad, "country").get().getTextContent();

        // Assert
        assertEquals("WP", telValueAttrs.getNamedItem("use").getNodeValue());
        assertEquals("tel:" + phoneNumber, telValueAttrs.getNamedItem("value").getNodeValue());
        assertEquals("Adresse 1", streetAddressLine);
        assertEquals("1234", postalCode);
        assertEquals("City", city);
        assertEquals("DK", country);
    }

    @Test
    public void givenSkrResponseWitManuallyEnteredLanguageWhenCallingGetClinicalDocumentThenLanguageShoudBeTranslatedToDanishInDocument()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document = getV3Document("/testfiles/GetPersonalDataCard_response_with_language.xml", properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node languageObservation = findObservationWithTemplateRootId(section, spokenLanguageRootId).get();
        Node value = getNode(languageObservation, "value").get();
        String languageDisplayName = value.getAttributes().getNamedItem("displayName").getNodeValue();
        assertEquals("Tysk", languageDisplayName);
    }

    @Test
    public void givenSkrResponseWithAllManuallyEnteredInfoWhenCallingGetClinicalDocumentThenCorrectRootIdsShouldBeInObservation()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document = getV3Document("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3.xml", properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Stream<String> manuallyEnteredRootIds = Stream.of(spokenLanguageRootId,
                                                          temporaryAddressRootId,
                                                          dentistRootId,
                                                          contactInfoRootId,
                                                          relativeRootId);
        manuallyEnteredRootIds.forEach(templateId -> {
            Node observation = findObservationWithTemplateRootId(section, templateId).get();
            NamedNodeMap id = getNode(observation, "id").get().getAttributes();
            assertEquals(templateId, "1.2.208.184", id.getNamedItem("root").getNodeValue());
            assertNull(templateId, id.getNamedItem("assigningAuthorityName"));
            UUID shouldNotThrow = UUID.fromString(id.getNamedItem("extension").getNodeValue());
        });
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenCoverageValueShoudHaveCorrectRoot()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document = getV3Document("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3.xml", properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node coverageObservation = findObservationWithTemplateRootId(section, "1.2.208.184.16.1.10.20.1.27").get();
        Node value = getNode(coverageObservation, "value").get();
        assertEquals("1.2.208.176.2.7", value.getAttributes().getNamedItem("root").getNodeValue());
    }

    private static Optional<Node> findObservationWithTemplateRootId(Node node, String rootId) {
        for (int i = 0; i < node.getChildNodes().getLength(); i++) {
            Node n = node.getChildNodes().item(i).getChildNodes().item(0);
            Optional<Node> elementsUnderSection = ofNullable(n);
            boolean isFound = elementsUnderSection.flatMap(e -> getNode(e, "templateId")
                            .map(tid -> tid.getAttributes().getNamedItem("root"))
                            .filter(root -> root.getNodeValue().equals(rootId)))
                    .isPresent();
            if (isFound) return of(n);
        }
        return empty();
    }

    @Test
    public void givenMinimalRegisterDataWhenCallingGetClinicalDocumentThenAddressShouldHaveCorrectTime()
            throws JAXBException, IOException, ParserConfigurationException, SAXException {
        Document document = getV3Document("/testfiles/GetPersonalDataCard_response_Clinical_CDA_V3.xml", properties);
        Node section =
                getNode(document, "ClinicalDocument", "component", "structuredBody", "component", "section").get();
        Node addressObservation = findObservationWithTemplateRootId(section, temporaryAddressRootId).get();
        Node highTime = getNode(addressObservation, "effectiveTime", "high").get();
        assertEquals("NI", highTime.getAttributes().getNamedItem("nullFlavor").getNodeValue());
    }

    private static Optional<Node> getNode(Node parent, String... names) {
        return getNode(parent, Stream.of(names).collect(Collectors.toList()));
    }

    private static Optional<Node> getNode(Node parent, List<String> names) {
        if (names.isEmpty()) { return of(parent); }
        for (int i = 0; i < parent.getChildNodes().getLength(); i++) {
            Node child = parent.getChildNodes().item(i);
            if (names.get(0).equals(child.getNodeName())) {
                return getNode(child, names.subList(1, names.size()));
            }
        }
        return empty();
    }

    private static RegisterData createRegisterData(String skrResponsePath) {
        return createRegisterData(skrResponsePath, expectedV3PhoneNumber);
    }

    private static RegisterData createRegisterData(String skrResponsePath, String yderPhoneNumber) {
        InputStream stream = DocumentProviderTest.class.getResourceAsStream(skrResponsePath);
        GetPersonalDataCardResponse personalDataCard;
        try {
            personalDataCard = (GetPersonalDataCardResponse)
                    JAXBContext.newInstance(GetPersonalDataCardResponse.class)
                            .createUnmarshaller()
                            .unmarshal(stream);
        } catch (JAXBException e) { throw new RuntimeException(e); }

        POCDMT000040ClinicalDocument clinicalDocument = personalDataCard.getClinicalDocument();
        String cprNr = clinicalDocument.getRecordTargets().get(0).getPatientRole().getIds().get(0)
                .getExtension();
        RegisterData registerData;
        try {
            registerData = PDC_2_0ClinicalDocumentAssemblerTest.createRegisterData(cprNr);
        } catch (DatatypeConfigurationException e) { throw new RuntimeException(e); }
        YderNr generalPractitionerYderNr = new YderNr("321321");
        AssociatedGeneralPractitionerStructureType a = new AssociatedGeneralPractitionerStructureType();
        a.setAssociatedGeneralPractitionerIdentifier(new BigInteger(generalPractitionerYderNr.value));
        CurrentGeneralPractitionerStructureType b = new CurrentGeneralPractitionerStructureType();
        b.setAssociatedGeneralPractitionerStructure(a);
        registerData.getScesData().getPersonWithHealthCareInformationStructure()
                .getPersonHealthCareInformationStructure()
                .setCurrentGeneralPractitionerStructure(b);
        registerData.setClinicalDocument(clinicalDocument);

        GeneralPractitioner gp = new GeneralPractitioner();
        gp.setYdernr(generalPractitionerYderNr);
        gp.setAddress("Addresses 2");
        gp.setPostcode("4321");
        gp.setCity("City 2");
        gp.setOrganizationName("general practioner org name");
        gp.setPersonName("læge navn");
        gp.setPhone("12345678");
        registerData.setGeneralPractitioner(gp);

        dentistFromDocument(yderPhoneNumber, clinicalDocument).ifPresent(registerData::setDentist);

        return registerData;
    }

    private static Optional<Dentist> dentistFromDocument(String yderPhoneNumber,
                                                         POCDMT000040ClinicalDocument clinicalDocument) {
        return of(clinicalDocument).map(POCDMT000040ClinicalDocument::getComponent)
                .map(POCDMT000040Component2::getStructuredBody)
                .flatMap(pocdmt000040StructuredBody -> pocdmt000040StructuredBody.getComponents().stream().findFirst())
                .map(POCDMT000040Component3::getSection)
                .flatMap(pocdmt000040Section -> pocdmt000040Section.getEntries().stream().findFirst())
                .map(POCDMT000040Entry::getHealthProvider)
                .map(HealthProvider::getId)
                .map(II::getExtension)
                .map(dentistYderNr -> {
                    Dentist dentist = new Dentist();
                    dentist.setYdernr(new YderNr(dentistYderNr));
                    dentist.setAddress("Adresse 1");
                    dentist.setPostcode("1234");
                    dentist.setCity("City");
                    dentist.setOrganizationName(expectedPracticeName);
                    dentist.setPersonName(expectedDentistName);
                    dentist.setPhone(yderPhoneNumber);
                    return dentist;
                });
    }

    private static Document getV2Document(String skrResponseFilePath, dk.stamkort.util.Properties properties)
            throws JAXBException, SAXException, IOException, ParserConfigurationException {
        return getDocument("urn:ad:dk:medcom:pdc-v2.0:full", createRegisterData(skrResponseFilePath), properties);
    }

    private static Document getV3Document(String skrResponseFilePath, dk.stamkort.util.Properties properties)
            throws JAXBException, SAXException, IOException, ParserConfigurationException {
        return getV3Document(createRegisterData(skrResponseFilePath), properties);
    }

    private static Document getV3Document(RegisterData rd, dk.stamkort.util.Properties properties)
            throws JAXBException, SAXException, IOException, ParserConfigurationException {
        return getDocument("urn:ad:dk:medcom:pdc-v3.0:full", rd, properties);
    }

    private static Document getDocument(String version, RegisterData rd, dk.stamkort.util.Properties properties)
            throws JAXBException, SAXException, IOException, ParserConfigurationException {
        String clinicalDocument = new ClinicalDocumentProviderImpl((u1, u2) -> rd, properties)
                .getClinicalDocument("1234567890", new RequestContextVO(), version);
        return DocumentBuilderFactory.newInstance().newDocumentBuilder()
                .parse(new InputSource(new StringReader(clinicalDocument)));
    }
}
