/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.stamkort.OID;
import dk.stamkort.domain.YderNr;
import dk.stamkort.domain.YderVO;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import org.apache.commons.io.IOUtils;
import org.hl7.v3.*;
import org.springframework.util.ResourceUtils;

import javax.xml.bind.JAXBElement;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.List;

public class DocumentProviderTestData {
    public static final String stubbedCPRE(String ydernr) throws IOException {
    	return stubbedCPRE(ydernr, "01");
    }

    public static final String stubbedCPREMissingFamilyName(String ydernr) throws IOException {
        return stubbedCPREMissingFamilyName(ydernr, "01");
    }

    public static final String stubbedCPREMissingGivenName(String ydernr) throws IOException {
        return stubbedCPREMissingGivenName(ydernr, "01");
    }

    public static final String stubbedCPRE(String ydernr, String status) throws IOException {
        File file = ResourceUtils.getFile("classpath:mock/sces-1.0.4-body.xml");
        FileInputStream fis = new FileInputStream(file);
        String data = IOUtils.toString(fis, StandardCharsets.UTF_8);

        return data.replace("####YDERNR####", ydernr).replace("####STATUS####", status);
    }

    public static final String stubbedCPREMissingFamilyName(String ydernr, String status) throws IOException {
        File file = ResourceUtils.getFile("classpath:mock/sces-1.0.4-body-missingFamilyName.xml");
        FileInputStream fis = new FileInputStream(file);
        String data = IOUtils.toString(fis, StandardCharsets.UTF_8);

        return data.replace("####YDERNR####", ydernr).replace("####STATUS####", status);
    }

    public static final String stubbedCPREMissingGivenName(String ydernr, String status) throws IOException {
        File file = ResourceUtils.getFile("classpath:mock/sces-1.0.4-body-missingGivenName.xml");
        FileInputStream fis = new FileInputStream(file);
        String data = IOUtils.toString(fis, StandardCharsets.UTF_8);

        return data.replace("####YDERNR####", ydernr).replace("####STATUS####", status);
    }

    public static GetPersonalDataCardResponse createPersonalDataCardResponse(String ydernr) {
        GetPersonalDataCardResponse response = new GetPersonalDataCardResponse();
        response.setClinicalDocument(createDocument(ydernr));
        return response;
    }

    private static POCDMT000040ClinicalDocument createDocument(String ydernr) {
        final POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();
        final II templateId = new II();
        templateId.setRoot("1.2.208.184.15.1");
        doc.getTemplateIds().add(templateId);
        doc.setClassCode(ActClinicalDocument.DOCCLIN);
        doc.getMoodCodes().add("EVN");
        final CS dk = new CS();
        dk.setCode("DK");
        doc.getRealmCodes().add(dk);
        doc.setComponent(createComponent(ydernr));
        return doc;
    }

    private static POCDMT000040Component2 createComponent(String ydernr) {
        final POCDMT000040Component2 component = new POCDMT000040Component2();
        final POCDMT000040StructuredBody structuredBody = new POCDMT000040StructuredBody();
        final POCDMT000040Component3 bodyComponent = new POCDMT000040Component3();
        final POCDMT000040Section section = new POCDMT000040Section();
        final StrucDocText fsk = new StrucDocText();
        fsk.getContent().add("FSK");
        section.setText(fsk);
        section.getEntries().add(createHealthProviderEntry(ydernr));
        bodyComponent.setSection(section);
        structuredBody.getComponents().add(bodyComponent);
        component.setStructuredBody(structuredBody);
        return component;
    }

    private static POCDMT000040Entry createHealthProviderEntry(String ydernr) {
        final POCDMT000040Entry entry = new POCDMT000040Entry();

        final HealthProvider healthProvider = new HealthProvider();
        healthProvider.setId(createId("f7ef1b08-78c2-4d6b-9158-6d17e2319a69"));
        healthProvider.setOrganization(createOrganization(ydernr));
        healthProvider.setDataEnterer(new POCDMT000040Author());
        healthProvider.getDataEnterer().setAssignedAuthor(new POCDMT000040AssignedAuthor());
        healthProvider.getDataEnterer().getAssignedAuthor().setAssignedPerson(new POCDMT000040Person());

        List<String> givens = Arrays.asList("Hans");
        healthProvider.getDataEnterer().getAssignedAuthor().setAssignedPerson(createAssignedPerson(createPN(givens, "Hansen")));

        healthProvider.getDataEnterer().setTime(new TS());
        healthProvider.getDataEnterer().getTime().setValue("20180704131302+0200");


        entry.setHealthProvider(healthProvider);

        return entry;
    }

    private static POCDMT000040Organization createOrganization(String ydernr) {
        POCDMT000040Organization organization = o.createPOCDMT000040Organization();

        organization.getNames().add(createOrganisationName("OrgName From Register"));
        organization.getAddrs().add(createAddress("Vej 42", "8000", "Aarhus C", "DK"));

        if (ydernr != null) {
            organization.getIds().add(createId(ydernr));
        }

        TEL phone = o.createTEL();
        phone.setValue("tel:11223344");
        phone.getUses().add("WP");
        organization.getTelecoms().add(phone);

        return organization;
    }

    public static II createId(String extension) {
        final II id = o.createII();
        id.setRoot(OID.YDERNUMMER);
        id.setExtension(extension);
        id.setAssigningAuthorityName("Yder");
        return id;
    }

    private static ObjectFactory o = new ObjectFactory();

    public static AD createAddress(String addressLine1, String postalCode, String city, String country) {
        AD address = o.createAD();
        address.setIsNotOrdered(false);
        AdxpStreetAddressLine adxpStreetAddressLine = o.createAdxpStreetAddressLine();
        adxpStreetAddressLine.getContent().add(addressLine1);
        address.getContent().add(o.createADStreetAddressLine(adxpStreetAddressLine));
        AdxpPostalCode adxpPostalCode = o.createAdxpPostalCode();
        adxpPostalCode.getContent().add(postalCode);
        address.getContent().add(o.createADPostalCode(adxpPostalCode));
        AdxpCity adxpCity = o.createAdxpCity();
        adxpCity.getContent().add(city);
        address.getContent().add(o.createADCity(adxpCity));
        address.getUses().add("WP");
        AdxpCountry adxpCountry = o.createAdxpCountry();
        adxpCountry.getContent().add(country);
        address.getContent().add(o.createADCountry(adxpCountry));
        return address;
    }
    public static ON createOrganisationName(String text) {
        final ON name = new ON();
        name.getContent().add(text);
        return name;
    }

    public static YderVO createYderData(String yderNr) {
        YderVO yder = new YderVO();
        yder.setYdernr(new YderNr(yderNr));
        yder.setOrganizationName("Testyder");
        yder.setAddress("Ydervej 42");
        yder.setPostcode("1234");
        yder.setCity("Yderby");
        yder.setPhone("12345678");
        yder.setEmail("test@testyder.dk");
        return yder;
    }

    public static POCDMT000040Person createAssignedPerson(PN name) {
        POCDMT000040Person assignedPerson = o.createPOCDMT000040Person();

        assignedPerson.getNames().add(name);

        return assignedPerson;
    }

    public static PN createPN(List<String> givens, String family) {

        PN name = o.createPN();

        // given
        givens.forEach(value -> {
            EnGiven enGiven = o.createEnGiven();
            enGiven.getContent().add(value);
            JAXBElement<EnGiven> enGivenType = o.createENGiven(enGiven);
            name.getContent().add(enGivenType);
        });

        // family
        EnFamily enFamily = o.createEnFamily();
        enFamily.getContent().add(family);
        JAXBElement<EnFamily> enFamilyType = o.createENFamily(enFamily);
        name.getContent().add(enFamilyType);

        return name;
    }


}
