/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.hsuid.*;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.stamkort.auditlog.AuditLogger;
import dk.stamkort.conf.FskConfig;
import dk.stamkort.conf.TestDAOConfig;
import dk.stamkort.context.FskWhitelistedDevelopmentSecurityProvider;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.*;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.HasTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.HasLivingWillResponse;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.HasOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType.DocumentResponse;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import org.hl7.v3.*;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mockito;
import org.mockito.stubbing.Answer;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.soap.SOAPException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceContext;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.StringReader;
import java.util.*;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = { TestDAOConfig.class, FskConfig.class, DocumentProviderWSTest.TestConf.class })
@Transactional
public class DocumentProviderWSTest {
    static AuditLogger auditLogger = Mockito.mock(AuditLogger.class);
    private static final String nameNotFilled = "Ikke udfyldt";

    @Configuration
    static class TestConf {

        @Autowired
        WebServiceContext context;

        @Bean
        public DocumentProviderWS documentProviderWS() {
            DocumentProviderWS wp = new DocumentProviderWS();
            wp.setWebServiceContext(context);
            return wp;
        }

        @Bean
        public OrganDonorClient organDonorClient(Properties properties) {
            return mock(OrganDonorClient.class);
        }

        @Bean
        public LivingWillClient livingWillClient(Properties properties) {
            return mock(LivingWillClient.class);
        }

        @Bean
        public TreatmentWillClient treatmentWillClient(Properties properties) {
            return mock(TreatmentWillClient.class);
        }

        @Bean
        public PersonalDataCardRegisterClient personalDataCardRegisterClient(Properties properties) {
            return mock(PersonalDataCardRegisterClient.class);
        }

        @Bean
        public SCESClient scesClient(Properties properties) {
            return mock(SCESClient.class);
        }

        @Bean
        public YderClient yderClient(Properties properties) {
            return mock(YderClient.class);
        }

        @Bean
        public AuditLogger auditLogger() {
            return auditLogger;
        }
    }

    @Autowired
    RegistryIndexDAO indexDAO;

    @Autowired
    Properties properties;

    @Autowired
    DocumentProviderWS ws;

    @Autowired
    public PersonalDataCardServiceImpl personalDataCardService;

    @Before
    public void setup() throws SOAPException, JAXBException, IOException, DGWSClientException, TransformerException {
        System.setProperty("dk.sds.nsp.security.provider", FskWhitelistedDevelopmentSecurityProvider.class.getName());

        HasLivingWillResponse lwResponse = new HasLivingWillResponse();
        lwResponse.setWillExists(true);
        HasTreatmentWillResponse twResponse = new HasTreatmentWillResponse();
        twResponse.setWillExists(true);
        HasOrganDonorRegistrationResponse odResponse = new HasOrganDonorRegistrationResponse();
        odResponse.setRegistrationExists(false);
        when(personalDataCardService.livingWillClient.call(anyString(), any())).thenReturn(lwResponse);
        when(personalDataCardService.treatmentWillClient.call(anyString(),any())).thenReturn(twResponse);
        when(personalDataCardService.organDonorClient.call(anyString(),any())).thenReturn(odResponse);

        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("12345"), personalDataCardService.cpre, properties);

        when(personalDataCardService.yderClient.call(anyString(),any())).thenAnswer(
                (Answer<YderVO>) invocation -> DocumentProviderTestData.createYderData(invocation.getArgument(0)));

        // Sikker at loggen er resat efter hver test
        Mockito.reset(auditLogger);
    }

    @Test
    public void testDocumentLookupNoYdernrSucces() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(null);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(),any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("0"), personalDataCardService.cpre, properties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");

        POCDMT000040PatientRole patientRole = clinicalDocument.getRecordTargets().get(0).getPatientRole();
        assertEquals(cpr, patientRole.getIds().get(0).getExtension());

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(8, entries.size());
        POCDMT000040Entry firstCustodianEntry = entries.get(0);
        POCDMT000040Entry secondCustodianEntry = entries.get(1);
        POCDMT000040Entry nameAndAddressEntry = entries.get(2);
        POCDMT000040Entry healthInsuranceGroup = entries.get(3);
        POCDMT000040Entry organDonorEntry = entries.get(4);
        POCDMT000040Entry treatmentWillEntry = entries.get(5);
        POCDMT000040Entry livingWillEntry = entries.get(6);
        POCDMT000040Entry dentistEntry = entries.get(7);

        checkCustodian(firstCustodianEntry, secondCustodianEntry);
        checkNameAndAddress(nameAndAddressEntry);
        checkSikringsgruppe(healthInsuranceGroup);
        checkOrganDonorAndLivingWillAndTreatmentWill(organDonorEntry, livingWillEntry, treatmentWillEntry);

        assertNotNull(dentistEntry);
        EN dentistEn = (EN) dentistEntry.getObservation().getValues().get(0);
        AD dentistAd = (AD)dentistEntry.getObservation().getValues().get(1);

        // No yder provided
        assertTrue(dentistEntry.getObservation().getValues().stream().noneMatch(II.class::isInstance));
        assertEquals("OrgName From Register", getJAXBElementsFromName(EnGiven.class, dentistEn.getContent()).getContent().get(0));
        assertEquals("Vej 42", getJAXBElementsFromName(AdxpStreetAddressLine.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("8000", getJAXBElementsFromName(AdxpPostalCode.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("Aarhus C", getJAXBElementsFromName(AdxpCity.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("DK", getJAXBElementsFromName(AdxpCountry.class, dentistAd.getContent()).getContent().get(0));

        // Check that the empty CPRE AssociatedGeneralPractitionerStructure causes no header participant
        assertEquals(0, clinicalDocument.getParticipants().size());
    }

    @Test
    public void testDocumentLookupNoYdernrFail_missingFamilyName() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(null);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(),any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPREMissingFamilyName("0"), personalDataCardService.cpre, properties);

        try {
            getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");
            Assert.fail("Should not pass");
        } catch (Exception e) {
            Assert.assertTrue(e instanceof RuntimeException);
            Assert.assertEquals(e.getMessage(), "Index: 0, Size: 0");
        }
    }

    @Test
    public void testDocumentLookupNoYdernrSucces_missingGivenName() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(null);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(),any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPREMissingGivenName("0"), personalDataCardService.cpre, properties);

        try {
            getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");
        } catch (Exception e) {
            Assert.assertTrue(e instanceof RuntimeException);
            Assert.assertEquals(e.getMessage(), "Index: 0, Size: 0");
        }
    }

    @Test
    public void testDocumentLookup_PDC_2_0_WithYdernrSucces() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        testDocumentLookup_PDC_2_0_WithYdernrSucces("012345");
    }

    @Test
    public void testDocumentLookup_PDC_2_0_WithYdernrSuccesNoYderNrPadding() throws Exception {
        testDocumentLookup_PDC_2_0_WithYdernrSucces("12345");
    }

    @Test
    public void testDocumentLookupWithYdernrSucces() throws Exception {
        testDocumentLookupWithYdernrSucces("012345");
    }

    @Test
    public void testDocumentLookupWithYdernrSuccesNoYderNrPadding() throws Exception {
        testDocumentLookupWithYdernrSucces("12345");
    }

    @Test
    public void testDocumentLookupCprNotMatching() throws Exception {
        String cpr = "1122334455";
        String otherCpr = "1122334466";
        String fskDocumentId = UUID.randomUUID().toString();

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(null);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(), any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("0"), personalDataCardService.cpre, properties);

        String error = getClinicalDocumentError(fskDocumentId, cpr, otherCpr);
        assertTrue(error.contains("PatientId i sikkerhedsbilletten matcher ikke patientId i requestet."));

        // No document ids or patient ids in log
        Mockito.verify(auditLogger).logRetrieveDocumentSet(
                eq(new HashSet<>()), eq(new HashSet<>()), any(RequestContextVO.class));
    }

    private void testDocumentLookupWithYdernrSucces(String yderNr) throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(yderNr);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(), any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE(yderNr), personalDataCardService.cpre, properties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");

        POCDMT000040PatientRole patientRole = clinicalDocument.getRecordTargets().get(0).getPatientRole();
        assertEquals(cpr, patientRole.getIds().get(0).getExtension());

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(8, entries.size());
        POCDMT000040Entry firstCustodianEntry = entries.get(0);
        POCDMT000040Entry secondCustodianEntry = entries.get(1);
        POCDMT000040Entry nameAndAddressEntry = entries.get(2);
        POCDMT000040Entry healthInsuranceGroup = entries.get(3);
        POCDMT000040Entry organDonorEntry = entries.get(4);
        POCDMT000040Entry treatmentWillEntry = entries.get(5);
        POCDMT000040Entry livingWillEntry = entries.get(6);
        POCDMT000040Entry dentistEntry = entries.get(7);

        checkCustodian(firstCustodianEntry, secondCustodianEntry);
        checkNameAndAddress(nameAndAddressEntry);
        checkSikringsgruppe(healthInsuranceGroup);
        checkOrganDonorAndLivingWillAndTreatmentWill(organDonorEntry, livingWillEntry, treatmentWillEntry);

        // Check that healthprovider has been enriched with data from yderregister
        assertNotNull(dentistEntry);
        II dentistIi = (II) dentistEntry.getObservation().getValues().get(0);
        EN dentistEn = (EN) dentistEntry.getObservation().getValues().get(1);
        AD dentistAd = (AD) dentistEntry.getObservation().getValues().get(2);

        assertEquals("012345", dentistIi.getExtension());
        assertEquals("OrgName From Register", getJAXBElementsFromName(EnGiven.class, dentistEn.getContent()).getContent().get(0));
        assertEquals("Vej 42", getJAXBElementsFromName(AdxpStreetAddressLine.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("8000", getJAXBElementsFromName(AdxpPostalCode.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("Aarhus C", getJAXBElementsFromName(AdxpCity.class, dentistAd.getContent()).getContent().get(0));
        assertEquals("DK", getJAXBElementsFromName(AdxpCountry.class, dentistAd.getContent()).getContent().get(0));

        // Check that CPRE AssociatedGeneralPractitionerStructure is mapped to header participant and enriched with data from yderregister
        POCDMT000040Organization providerOrganisation = clinicalDocument.getRecordTargets().get(0).getPatientRole().getProviderOrganization();
        assertEquals("012345", providerOrganisation.getIds().get(0).getExtension());
        assertEquals("Yderregisteret", providerOrganisation.getIds().get(0).getAssigningAuthorityName());
        assertEquals("1.2.208.176.1.4", providerOrganisation.getIds().get(0).getRoot());
        assertEquals("Testyder", providerOrganisation.getNames().get(0).getContent().get(0));
        assertEquals("tel:12345678", providerOrganisation.getTelecoms().get(0).getValue());
        assertEquals("WP", providerOrganisation.getTelecoms().get(0).getUses().get(0));
        assertEquals("Ydervej 42", getJAXBElementsFromName(AdxpStreetAddressLine.class, providerOrganisation.getAddrs().get(0).getContent()).getContent().get(0));
        assertEquals("1234", getJAXBElementsFromName(AdxpPostalCode.class, providerOrganisation.getAddrs().get(0).getContent()).getContent().get(0));
        assertEquals("Yderby", getJAXBElementsFromName(AdxpCity.class, providerOrganisation.getAddrs().get(0).getContent()).getContent().get(0));

        Mockito.verify(auditLogger).logRetrieveDocumentSet(
                eq(new HashSet<>(Arrays.asList(fskDocumentId))), eq(new HashSet<>(Arrays.asList(cpr))), any(RequestContextVO.class));

    }

    @Test
    public void testDocumentLookupWithStatus30() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("30");

        assertEquals("109: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }

    @Test
    public void testDocumentLookupWithStatus50() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("50");

        assertEquals("110: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }

    @Test
    public void testDocumentLookupWithStatus60() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("60");

        assertEquals("111: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }

    @Test
    public void testDocumentLookupWithStatus90() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("90");

        assertEquals("112: Det er ikke muligt at hente stamkortet, da borgeren er afgået ved døden for mere end 1 år siden: null", error);
    }

    @Test
    public void testDocumentLookupForNonExistingDocumentid() throws DAOException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        RegistryIndexVO indexVO = new RegistryIndexVO();
        indexVO.setPersonIdentifier(cpr);
        indexVO.setFSKDocumentId(fskDocumentId);
        indexDAO.createRegistryIndexVO(indexVO);

        Security securityHeader = null;
        Holder<Header> medcomHeader = null;
        HsuidHeader hsuidHeader = null;
        RetrieveDocumentSetRequestType body = new RetrieveDocumentSetRequestType();
        RetrieveDocumentSetRequestType.DocumentRequest documentRequest = new RetrieveDocumentSetRequestType.DocumentRequest();
        documentRequest.setDocumentUniqueId("dummy-foo-bar");
        documentRequest.setHomeCommunityId(properties.getDdsHomeCommunityId());
        documentRequest.setRepositoryUniqueId(properties.getDdsRepositoryUniqueId());
        body.getDocumentRequest().add(documentRequest);
        RetrieveDocumentSetResponseType response = ws.documentRepositoryRetrieveDocumentSet(body, securityHeader, medcomHeader, hsuidHeader);
        System.out.println(response);
        assertEquals("Did not expect any returned document for dummy document id", 0, response.getDocumentResponse().size());
    }

    private HsuidHeader createHsuidHeader(String cpr) {
        List<Attribute> attributes = new ArrayList<Attribute>();
        attributes.add(UserTypeAttribute.CITIZEN);
        attributes.add(new CitizenCivilRegistrationNumberAttribute(cpr));
        attributes.add(new ActingUserCivilRegistrationNumberAttribute(cpr));
        attributes.add(new SystemVendorNameAttribute("system Owner"));
        attributes.add(new SystemNameAttribute("system Name"));
        attributes.add(new SystemVersionAttribute("system Version"));
        attributes.add(new OperationsOrganisationNameAttribute("responsible Org"));

        return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader("system owner", attributes);
    }

    private String testDocumentLookupWithStatusCommon(String status)
            throws DGWSClientException, IOException, JAXBException, SOAPException, DAOException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse("012345");
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(), any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("012345", status), personalDataCardService.cpre, properties);

        return getClinicalDocumentError(fskDocumentId, cpr, cpr);
    }

    private void checkOrganDonorAndLivingWillAndTreatmentWill(POCDMT000040Entry organDonorEntry, POCDMT000040Entry livingWillEntry, POCDMT000040Entry treatmentWillEntry) {
        assertNotNull(organDonorEntry);
        assertEquals("OrganDonorRegistration", organDonorEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.100.1", organDonorEntry.getObservation().getCode().getCodeSystem());
        assertEquals("Registreret organdonor", organDonorEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, organDonorEntry.getObservation().getValues().size());
        assertEquals("false", ((II) organDonorEntry.getObservation().getValues().get(0)).getExtension());

        assertNotNull(livingWillEntry);
        assertEquals("LivingWillRegistration", livingWillEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.100.1", livingWillEntry.getObservation().getCode().getCodeSystem());
        assertEquals("Registreret livstestamente", livingWillEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, livingWillEntry.getObservation().getValues().size());
        assertEquals("true", ((II) livingWillEntry.getObservation().getValues().get(0)).getExtension());

        assertNotNull(treatmentWillEntry);
        assertEquals("TreatmentWillRegistration", treatmentWillEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.100.1", treatmentWillEntry.getObservation().getCode().getCodeSystem());
        assertEquals("Registreret behandlingstestamente", treatmentWillEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, treatmentWillEntry.getObservation().getValues().size());
        assertEquals("true", ((II) treatmentWillEntry.getObservation().getValues().get(0)).getExtension());
    }

    private POCDMT000040ClinicalDocument getClinicalDocument(String fskDocumentId, String cpr, String typeCode, String formatCode) throws DAOException, JAXBException {
        RetrieveDocumentSetResponseType response = getClinicalDocumentResponse(fskDocumentId, cpr, typeCode, formatCode, cpr);
        List<DocumentResponse> documentResponse = response.getDocumentResponse();
        RetrieveDocumentSetResponseType.DocumentResponse responseDocument = documentResponse.get(0);
        assertEquals("Returned documentId must match requested document id", fskDocumentId, responseDocument.getDocumentUniqueId());
        assertEquals(properties.getDdsHomeCommunityId(), responseDocument.getHomeCommunityId());
        assertNotNull(responseDocument.getDocument());
        System.out.println(new String(responseDocument.getDocument()));

        JAXBContext context = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        Unmarshaller unmarshaller = context.createUnmarshaller();
        ByteArrayInputStream bais = new ByteArrayInputStream(responseDocument.getDocument());
        // https://www.igorkromin.net/index.php/2017/03/15/unmarshalling-an-xml-fragment-representing-a-jaxb-object-without-xmlrootelement/
        return unmarshaller.unmarshal(new StreamSource(bais), POCDMT000040ClinicalDocument.class).getValue();
    }

    private String getClinicalDocumentError(String fskDocumentId, String cpr, String hsuidCpr) throws DAOException, JAXBException {
        RetrieveDocumentSetResponseType response = getClinicalDocumentResponse(fskDocumentId, cpr, "52460-3", "DK FSK Schema", hsuidCpr);

        return response.getRegistryResponse().getRegistryErrorList().getRegistryError().get(0).getCodeContext();
    }

    private RetrieveDocumentSetResponseType getClinicalDocumentResponse(String fskDocumentId, String cpr, String typeCode, String formatCode, String hsuidCpr)
            throws DAOException {
        RegistryIndexVO indexVO = new RegistryIndexVO();
        indexVO.setPersonIdentifier(cpr);
        indexVO.setFSKDocumentId(fskDocumentId);
        indexVO.setTypeCode(typeCode);
        indexVO.setFormatCode(formatCode);
        indexDAO.createRegistryIndexVO(indexVO);

        Security securityHeader = null;
        Holder<Header> medcomHeader = null;
        HsuidHeader hsuidHeader = createHsuidHeader(hsuidCpr);

        RetrieveDocumentSetRequestType body = new RetrieveDocumentSetRequestType();
        RetrieveDocumentSetRequestType.DocumentRequest documentRequest = new RetrieveDocumentSetRequestType.DocumentRequest();
        documentRequest.setDocumentUniqueId(fskDocumentId);
        documentRequest.setHomeCommunityId(properties.getDdsHomeCommunityId());
        documentRequest.setRepositoryUniqueId(properties.getDdsRepositoryUniqueId());
        body.getDocumentRequest().add(documentRequest);
        return ws.documentRepositoryRetrieveDocumentSet(body, securityHeader, medcomHeader, hsuidHeader);
    }

    private void testDocumentLookup_PDC_2_0_WithYdernrSucces(String yderNr) throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(yderNr);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(),any())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE(yderNr), personalDataCardService.cpre, properties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "PDC", "urn:ad:dk:medcom:pdc-v2.0:full");

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(8, entries.size());

        Mockito.verify(auditLogger).logRetrieveDocumentSet(
                eq(new HashSet<>(Arrays.asList(fskDocumentId))), eq(new HashSet<>(Arrays.asList(cpr))), any(RequestContextVO.class));
    }

    private void mockCPREResponse(String response, SCESClient cpre, Properties p) throws JAXBException,
            DGWSClientException {
        Unmarshaller unmarshaller = JAXBContext.newInstance(GetPersonWithHealthCareInformationOut.class).createUnmarshaller();
        GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformationOut = (GetPersonWithHealthCareInformationOut) unmarshaller.unmarshal(new StringReader(response));

        Mockito.when(cpre.call(anyString(), any(RequestContextVO.class))).thenReturn(getPersonWithHealthCareInformationOut);
    }

    private void checkSikringsgruppe(POCDMT000040Entry healthInsuranceGroup) {
        assertEquals("Sygesikringen", ((II)healthInsuranceGroup.getObservation().getValues().get(0)).getAssigningAuthorityName());
        assertEquals("1", ((II)healthInsuranceGroup.getObservation().getValues().get(0)).getExtension());
        assertEquals("1.2.208.176.1.99.99.99", ((II)healthInsuranceGroup.getObservation().getValues().get(0)).getRoot());
    }

    private void checkCustodian(POCDMT000040Entry firstCustodianEntry, POCDMT000040Entry secondCustodianEntry) {
        II ii = (II)firstCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        PN pn = (PN)firstCustodianEntry.getObservation().getValues().get(1);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        CD cd = (CD)firstCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());

        ii = (II)secondCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        pn = (PN)secondCustodianEntry.getObservation().getValues().get(1);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        cd = (CD)secondCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());
    }

    private void checkCustodianMissingFamilyName(POCDMT000040Entry firstCustodianEntry, POCDMT000040Entry secondCustodianEntry) {
        II ii = (II)firstCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        PN pn = (PN)firstCustodianEntry.getObservation().getValues().get(1);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        CD cd = (CD)firstCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());

        ii = (II)secondCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        pn = (PN)secondCustodianEntry.getObservation().getValues().get(1);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        cd = (CD)secondCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());
    }

    private void checkCustodianMissingGivenName(POCDMT000040Entry firstCustodianEntry, POCDMT000040Entry secondCustodianEntry) {
        II ii = (II)firstCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        PN pn = (PN)firstCustodianEntry.getObservation().getValues().get(1);
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        CD cd = (CD)firstCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());

        ii = (II)secondCustodianEntry.getObservation().getValues().get(0);
        assertEquals("CPR", ii.getAssigningAuthorityName());
        assertEquals("0406048357", ii.getExtension());
        assertEquals("1.2.208.176.1.2", ii.getRoot());

        pn = (PN)secondCustodianEntry.getObservation().getValues().get(1);
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        cd = (CD)secondCustodianEntry.getObservation().getValues().get(2);
        assertEquals("barn", cd.getCode());
        assertEquals("1.2.208.184.100.2", cd.getCodeSystem());
        assertEquals("Barn", cd.getDisplayName());
    }

    private void checkNameAndAddress(POCDMT000040Entry nameAndAddressEntry) {
        PN pn = (PN)nameAndAddressEntry.getObservation().getValues().get(0);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        AD ad = (AD)nameAndAddressEntry.getObservation().getValues().get(1);
        assertEquals("Valbyvej 235", getJAXBElementsFromName(AdxpStreetAddressLine.class, ad.getContent()).getContent().get(0));
        assertEquals("1926", getJAXBElementsFromName(AdxpPostalCode.class, ad.getContent()).getContent().get(0));
        assertEquals("Frederiksberg C", getJAXBElementsFromName(AdxpCity.class, ad.getContent()).getContent().get(0));
        assertEquals("DK\n                        ", getJAXBElementsFromName(AdxpCountry.class, ad.getContent()).getContent().get(0));
    }

    private void checkNameAndAddressMissingFamilyName(POCDMT000040Entry nameAndAddressEntry) {
        PN pn = (PN)nameAndAddressEntry.getObservation().getValues().get(0);
        assertEquals("Jakob", getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        AD ad = (AD)nameAndAddressEntry.getObservation().getValues().get(1);
        assertEquals("Valbyvej 235", getJAXBElementsFromName(AdxpStreetAddressLine.class, ad.getContent()).getContent().get(0));
        assertEquals("1926", getJAXBElementsFromName(AdxpPostalCode.class, ad.getContent()).getContent().get(0));
        assertEquals("Frederiksberg C", getJAXBElementsFromName(AdxpCity.class, ad.getContent()).getContent().get(0));
        assertEquals("DK\n                        ", getJAXBElementsFromName(AdxpCountry.class, ad.getContent()).getContent().get(0));
    }

    private void checkNameAndAddressMissingGiveName(POCDMT000040Entry nameAndAddressEntry) {
        PN pn = (PN)nameAndAddressEntry.getObservation().getValues().get(0);
        assertEquals(nameNotFilled, getJAXBElementsFromName(EnGiven.class, pn.getContent()).getContent().get(0));
        assertEquals("Pedersen", getJAXBElementsFromName(EnFamily.class, pn.getContent()).getContent().get(0));

        AD ad = (AD)nameAndAddressEntry.getObservation().getValues().get(1);
        assertEquals("Valbyvej 235", getJAXBElementsFromName(AdxpStreetAddressLine.class, ad.getContent()).getContent().get(0));
        assertEquals("1926", getJAXBElementsFromName(AdxpPostalCode.class, ad.getContent()).getContent().get(0));
        assertEquals("Frederiksberg C", getJAXBElementsFromName(AdxpCity.class, ad.getContent()).getContent().get(0));
        assertEquals("DK\n                        ", getJAXBElementsFromName(AdxpCountry.class, ad.getContent()).getContent().get(0));
    }

    private <T> T getJAXBElementsFromName(Class<T> clazz, List<Serializable> elements) {
        return elements.stream()
                .filter(o -> o instanceof JAXBElement)
                .map(JAXBElement.class::cast)
                .filter(jaxbElement -> jaxbElement.getDeclaredType().equals(clazz))
                .reduce((jaxbElement, jaxbElement2) -> {
                    throw new IllegalStateException("Multiple element found of type " + clazz);
                })
                .map(JAXBElement::getValue)
                .filter(clazz::isInstance)
                .map(clazz::cast)
                .orElseThrow(() -> new RuntimeException("Element of not found of type: " + clazz))
                ;
    }
}
