/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.*;
import dk.stamkort.integrations.fgvhr.FgvhrClient;
import dk.stamkort.integrations.fgvhr.FgvhrClientException;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.GetPersonalDataCardResponse;
import oio.medcom.cprservice._1_0.*;
import org.hl7.v3.*;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import java.math.BigInteger;

import static dk.stamkort.webservice.ondemand.PersonalDataCardServiceTest.EnableClient.*;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

@RunWith(SpringJUnit4ClassRunner.class)
public class PersonalDataCardServiceTest {

    @Mock
    private Properties properties;

    @Mock
    public OrganDonorClient organDonorClient;

    @Mock
    public LivingWillClient livingWillClient;

    @Mock
    public TreatmentWillClient treatmentWillClient;

    @Mock
    public PersonalDataCardRegisterClient personalDataCardRegisterClient;

    @Mock
    public SCESClient cpre;

    @Mock
    public YderClient yderClient;

    @Mock
    public FgvhrClient fgvhrClient;

    @InjectMocks
    PersonalDataCardServiceImpl pdcs;

    enum EnableClient {Fgvhr, Skr, Sces}

    void init(EnableClient e) {
        when(properties.getDgwsClientPoolSize()).thenReturn(1);
        when(properties.isFgvhrIntegrationEnabled()).thenReturn(e == Fgvhr);
        when(properties.isBtrIntegrationEnabled()).thenReturn(false);
        when(properties.isLtrIntegrationEnabled()).thenReturn(false);
        when(properties.isOdrIntegrationEnabled()).thenReturn(false);
        when(properties.isScesIntegrationEnabled()).thenReturn(e == Sces);
        when(properties.isSkrIntegrationEnabled()).thenReturn(e == Skr);
        when(properties.getDgwsClientTimeoutMillis()).thenReturn(1000L);
        try {
            when(yderClient.call(anyString(), any())).thenReturn(new YderVO());
        } catch (DGWSClientException ex) {
            throw new RuntimeException(ex);
        }
        pdcs.init();
    }

    @Test
    public void givenRegisterDataWhenFghvrFailsThenResuscitationShouldNotBePresent() throws FgvhrClientException {
        init(Fgvhr);
        when(fgvhrClient.hasNoResuscitation(anyString()))
                .thenThrow(new FgvhrClientException("", new RuntimeException()));
        assertFalse(pdcs.getRegisterData("", new RequestContextVO()).getHasNoResuscitation().isPresent());
    }

    @Test
    public void givenRegisterDataWhenSkrReturnsSorThenYderRegisterShouldNotBeCalled() throws DGWSClientException {
        init(Skr);
        String dummySorCode = "1234567890";
        when(personalDataCardRegisterClient.call(anyString(), any())).thenReturn(mockResponse(dummySorCode, "SOR"));
        pdcs.getRegisterData("", new RequestContextVO());
        verify(yderClient, never()).call(anyString(), any());
    }

    @Test
    public void givenRegisterDataWhenSkrReturnsYderThenDentistShouldBeSet() throws DGWSClientException {
        init(Skr);
        when(personalDataCardRegisterClient.call(anyString(), any())).thenReturn(mockResponse("1234", "Yder"));
        assertTrue(pdcs.getRegisterData("", new RequestContextVO()).getDentist().isPresent());
        verify(yderClient, times(1)).call(anyString(), any());
    }

    @Test
    public void givenRegisterDataWhenScesReturnsYderThenGeneralPractitionerShouldBeSet() throws DGWSClientException {
        init(Sces);
        when(cpre.call(anyString(), any())).thenReturn(mockResponse());
        assertTrue(pdcs.getRegisterData("", new RequestContextVO()).getGeneralPractitioner().isPresent());
        verify(yderClient, times(1)).call(anyString(), any());
    }

    private static GetPersonalDataCardResponse mockResponse(String code, String type) {
        II a = new II(); a.setExtension(code); a.setAssigningAuthorityName(type);
        POCDMT000040Organization b = new POCDMT000040Organization(); b.getIds().add(a);
        HealthProvider c = new HealthProvider(); c.setOrganization(b);
        POCDMT000040Entry d = new POCDMT000040Entry(); d.setHealthProvider(c);
        POCDMT000040Section e = new POCDMT000040Section(); e.getEntries().add(d);
        POCDMT000040Component3 f = new POCDMT000040Component3(); f.setSection(e);
        POCDMT000040StructuredBody g = new POCDMT000040StructuredBody(); g.getComponents().add(f);
        POCDMT000040Component2 h = new POCDMT000040Component2(); h.setStructuredBody(g);
        POCDMT000040ClinicalDocument i = new POCDMT000040ClinicalDocument(); i.setComponent(h);
        GetPersonalDataCardResponse j = new GetPersonalDataCardResponse(); j.setClinicalDocument(i);
        return j;
    }

    private static GetPersonWithHealthCareInformationOut mockResponse() {
        AssociatedGeneralPractitionerStructureType a = new AssociatedGeneralPractitionerStructureType();
        a.setAssociatedGeneralPractitionerIdentifier(new BigInteger("321312"));
        CurrentGeneralPractitionerStructureType b = new CurrentGeneralPractitionerStructureType();
        b.setAssociatedGeneralPractitionerStructure(a);
        PersonHealthCareInformationStructureType c = new PersonHealthCareInformationStructureType();
        c.setCurrentGeneralPractitionerStructure(b);
        PersonWithHealthCareInformationStructureType d = new PersonWithHealthCareInformationStructureType();
        d.setPersonHealthCareInformationStructure(c);
        GetPersonWithHealthCareInformationOut e = new GetPersonWithHealthCareInformationOut();
        e.setPersonWithHealthCareInformationStructure(d);
        return e;
    }
}