/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort;

import dk.nsp.test.idp.NspSoapClient;
import dk.nsp.test.idp.model.Identity;
import dk.stamkort.util.skr.SkrFacade;
import org.apache.commons.io.IOUtils;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.hl7.v3.POCDMT000040Component3;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Before;
import org.w3c.dom.Document;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

import static dk.nsp.test.idp.EmployeeIdentities.lægeCharlesBabbage;
import static dk.nsp.test.idp.OrganizationIdentities.sundhedsdatastyrelsen;
import static dk.stamkort.AbstractITCase.NspSoapResponseData.from;

public class AbstractITCase {
    SkrFacade skrFacade;
    URI skrEndpoint = new URI("http://test1-cnsp.ekstern-test.nspop.dk:8080/skr/skr");

    public static final String TEST_PERSON = "0509900267";

    protected String host = "localhost";
    protected String port = "8080";
    private String docid = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000003";

    public static class NspSoapResponseData {
        public final String content;
        public final int status;
        public final String statusMessage;
        public final boolean isFault;

        public NspSoapResponseData(String content, int status, String statusMessage, boolean isFault) {
            this.content = content;
            this.status = status;
            this.statusMessage = statusMessage;
            this.isFault = isFault;
        }

        public static NspSoapResponseData from(NspSoapClient.NspSoapResponse res) {
            try { return new NspSoapResponseData(IOUtils.toString(res.getResponse()),
                        res.getStatus(),
                        res.getStatusMessage(),
                        res.isFault()); }
            catch (IOException e) { throw new RuntimeException(e); }
        }
    }

    public final boolean isTestLocal() { return host.equals("localhost"); }

    public static NspSoapResponseData executeNspRequest(URI uri, String soapAction, Identity id, String req) {
        try(NspSoapClient c = new NspSoapClient(System.out::println)) {
            try(NspSoapClient.NspSoapResponse res = c.request(uri, soapAction)
                    .as(id)
                    .execute(req)) { return from(res); }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static <T> T parseNspSoapResponse(NspSoapResponseData response, Class<T> c)
            throws SOAPException, IOException, JAXBException {
        final Document message = MessageFactory
                .newInstance()
                .createMessage(null, new ByteArrayInputStream(response.content.getBytes()))
                .getSOAPBody()
                .extractContentAsDocument();
        return JAXBContext.newInstance(c)
                .createUnmarshaller()
                .unmarshal(message, c)
                .getValue();
    }

    public AbstractITCase() throws URISyntaxException {
        String host = System.getProperty("FSK_TEST_HOST");
        String port = System.getProperty("FSK_TEST_PORT");
        String docid = System.getProperty("FSK_TEST_DOCID");
        String skrEndpointProp = System.getProperty("SKR_ENDPOINT");

        if (host != null && host.length() >0 ) {
            this.host = host;
        }

        if (port != null && port.length() > 0) {
            this.port = port;
        }

        if (docid != null && docid.length() > 0) {
            this.docid = docid;
        }

        if (skrEndpointProp != null && skrEndpointProp.length() > 0) {
            skrEndpoint = new URI(skrEndpointProp);
        }

        skrFacade = new SkrFacade(skrEndpoint);
    }

    @Before
    public void setup() throws Exception { cleanup(); }

    /**
     @deprecated Use RetrieveDocumentSet(String, String, boolean) instead.
     */
    @Deprecated
    protected NspSoapResponseData RequestRetrieveDocumentSet(String soapAction, boolean systemIdCard)
            throws Exception {
        final URI endpoint = new URI("http://" + host + ":" + port + "/fsk/services/fsk");
        final String requestXml =
                IOUtils.toString(getClass().getResourceAsStream("/testrequest.xml"), StandardCharsets.UTF_8)
                .replace("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000000", docid);
        return executeNspRequest(endpoint,
                soapAction,
                systemIdCard
                        ? sundhedsdatastyrelsen()
                        : lægeCharlesBabbage(),
                requestXml);
    }

    public NspSoapResponseData RetrieveDocumentSet(String docId, String soapAction, boolean systemIdCard)
            throws Exception {
        final String requestXml =
                IOUtils.toString(getClass().getResourceAsStream("/testrequest.xml"), StandardCharsets.UTF_8)
                .replace("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000000", docId);
        return executeNspRequest(new URI("http://" + host + ":" + port + "/fsk/services/fsk"),
                                 soapAction,
                                 systemIdCard
                                         ? sundhedsdatastyrelsen()
                                         : lægeCharlesBabbage(),
                                 requestXml);
    }

    protected void cleanup() throws Exception {
        // Find id'er fra skr
        POCDMT000040ClinicalDocument clinicalDocument = skrFacade.getSkr(TEST_PERSON);

        // sprog fra TEST_PERSON
        List<String> languageIds = getLanguageIds(clinicalDocument);
        for (String languageId : languageIds) {
            skrFacade.deleteLanguage(TEST_PERSON, languageId);
        }
        // midlertidig adresse fra TEST_PERSON
        List<String> tempAddrIds = getTempAddrIds(clinicalDocument);
        for (String tempAddrId: tempAddrIds) {
            skrFacade.deleteTemporaryAddress(TEST_PERSON, tempAddrId);
        }
        // tandlæge fra TEST_PERSON
        List<String> dentistIds = getDentistsIds(clinicalDocument);
        for (String dentistId: dentistIds) {
            skrFacade.deleteDentist(TEST_PERSON, dentistId);
        }
        // pårørende fra TEST_PERSON
        List<String> relativeIds = getRelativeIds(clinicalDocument);
        for (String relativeId: relativeIds) {
            skrFacade.deleteRelative(TEST_PERSON, relativeId);
        }
    }

    private List<String> getRelativeIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> relativeIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return relativeIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return relativeIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entry : component.getSection().getEntries()) {
                if (entry.getRelatedPerson() != null) {
                    relativeIds.add(entry.getRelatedPerson().getId().getExtension());
                }
            }
        }
        return relativeIds;
    }

    private List<String> getDentistsIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> dentistIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return dentistIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return dentistIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entry : component.getSection().getEntries()) {
                if (entry.getHealthProvider() != null) {
                    dentistIds.add(entry.getHealthProvider().getId().getExtension());
                }
            }
        }
        return dentistIds;
    }

    private List<String> getLanguageIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> languageIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return languageIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return languageIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entry : component.getSection().getEntries()) {
                if (entry.getLanguage() != null) {
                    languageIds.add(entry.getLanguage().getId().getExtension());
                }
            }
        }
        return languageIds;
    }

    private List<String> getTempAddrIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> tempAddrIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return tempAddrIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return tempAddrIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entry : component.getSection().getEntries()) {
                if (entry.getTemporaryAddress() != null) {
                    tempAddrIds.add(entry.getTemporaryAddress().getId().getExtension());
                }
            }
        }
        return tempAddrIds;
    }
}
