/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort;

import org.hl7.v3.II;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Function;

import static dk.stamkort.util.XmlExt.fromXml;
import static dk.stamkort.webservice.hl7cda.assembler.PDC_3_0ClinicalDocumentAssembler.findObservation;
import static org.junit.Assert.*;

public class OnDemandServiceITCase extends AbstractITCase {
    public OnDemandServiceITCase() throws URISyntaxException { }

    @Before
    public void setupTestData() throws Exception {
        // Spoken language
        skrFacade.addLanguage(TEST_PERSON, "de");
        // Temporary Address
        skrFacade.addTemporaryAddress(TEST_PERSON);
        // Dentist
        skrFacade.addDentist(TEST_PERSON);
    }

    @Test
    // Even though service no longer supports version 1 we have a test case that requests a version 1 document.
    // This is to ensure that the service still answers nicely to version 1 requests.
    public void callOnDemandService_PDC_1_0ClinicalDocument() throws Exception {
        String response = RequestRetrieveDocumentSet("urn:ihe:iti:2007:RetrieveDocumentSet", false).content;
        assertTrue(response.contains("ClinicalDocument"));
    }

    @Test
    public void callOnDemandService_PDC_2_0ClinicalDocument_correctIds() throws Exception {
        String responseA = RequestRetrieveDocumentSet("urn:ihe:iti:2007:RetrieveDocumentSet", false).content;

        String xmlResponseA = responseA.substring(responseA.indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?><ClinicalDocument moodCode=\"EVN\""), responseA.indexOf("</ClinicalDocument>")+"</ClinicalDocument>".length());
        assertNotNull(responseA);
        assertTrue(responseA.contains("ClinicalDocument"));

        JAXBContext contextA = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        JAXBElement<POCDMT000040ClinicalDocument> rA = contextA.createUnmarshaller().unmarshal(new StreamSource(new StringReader(xmlResponseA)), POCDMT000040ClinicalDocument.class);

        String languageIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(5).getObservation().getIds().get(0).getExtension();
        String tempAddrIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(6).getObservation().getIds().get(0).getExtension();
        String dentistIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(7).getObservation().getIds().get(0).getExtension();

        // Run again
        String responseB = RequestRetrieveDocumentSet("urn:ihe:iti:2007:RetrieveDocumentSet", false).content;

        String xmlResponseB = responseB.substring(responseB.indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?><ClinicalDocument moodCode=\"EVN\""), responseB.indexOf("</ClinicalDocument>")+"</ClinicalDocument>".length());
        assertNotNull(responseB);
        assertTrue(responseB.contains("ClinicalDocument"));

        JAXBContext contextB = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        JAXBElement<POCDMT000040ClinicalDocument> rB = contextB.createUnmarshaller().unmarshal(new StreamSource(new StringReader(xmlResponseB)), POCDMT000040ClinicalDocument.class);

        String languageIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(5).getObservation().getIds().get(0).getExtension();
        String tempAddrIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(6).getObservation().getIds().get(0).getExtension();
        String dentistIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(7).getObservation().getIds().get(0).getExtension();

        // Assert the same Ids are provided in both cal
        assertEquals(tempAddrIdA, tempAddrIdB);
        assertEquals(dentistIdA, dentistIdB);
        assertEquals(languageIdA, languageIdB);
        assertNull(rA.getValue().getTemplateIds().get(0).getExtension());
        String birthTimeValue = getBirthTimeValue(xmlResponseB);
        assertEquals("19900905", birthTimeValue);
    }

    @Test
    public void callOnDemandService_PDC_3_0ClinicalDocument_fgvhrInTheText() throws Exception {
        // Arrange
        String v3docId = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000004";

        // Act
        String responseDoc = requestDocument(v3docId);

        // Assert
        String resuscitateNarrativeText =
                extractSubstring(responseDoc, "<td>Fravalg af genoplivning ved hjertestop</td>", "</tr>")
                        .orElseThrow(NoSuchElementException::new);
        assertTrue(resuscitateNarrativeText.contains("Fravalg af genoplivning ved hjertestop"));
        assertTrue(resuscitateNarrativeText.contains("Aktiv") || resuscitateNarrativeText.contains("Inaktiv"));
        String birthTimeValue = getBirthTimeValue(responseDoc);
        assertEquals("19900905000000+0000", birthTimeValue);
    }

    @Test
    public void callOnDemandService_PDC_3_0ClinicalDocument_dentistInTheText() throws Exception {
        // Arrange
        String v3docId = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000004";

        // Act
        String responseDoc = requestDocument(v3docId);

        // Assert
        String dentistNarrativeText = extractSubstring(responseDoc, "<td>Tandlæge indtastet</td>", "</tr>")
                        .orElseThrow(NoSuchElementException::new);
        assertFalse(dentistNarrativeText.contains("NA"));
        assertTrue(dentistNarrativeText.contains("Tandklinikken smil. nu"));
        String birthTimeValue = getBirthTimeValue(responseDoc);
        assertEquals("19900905000000+0000", birthTimeValue);
    }

    @Test
    public void callOnDemandService_PDC_3_0ClinicalDocument_correctPDCVersion() throws Exception {
        // Arrange
        String v3docId = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000004";
        String searchFor = "<id root=\"1.2.208.184.100.10\" extension=\"pdc-v3.0\"";

        // Act
        String responseDoc = requestDocument(v3docId);

        // Assert
        assertTrue(responseDoc.contains(searchFor));
    }

    @Test
    public void callOnDemandService_PDC_3_0ClinicalDocument_correctIds() throws Exception {
        String languageTemplateIdRoot = "1.2.208.184.16.1.10.20.1.20",
                temporaryAddressTemplateIdRoot = "1.2.208.184.16.1.10.20.1.21",
                dentistTemplateIdRoot = "1.2.208.184.16.1.10.20.1.22",
                fgvhrTemplateIdRoot = "1.2.208.184.16.1.10.20.1.31",
                v3docId = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000004";
        Function<POCDMT000040ClinicalDocument, Function<String, String>> findIdInDoc =
                doc -> templateIdRoot -> findObservation(doc, templateIdRoot).findFirst().get().getIds().get(0)
                        .getExtension();

        POCDMT000040ClinicalDocument docA = fromXml(requestDocument(v3docId), POCDMT000040ClinicalDocument.class);
        Function<String, String> findIdA = findIdInDoc.apply(docA);
        String languageIdA = findIdA.apply(languageTemplateIdRoot),
                tempAddrIdA = findIdA.apply(temporaryAddressTemplateIdRoot),
                dentistIdA = findIdA.apply(dentistTemplateIdRoot);

        POCDMT000040ClinicalDocument docB = fromXml(requestDocument(v3docId), POCDMT000040ClinicalDocument.class);
        Function<String, String> findIdB = findIdInDoc.apply(docB);
        String languageIdB = findIdB.apply(languageTemplateIdRoot),
                tempAddrIdB = findIdB.apply(temporaryAddressTemplateIdRoot),
                dentistIdB = findIdB.apply(dentistTemplateIdRoot);

        assertEquals(tempAddrIdA, tempAddrIdB);
        assertEquals(languageIdA, languageIdB);
        assertEquals(dentistIdA, dentistIdB);
        assertEquals("3.0", docA.getTemplateIds().get(0).getExtension());

        II ii = (II) findObservation(docA, fgvhrTemplateIdRoot).findFirst().get().getValues().get(0);
        String hasNoResuscitation = ii.getExtension();
        assertEquals("false", hasNoResuscitation);
    }

    @Test
    public void callOnDemandService_PDC_1_0ClinicalDocument_withCitizenIdNullInHsuid() throws Exception {
        requestDocument("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000002", false);
    }


    @Test
    public void callOnDemandService_PDC_1_0ClinicalDocument_WithSystemIdCardAndHsuid() throws Exception {
        requestDocument("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000002", true);
    }


    private String requestDocument(String docId) throws Exception { return requestDocument(docId, false); }

    private String requestDocument(String docId, boolean systemIdCard) throws Exception {
        final String response = RetrieveDocumentSet(docId,
                                                    "urn:ihe:iti:2007:RetrieveDocumentSet",
                                                    systemIdCard).content;
        assertTrue(response.contains("ClinicalDocument"));
        return response.substring(response.indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\" " +
                                                           "standalone=\"yes\"?><ClinicalDocument " +
                                                           "moodCode=\"EVN\""),
                                  response.indexOf("</ClinicalDocument>") +
                                          "</ClinicalDocument>".length());
    }

    private Optional<String> extractSubstring(String input, String startTag, String endTag) {
        int startPos = input.indexOf(startTag);
        if (startPos != -1) {
            int endPos = input.indexOf(endTag, startPos);
            if (endPos != -1) {
                return Optional.of(input.substring(startPos, endPos + endTag.length()));
            }
        }
        return Optional.empty();
    }

    public static String getBirthTimeValue(String xmlResponse) throws Exception {
        // Convert String to XML Document
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = factory.newDocumentBuilder();
        Document doc = builder.parse(new InputSource(new StringReader(xmlResponse)));

        // Prepare XPath to extract the birthTime value
        XPathFactory xpathFactory = XPathFactory.newInstance();
        XPath xpath = xpathFactory.newXPath();
        String expression = "//birthTime/@value";
        String birthTimeValue = (String) xpath.evaluate(expression, doc, XPathConstants.STRING);

        return birthTimeValue;
    }
}