/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util.skr;

import dk.stamkort.OID;
import dk.sundhedsdatastyrelsen.fsk.extensions.*;
import dk.sundhedsdatastyrelsen.skr._2024._01._01.*;
import org.hl7.v3.*;

import java.net.URI;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;

import static dk.nsp.test.idp.EmployeeIdentities.lægeCharlesBabbage;
import static dk.stamkort.AbstractITCase.*;
import static dk.stamkort.util.XmlExt.toXml;

public class SkrFacade {
    public static final String DATAENTERER_ID_CHARLES_BABBAGE = "0611809735";
    private final org.hl7.v3.ObjectFactory objectFactory = new org.hl7.v3.ObjectFactory();
    private final URI endpoint;

    public SkrFacade(URI endpoint) { this.endpoint = endpoint; }

    public void addLanguage(String cpr, String lang) throws Exception {
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#CreateLanguage";
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(),toXml(createCreateLanguageRequest(cpr, lang)))
                .isFault)
        { throw new Exception(); }
    }

    public void addLanguageWithCustomDataEnterer(String cpr, String lang, POCDMT000040Author dataEnterer)
            throws Exception {
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#CreateLanguage";
        CreateLanguageRequest request = createCreateLanguageRequest(cpr, lang); request.setDataEnterer(dataEnterer);
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request)).isFault) { throw new Exception(); }
    }

    public void addLanguageWithCustomDataEnterer(String cpr, String lang, String dataEntererGivenName,
                                                 String dataEntererFamilyName)
            throws Exception {
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#CreateLanguage";
        CreateLanguageRequest request = createCreateLanguageRequest(cpr, lang);
        request.setDataEnterer(createDataEntererElement(dataEntererGivenName, dataEntererFamilyName));
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request)).isFault) {throw new Exception();}
    }

    public CreateLanguageRequest createCreateLanguageRequest(String cpr, String lang) {
        CreateLanguageRequest request = new CreateLanguageRequest();

        II ii = new II();
        ii.setAssigningAuthorityName("CPR");
        ii.setExtension(cpr);
        ii.setRoot("1.2.208.176.1.2");
        request.setId(ii);

        Language language = new Language();
        LanguageCodeType languageCode = new LanguageCodeType();
        languageCode.setValue(lang);
        languageCode.setCodeSystem("1.2.208.184.100.1");
        language.setLanguageCode(languageCode);

        request.setLanguage(language);
        request.setDataEnterer(createDataEntererElement());
        return request;
    }

    public void deleteLanguage(String cpr, String languageIdString) throws Exception {
        if (languageIdString == null) { return; }
        DeleteLanguageRequest request = new DeleteLanguageRequest();
        request.setId(createCprIdElement(cpr));

        II languageId = createId(languageIdString);

        request.setLanguageId(languageId);
        request.setDataEnterer(createDataEntererElement());

        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#DeleteLanguage";
        NspSoapResponseData response = executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request));
        if (response.isFault && !response.content.contains("did not match any current data")) { throw new Exception(); }
    }

    public UpdateContactInformationRequest createUpdateContactInformationRequest(String cpr) {
        UpdateContactInformationRequest request = new UpdateContactInformationRequest();
        request.setId(createCprIdElement(cpr));
        request.setDataEnterer(createDataEntererElement());
        TEL tel = createTel();
        request.getTelecoms().add(tel);
        return request;
    }

    public void addTemporaryAddress(String cpr) throws Exception {
        CreateTemporaryAddressRequest request = createCreateTemporaryAddressRequest(cpr);
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#CreateTemporaryAddress";
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request)).isFault) {throw new Exception();}
    }

    public CreateTemporaryAddressRequest createCreateTemporaryAddressRequest(String cpr) {
        CreateTemporaryAddressRequest request = new CreateTemporaryAddressRequest();
        request.setId(createCprIdElement(cpr));
        TemporaryAddress address = new TemporaryAddress();

        AD ad = createAddress("Temporary Street Name 20");
        addUseablePeriodToAddress(ad, "20191001");
        addUseablePeriodToAddress(ad, "20191010");
        address.setAddr(ad);

        request.setTemporaryAddress(address);
        request.setDataEnterer(createDataEntererElement());
        return request;
    }

    public void addUseablePeriodToAddress(AD address, String usablePeriod) {
        SXCMTS usablePeriodElm = new SXCMTS();
        usablePeriodElm.setValue(usablePeriod);
        address.getContent().add(objectFactory.createADUseablePeriod(usablePeriodElm));
    }

    public void deleteTemporaryAddress(String cpr, String tempAddrId) throws Exception {
        if (tempAddrId == null) {return;}
        DeleteTemporaryAddressRequest request = new DeleteTemporaryAddressRequest();
        request.setId(createCprIdElement(cpr));

        II temporaryAddressId = createId(tempAddrId);

        request.setTemporaryAddressId(temporaryAddressId);
        request.setDataEnterer(createDataEntererElement());

        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#DeleteTemporaryAddress";
        NspSoapResponseData response =
                executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request));
        if (response.isFault && !response.content.contains("did not match any current data")) {throw new Exception();}
    }

    private II createId(String idString) {
        II id = new II();
        id.setExtension(idString);
        id.setRoot("1.2.208.184.15.2");
        id.setAssigningAuthorityName("FSK");
        return id;
    }

    public AD createAddress(String streetName) {
        return createAddress("8000", "Aarhus C", "Denmark", streetName);
    }

    public AD createEmptyAddress() {
        AD address = new AD();
        address.getUses().add("H");
        address.setIsNotOrdered(false);
        return address;
    }

    public AD createAddressWithUseablePeriod(String startDate, String endDate, String postalCode, String city,
                                             String country, String... streetNames) {
        AD address = createAddress(postalCode, city, country, streetNames);
        if (startDate != null) {
            addUseablePeriodToAddress(address, startDate);
            if (endDate != null) {
                addUseablePeriodToAddress(address, endDate);
            }
        }
        return address;
    }

    public AD createAddress(String postalCode, String city, String country, String... streetNames) {
        AD address = new AD();
        address.setIsNotOrdered(false);
        address.getUses().add("H");
        if (postalCode != null) {
            AdxpPostalCode postal = new AdxpPostalCode();
            postal.getContent().add(postalCode);
            address.getContent().add(objectFactory.createADPostalCode(postal));
        }
        if (city != null) {
            AdxpCity cityElement = new AdxpCity();
            cityElement.getContent().add(city);
            address.getContent().add(objectFactory.createADCity(cityElement));
        }
        if (country != null) {
            AdxpCountry countryElement = new AdxpCountry();
            countryElement.getContent().add(country);
            address.getContent().add(objectFactory.createADCountry(countryElement));
        }
        for (String streetName : streetNames) {
            AdxpStreetAddressLine streetAddressElement = new AdxpStreetAddressLine();
            streetAddressElement.getContent().add(streetName);
            address.getContent().add(objectFactory.createADStreetAddressLine(streetAddressElement));
        }

        return address;
    }


    public enum ContactType {
        H, WP, MC
    }

    public void addDentist(String cpr) throws Exception {
        CreateHealthProviderRequest request = createCreateHealthProviderRequest(cpr);
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#CreateHealthProvider";
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request)).isFault) {throw new Exception();}
    }

    public CreateHealthProviderRequest createCreateHealthProviderRequest(String cpr) {
        CreateHealthProviderRequest request = new CreateHealthProviderRequest();
        request.setId(createCprIdElement(cpr));
        HealthProvider healthProvider = new HealthProvider();
        CD providerType = new CD();
        providerType.setCode("tandlæge");
        providerType.setCodeSystem("1.2.208.184.15.12");
        providerType.setDisplayName("dentist");
        healthProvider.setProviderType(providerType);
        POCDMT000040Organization organisation = createOrganizationElement();
        organisation.getTelecoms().add(createTel());

        AD address = new AD();
        address.setIsNotOrdered(false);
        address.getUses().add("H");
        AdxpStreetAddressLine streetAddress = new AdxpStreetAddressLine();
        streetAddress.getContent().add("Street Name 2");
        AdxpPostalCode postal = new AdxpPostalCode();
        postal.getContent().add("8000");
        AdxpCity city = new AdxpCity();
        city.getContent().add("Aarhus C");
        AdxpCountry country = new AdxpCountry();
        country.getContent().add("Denmark");
        address.getContent().add(objectFactory.createADStreetAddressLine(streetAddress));
        address.getContent().add(objectFactory.createADPostalCode(postal));
        address.getContent().add(objectFactory.createADCity(city));
        address.getContent().add(objectFactory.createADCountry(country));
        organisation.getAddrs().add(address);

        healthProvider.setOrganization(organisation);
        request.setHealthProvider(healthProvider);
        request.setDataEnterer(createDataEntererElement());
        return request;
    }

    public POCDMT000040Organization createOrganizationElement() {
        POCDMT000040Organization organisation = new POCDMT000040Organization();
        II organisationType = new II();
        organisationType.setAssigningAuthorityName("Yder");
        organisationType.setExtension("410845");
        organisationType.setRoot("1.2.208.176.1.4");
        organisation.getIds().add(organisationType);
        ON organisationName = new ON();
        organisationName.getContent().add("My Dentist");
        organisation.getNames().add(organisationName);
        return organisation;
    }

    public void deleteDentist(String cpr, String dentistIdString) throws Exception {
        if (dentistIdString == null) { return; }
        DeleteHealthProviderRequest request = new DeleteHealthProviderRequest();
        request.setId(createCprIdElement(cpr));

        II healthProviderId = createId(dentistIdString);

        request.setHealthProviderId(healthProviderId);
        request.setDataEnterer(createDataEntererElement());

        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#DeleteHealthProvider";
        NspSoapResponseData response =
                executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request));
        if (response.isFault && !response.content.contains("did not match any current data")) {throw new Exception();}
    }

    public TEL createTel(String contactType, String phone) {
        TEL tel = new TEL(); tel.setValue(phone); tel.getUses().add(contactType); return tel;
    }

    private TEL createTel() {
        return createTel(ContactType.WP.toString(), "tel:911");
    }


    public void deleteRelative(String cpr, String... relatives) throws Exception {
        DeleteRelativesRequest request = new DeleteRelativesRequest();
        request.setId(createCprIdElement(cpr));
        request.setDataEnterer(createDataEntererElement());

        Arrays.stream(relatives).forEach(x -> request.getRelativeIds().add(createRelatedId(x)));

        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#DeleteRelatives";
        if (executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request)).isFault) {throw new Exception();}
    }

    private II createRelatedId(String uuid) {
        II ii = new II(); ii.setExtension(uuid); ii.setRoot("1.2.208.184.15.3"); ii.setAssigningAuthorityName("FSK");
        return ii;
    }

    public CreateRelativesRequest createCreateRelativesRequest(String cpr) {
        CreateRelativesRequest request = new CreateRelativesRequest();
        request.setId(createCprIdElement(cpr));
        request.setDataEnterer(createDataEntererElement());
        RelatedPerson relatedPerson = createRelatedPerson();
        request.getRelatedPersons().add(relatedPerson);
        return request;
    }

    private RelatedPerson createRelatedPerson() {
        RelatedPerson relatedPerson = new RelatedPerson();
        CD relationshipType = new CD();
        relationshipType.setCode("barn");
        relationshipType.setCodeSystem("1.2.208.184.15.4");
        relatedPerson.setRelationshipType(relationshipType);
        POCDMT000040AssociatedEntity associatedEntity = new POCDMT000040AssociatedEntity();
        associatedEntity.getClassCodes().add("CON");

        associatedEntity.getTelecoms().add(createTel());
        associatedEntity.getAddrs().add(createAddress("A relative address"));
        POCDMT000040Person associatedPerson = new POCDMT000040Person();
        associatedPerson.getNames().add(createName("Larsen", "Casper"));

        associatedEntity.setAssociatedPerson(associatedPerson);
        relatedPerson.setAssociatedEntity(associatedEntity);
        return relatedPerson;
    }

    public POCDMT000040ClinicalDocument getSkr(String cpr) throws Exception {
        GetPersonalDataCardRequest request = new GetPersonalDataCardRequest();
        request.setId(createCprIdElement(cpr));
        final String action = "http://sundhedsdatastyrelsen.dk/skr/2024/01/01#GetPersonalDataCard";
        NspSoapResponseData response = executeNspRequest(endpoint, action, lægeCharlesBabbage(), toXml(request));
        if (response.isFault) { throw new Exception(); }
        return parseNspSoapResponse(response, GetPersonalDataCardResponse.class).getClinicalDocument();
    }

    public PN createName(String family, String... given) {
        PN pn = new PN();
        for (String g : given) {
            EnGiven enGiven = new EnGiven();
            enGiven.getContent().add(g);
            pn.getContent().add(new org.hl7.v3.ObjectFactory().createENGiven(enGiven));
        }
        if (family != null) {
            EnFamily enFamily = new EnFamily();
            enFamily.getContent().add(family);
            pn.getContent().add(new org.hl7.v3.ObjectFactory().createENFamily(enFamily));
        }
        return pn;
    }

    public POCDMT000040Author createDataEntererElement() {
        return createDataEntererElement("Charles", "Babbage");
    }

    private POCDMT000040Author createDataEntererElement(String dataEntererGivenName, String dataEntererFamilyName) {
        POCDMT000040Author dataEnterer = new POCDMT000040Author();
        TS ts = new TS();
        ts.setValue(ZonedDateTime.now().format(DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ"))); //
        dataEnterer.setTime(ts);
        POCDMT000040AssignedAuthor assignedAuthor = new POCDMT000040AssignedAuthor();
        assignedAuthor.getIds().add(createCprIdElement(DATAENTERER_ID_CHARLES_BABBAGE));
        POCDMT000040Person assignedPerson = new POCDMT000040Person();
        PN pn = new PN();

        EnGiven enGiven = new EnGiven();
        enGiven.getContent().add(dataEntererGivenName);
        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENGiven(enGiven));

        EnFamily enFamily = new EnFamily();
        enFamily.getContent().add(dataEntererFamilyName);

        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENFamily(enFamily));

        assignedPerson.getNames().add(pn);
        assignedAuthor.setAssignedPerson(assignedPerson);
        dataEnterer.setAssignedAuthor(assignedAuthor);

        return dataEnterer;
    }

    public II createCprIdElement(String cpr) {
        return createIdElement("CPR", cpr, OID.CPR);
    }

    public II createRelativeIdElement(String id) {
        return createIdElement(null, id, OID.RELATED);
    }

    public II createTemporaryAddressIdElement(String id) {
        return createIdElement(null, id, OID.TMP_ADDRESS_OID);
    }

    public II createLanguageIdElement(String id) {
        return createIdElement(null, id, OID.LANGUAGE_OID);
    }

    public II createHealthProviderIdElement(String id) {
        return createIdElement(null, id, OID.HEALTH_PROVIDER);
    }

    private II createIdElement(String assigningAuthorityName, String extension, String root) {
        II ii = new II();
        ii.setAssigningAuthorityName(assigningAuthorityName);
        ii.setExtension(extension);
        ii.setRoot(root);
        return ii;
    }
}
